/* Instruction scheduling pass.
   Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997, 1998,
   1999, 2000, 2001 Free Software Foundation, Inc.
   Contributed by Michael Tiemann (tiemann@cygnus.com) Enhanced by,
   and currently maintained by, Jim Wilson (wilson@cygnus.com)

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

GNU CC is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to the Free
the Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

/* This pass implements list scheduling within basic blocks.  It is
   run twice: (1) after flow analysis, but before register allocation,
   and (2) after register allocation.

   The first run performs interblock scheduling, moving insns between
   different blocks in the same "region", and the second runs only
   basic block scheduling.

   Interblock motions performed are useful motions and speculative
   motions, including speculative loads.  Motions requiring code
   duplication are not supported.  The identification of motion type
   and the check for validity of speculative motions requires
   construction and analysis of the function's control flow graph.

   The main entry point for this pass is schedule_insns(), called for
   each function.  The work of the scheduler is organized in three
   levels: (1) function level: insns are subject to splitting,
   control-flow-graph is constructed, regions are computed (after
   reload, each region is of one block), (2) region level: control
   flow graph attributes required for interblock scheduling are
   computed (dominators, reachability, etc.), data dependences and
   priorities are computed, and (3) block level: insns in the block
   are actually scheduled.  */

#include "config.h"
#include "system.h"
#include "toplev.h"
#include "rtl.h"
#include "tm_p.h"
#include "hard-reg-set.h"
#include "basic-block.h"
#include "regs.h"
#include "function.h"
#include "flags.h"
#include "insn-config.h"
#include "insn-attr.h"
#include "except.h"
#include "toplev.h"
#include "recog.h"
#include "sched-int.h"
/* CYGNUS LOCAL -- vmakarov/RCSP */
#include "output.h"
/* END CYGNUS LOCAL -- vmakarov/RCSP */

#ifdef INSN_SCHEDULING
/* Some accessor macros for h_i_d members only used within this file.  */
#define INSN_REF_COUNT(INSN)	(h_i_d[INSN_UID (INSN)].ref_count)
#define FED_BY_SPEC_LOAD(insn)	(h_i_d[INSN_UID (insn)].fed_by_spec_load)
#define IS_LOAD_INSN(insn)	(h_i_d[INSN_UID (insn)].is_load_insn)

#define MAX_RGN_BLOCKS 10
#define MAX_RGN_INSNS 100

/* nr_inter/spec counts interblock/speculative motion for the function.  */
static int nr_inter, nr_spec;

/* Control flow graph edges are kept in circular lists.  */
typedef struct
{
  int from_block;
  int to_block;
  int next_in;
  int next_out;
}
haifa_edge;
static haifa_edge *edge_table;

#define NEXT_IN(edge) (edge_table[edge].next_in)
#define NEXT_OUT(edge) (edge_table[edge].next_out)
#define FROM_BLOCK(edge) (edge_table[edge].from_block)
#define TO_BLOCK(edge) (edge_table[edge].to_block)

/* Number of edges in the control flow graph.  (In fact, larger than
   that by 1, since edge 0 is unused.)  */
static int nr_edges;

/* Circular list of incoming/outgoing edges of a block.  */
static int *in_edges;
static int *out_edges;

#define IN_EDGES(block) (in_edges[block])
#define OUT_EDGES(block) (out_edges[block])

static int is_cfg_nonregular PARAMS ((void));
static int build_control_flow PARAMS ((struct edge_list *));
static void new_edge PARAMS ((int, int));

/* A region is the main entity for interblock scheduling: insns
   are allowed to move between blocks in the same region, along
   control flow graph edges, in the 'up' direction.  */
typedef struct
{
  int rgn_nr_blocks;		/* Number of blocks in region.  */
  int rgn_blocks;		/* cblocks in the region (actually index in rgn_bb_table).  */
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
  char rgn_may_be_pipelined_p;	/* True if region is loop may be pipelined.  */
  char rgn_do_not_schedule_p;	/* TRUE if rgn is in pipelined loop.  */
  rtx rgn_pipelined_loop;	/* pipelined loop or NULL if failure.  */
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
}
region;

/* Number of regions in the procedure.  */
static int nr_regions;

/* Table of region descriptions.  */
static region *rgn_table;

/* Array of lists of regions' blocks.  */
static int *rgn_bb_table;

/* Topological order of blocks in the region (if b2 is reachable from
   b1, block_to_bb[b2] > block_to_bb[b1]).  Note: A basic block is
   always referred to by either block or b, while its topological
   order name (in the region) is refered to by bb.  */
static int *block_to_bb;

/* The number of the region containing a block.  */
static int *containing_rgn;

#define RGN_NR_BLOCKS(rgn) (rgn_table[rgn].rgn_nr_blocks)
#define RGN_BLOCKS(rgn) (rgn_table[rgn].rgn_blocks)
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
#define RGN_MAY_BE_PIPELINED_P(rgn) (rgn_table[rgn].rgn_may_be_pipelined_p)
#define RGN_PIPELINED_LOOP(rgn) (rgn_table[rgn].rgn_pipelined_loop)
#define RGN_DO_NOT_SCHEDULE_P(rgn) (rgn_table[rgn].rgn_do_not_schedule_p)
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
#define BLOCK_TO_BB(block) (block_to_bb[block])
#define CONTAINING_RGN(block) (containing_rgn[block])

void debug_regions PARAMS ((void));
static void find_single_block_region PARAMS ((void));
static void find_rgns PARAMS ((struct edge_list *, sbitmap *));
static int too_large PARAMS ((int, int *, int *));

extern void debug_live PARAMS ((int, int));

/* Blocks of the current region being scheduled.  */
static int current_nr_blocks;
static int current_blocks;

/* The mapping from bb to block.  */
#define BB_TO_BLOCK(bb) (rgn_bb_table[current_blocks + (bb)])

/* Bit vectors and bitset operations are needed for computations on
   the control flow graph.  */

typedef unsigned HOST_WIDE_INT *bitset;
typedef struct
{
  int *first_member;		/* Pointer to the list start in bitlst_table.  */
  int nr_members;		/* The number of members of the bit list.  */
}
bitlst;

static int bitlst_table_last;
static int bitlst_table_size;
static int *bitlst_table;

static char bitset_member PARAMS ((bitset, int, int));
static void extract_bitlst PARAMS ((bitset, int, int, bitlst *));

/* Target info declarations.

   The block currently being scheduled is referred to as the "target" block,
   while other blocks in the region from which insns can be moved to the
   target are called "source" blocks.  The candidate structure holds info
   about such sources: are they valid?  Speculative?  Etc.  */
typedef bitlst bblst;
typedef struct
{
  char is_valid;
  char is_speculative;
  int src_prob;
  bblst split_bbs;
  bblst update_bbs;
}
candidate;

static candidate *candidate_table;

/* A speculative motion requires checking live information on the path
   from 'source' to 'target'.  The split blocks are those to be checked.
   After a speculative motion, live information should be modified in
   the 'update' blocks.

   Lists of split and update blocks for each candidate of the current
   target are in array bblst_table.  */
static int *bblst_table, bblst_size, bblst_last;

#define IS_VALID(src) ( candidate_table[src].is_valid )
#define IS_SPECULATIVE(src) ( candidate_table[src].is_speculative )
#define SRC_PROB(src) ( candidate_table[src].src_prob )

/* The bb being currently scheduled.  */
static int target_bb;

/* List of edges.  */
typedef bitlst edgelst;

/* Target info functions.  */
static void split_edges PARAMS ((int, int, edgelst *));
static void compute_trg_info PARAMS ((int));
void debug_candidate PARAMS ((int));
void debug_candidates PARAMS ((int));

/* Bit-set of bbs, where bit 'i' stands for bb 'i'.  */
typedef bitset bbset;

/* Number of words of the bbset.  */
static int bbset_size;

/* Dominators array: dom[i] contains the bbset of dominators of
   bb i in the region.  */
static bbset *dom;

/* bb 0 is the only region entry.  */
#define IS_RGN_ENTRY(bb) (!bb)

/* Is bb_src dominated by bb_trg.  */
#define IS_DOMINATED(bb_src, bb_trg)                                 \
( bitset_member (dom[bb_src], bb_trg, bbset_size) )

/* Probability: Prob[i] is a float in [0, 1] which is the probability
   of bb i relative to the region entry.  */
static float *prob;

/* The probability of bb_src, relative to bb_trg.  Note, that while the
   'prob[bb]' is a float in [0, 1], this macro returns an integer
   in [0, 100].  */
#define GET_SRC_PROB(bb_src, bb_trg) ((int) (100.0 * (prob[bb_src] / \
						      prob[bb_trg])))

/* Bit-set of edges, where bit i stands for edge i.  */
typedef bitset edgeset;

/* Number of edges in the region.  */
static int rgn_nr_edges;

/* Array of size rgn_nr_edges.  */
static int *rgn_edges;

/* Number of words in an edgeset.  */
static int edgeset_size;

/* Number of bits in an edgeset.  */
static int edgeset_bitsize;

/* Mapping from each edge in the graph to its number in the rgn.  */
static int *edge_to_bit;
#define EDGE_TO_BIT(edge) (edge_to_bit[edge])

/* The split edges of a source bb is different for each target
   bb.  In order to compute this efficiently, the 'potential-split edges'
   are computed for each bb prior to scheduling a region.  This is actually
   the split edges of each bb relative to the region entry.

   pot_split[bb] is the set of potential split edges of bb.  */
static edgeset *pot_split;

/* For every bb, a set of its ancestor edges.  */
static edgeset *ancestor_edges;

static void compute_dom_prob_ps PARAMS ((int));

#define ABS_VALUE(x) (((x)<0)?(-(x)):(x))
#define INSN_PROBABILITY(INSN) (SRC_PROB (BLOCK_TO_BB (BLOCK_NUM (INSN))))
#define IS_SPECULATIVE_INSN(INSN) (IS_SPECULATIVE (BLOCK_TO_BB (BLOCK_NUM (INSN))))
#define INSN_BB(INSN) (BLOCK_TO_BB (BLOCK_NUM (INSN)))

/* Parameters affecting the decision of rank_for_schedule().
   ??? Nope.  But MIN_PROBABILITY is used in copmute_trg_info.  */
#define MIN_DIFF_PRIORITY 2
#define MIN_PROBABILITY 40
#define MIN_PROB_DIFF 10

/* Speculative scheduling functions.  */
static int check_live_1 PARAMS ((int, rtx));
static void update_live_1 PARAMS ((int, rtx));
static int check_live PARAMS ((rtx, int));
static void update_live PARAMS ((rtx, int));
static void set_spec_fed PARAMS ((rtx));
static int is_pfree PARAMS ((rtx, int, int));
static int find_conditional_protection PARAMS ((rtx, int));
static int is_conditionally_protected PARAMS ((rtx, int, int));
static int may_trap_exp PARAMS ((rtx, int));
static int haifa_classify_insn PARAMS ((rtx));
static int is_prisky PARAMS ((rtx, int, int));
static int is_exception_free PARAMS ((rtx, int, int));

static void add_branch_dependences PARAMS ((rtx, rtx));
static void compute_block_backward_dependences PARAMS ((int));
void debug_dependencies PARAMS ((void));

static void init_regions PARAMS ((void));
/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
static void remove_new_cpu_cycle_marks PARAMS ((int));
#endif
/* END CYGNUS LOCAL -- vmakarov/dfa */
static void schedule_region PARAMS ((int));
static void propagate_deps PARAMS ((int, struct deps *, int));
static void free_pending_lists PARAMS ((void));
/* CYGNUS LOCAL -- vmakarov/RCSP */
/* The following is for calling software pipeliner. */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
static int may_be_pipelined_p PARAMS ((rtx, rtx));
static void start_RCSP_interface PARAMS ((rtx, rtx, int));
static rtx call_RCSP PARAMS ((FILE *));
static void finish_RCSP_interface PARAMS ((void));
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */

/* Functions for construction of the control flow graph.  */

/* Return 1 if control flow graph should not be constructed, 0 otherwise.

   We decide not to build the control flow graph if there is possibly more
   than one entry to the function, if computed branches exist, of if we
   have nonlocal gotos.  */

static int
is_cfg_nonregular ()
{
  int b;
  rtx insn;
  RTX_CODE code;

  /* If we have a label that could be the target of a nonlocal goto, then
     the cfg is not well structured.  */
  if (nonlocal_goto_handler_labels)
    return 1;

  /* If we have any forced labels, then the cfg is not well structured.  */
  if (forced_labels)
    return 1;

  /* If this function has a computed jump, then we consider the cfg
     not well structured.  */
  if (current_function_has_computed_jump)
    return 1;

  /* If we have exception handlers, then we consider the cfg not well
     structured.  ?!?  We should be able to handle this now that flow.c
     computes an accurate cfg for EH.  */
  if (exception_handler_labels)
    return 1;

  /* If we have non-jumping insns which refer to labels, then we consider
     the cfg not well structured.  */
  /* Check for labels referred to other thn by jumps.  */
  for (b = 0; b < n_basic_blocks; b++)
    for (insn = BLOCK_HEAD (b);; insn = NEXT_INSN (insn))
      {
	code = GET_CODE (insn);
	if (GET_RTX_CLASS (code) == 'i' && code != JUMP_INSN)
	  {
	    rtx note = find_reg_note (insn, REG_LABEL, NULL_RTX);

	    if (note
		&& ! (GET_CODE (NEXT_INSN (insn)) == JUMP_INSN
		      && find_reg_note (NEXT_INSN (insn), REG_LABEL,
					XEXP (note, 0))))
	      return 1;
	  }

	if (insn == BLOCK_END (b))
	  break;
      }

  /* All the tests passed.  Consider the cfg well structured.  */
  return 0;
}

/* Build the control flow graph and set nr_edges.

   Instead of trying to build a cfg ourselves, we rely on flow to
   do it for us.  Stamp out useless code (and bug) duplication.

   Return nonzero if an irregularity in the cfg is found which would
   prevent cross block scheduling.  */

static int
build_control_flow (edge_list)
     struct edge_list *edge_list;
{
  int i, unreachable, num_edges;

  /* This already accounts for entry/exit edges.  */
  num_edges = NUM_EDGES (edge_list);

  /* Unreachable loops with more than one basic block are detected
     during the DFS traversal in find_rgns.

     Unreachable loops with a single block are detected here.  This
     test is redundant with the one in find_rgns, but it's much
    cheaper to go ahead and catch the trivial case here.  */
  unreachable = 0;
  for (i = 0; i < n_basic_blocks; i++)
    {
      basic_block b = BASIC_BLOCK (i);

      if (b->pred == NULL
	  || (b->pred->src == b
	      && b->pred->pred_next == NULL))
	unreachable = 1;
    }

  /* ??? We can kill these soon.  */
  in_edges = (int *) xcalloc (n_basic_blocks, sizeof (int));
  out_edges = (int *) xcalloc (n_basic_blocks, sizeof (int));
  edge_table = (haifa_edge *) xcalloc (num_edges, sizeof (haifa_edge));

  nr_edges = 0;
  for (i = 0; i < num_edges; i++)
    {
      edge e = INDEX_EDGE (edge_list, i);

      if (e->dest != EXIT_BLOCK_PTR
	  && e->src != ENTRY_BLOCK_PTR)
	new_edge (e->src->index, e->dest->index);
    }

  /* Increment by 1, since edge 0 is unused.  */
  nr_edges++;

  return unreachable;
}

/* Record an edge in the control flow graph from SOURCE to TARGET.

   In theory, this is redundant with the s_succs computed above, but
   we have not converted all of haifa to use information from the
   integer lists.  */

static void
new_edge (source, target)
     int source, target;
{
  int e, next_edge;
  int curr_edge, fst_edge;

  /* Check for duplicates.  */
  fst_edge = curr_edge = OUT_EDGES (source);
  while (curr_edge)
    {
      if (FROM_BLOCK (curr_edge) == source
	  && TO_BLOCK (curr_edge) == target)
	{
	  return;
	}

      curr_edge = NEXT_OUT (curr_edge);

      if (fst_edge == curr_edge)
	break;
    }

  e = ++nr_edges;

  FROM_BLOCK (e) = source;
  TO_BLOCK (e) = target;

  if (OUT_EDGES (source))
    {
      next_edge = NEXT_OUT (OUT_EDGES (source));
      NEXT_OUT (OUT_EDGES (source)) = e;
      NEXT_OUT (e) = next_edge;
    }
  else
    {
      OUT_EDGES (source) = e;
      NEXT_OUT (e) = e;
    }

  if (IN_EDGES (target))
    {
      next_edge = NEXT_IN (IN_EDGES (target));
      NEXT_IN (IN_EDGES (target)) = e;
      NEXT_IN (e) = next_edge;
    }
  else
    {
      IN_EDGES (target) = e;
      NEXT_IN (e) = e;
    }
}

/* BITSET macros for operations on the control flow graph.  */

/* Compute bitwise union of two bitsets.  */
#define BITSET_UNION(set1, set2, len)                                \
do { register bitset tp = set1, sp = set2;                           \
     register int i;                                                 \
     for (i = 0; i < len; i++)                                       \
       *(tp++) |= *(sp++); } while (0)

/* Compute bitwise intersection of two bitsets.  */
#define BITSET_INTER(set1, set2, len)                                \
do { register bitset tp = set1, sp = set2;                           \
     register int i;                                                 \
     for (i = 0; i < len; i++)                                       \
       *(tp++) &= *(sp++); } while (0)

/* Compute bitwise difference of two bitsets.  */
#define BITSET_DIFFER(set1, set2, len)                               \
do { register bitset tp = set1, sp = set2;                           \
     register int i;                                                 \
     for (i = 0; i < len; i++)                                       \
       *(tp++) &= ~*(sp++); } while (0)

/* Inverts every bit of bitset 'set'.  */
#define BITSET_INVERT(set, len)                                      \
do { register bitset tmpset = set;                                   \
     register int i;                                                 \
     for (i = 0; i < len; i++, tmpset++)                             \
       *tmpset = ~*tmpset; } while (0)

/* Turn on the index'th bit in bitset set.  */
#define BITSET_ADD(set, index, len)                                  \
{                                                                    \
  if (index >= HOST_BITS_PER_WIDE_INT * len)                         \
    abort ();                                                        \
  else                                                               \
    set[index/HOST_BITS_PER_WIDE_INT] |=			     \
      1 << (index % HOST_BITS_PER_WIDE_INT);                         \
}

/* Turn off the index'th bit in set.  */
#define BITSET_REMOVE(set, index, len)                               \
{                                                                    \
  if (index >= HOST_BITS_PER_WIDE_INT * len)                         \
    abort ();                                                        \
  else                                                               \
    set[index/HOST_BITS_PER_WIDE_INT] &=			     \
      ~(1 << (index%HOST_BITS_PER_WIDE_INT));                        \
}

/* Check if the index'th bit in bitset set is on.  */

static char
bitset_member (set, index, len)
     bitset set;
     int index, len;
{
  if (index >= HOST_BITS_PER_WIDE_INT * len)
    abort ();
  return (set[index / HOST_BITS_PER_WIDE_INT] &
	  1 << (index % HOST_BITS_PER_WIDE_INT)) ? 1 : 0;
}

/* Translate a bit-set SET to a list BL of the bit-set members.  */

static void
extract_bitlst (set, len, bitlen, bl)
     bitset set;
     int len;
     int bitlen;
     bitlst *bl;
{
  int i, j, offset;
  unsigned HOST_WIDE_INT word;

  /* bblst table space is reused in each call to extract_bitlst.  */
  bitlst_table_last = 0;

  bl->first_member = &bitlst_table[bitlst_table_last];
  bl->nr_members = 0;

  /* Iterate over each word in the bitset.  */
  for (i = 0; i < len; i++)
    {
      word = set[i];
      offset = i * HOST_BITS_PER_WIDE_INT;

      /* Iterate over each bit in the word, but do not
	 go beyond the end of the defined bits.  */
      for (j = 0; offset < bitlen && word; j++)
	{
	  if (word & 1)
	    {
	      bitlst_table[bitlst_table_last++] = offset;
	      (bl->nr_members)++;
	    }
	  word >>= 1;
	  ++offset;
	}
    }

}

/* Functions for the construction of regions.  */

/* Print the regions, for debugging purposes.  Callable from debugger.  */

void
debug_regions ()
{
  int rgn, bb;

  fprintf (sched_dump, "\n;;   ------------ REGIONS ----------\n\n");
  for (rgn = 0; rgn < nr_regions; rgn++)
    {
      fprintf (sched_dump, ";;\trgn %d nr_blocks %d:\n", rgn,
	       rgn_table[rgn].rgn_nr_blocks);
      fprintf (sched_dump, ";;\tbb/block: ");

      for (bb = 0; bb < rgn_table[rgn].rgn_nr_blocks; bb++)
	{
	  current_blocks = RGN_BLOCKS (rgn);

	  if (bb != BLOCK_TO_BB (BB_TO_BLOCK (bb)))
	    abort ();

	  fprintf (sched_dump, " %d/%d ", bb, BB_TO_BLOCK (bb));
	}

      fprintf (sched_dump, "\n\n");
    }
}

/* Build a single block region for each basic block in the function.
   This allows for using the same code for interblock and basic block
   scheduling.  */

static void
find_single_block_region ()
{
  int i;
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
  rtx first_bb_insn, last_bb_insn, insn;
  int in_pipelined_loop_p = 0;
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */

  for (i = 0; i < n_basic_blocks; i++)
    {
      rgn_bb_table[i] = i;
      RGN_NR_BLOCKS (i) = 1;
      RGN_BLOCKS (i) = i;
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
      RGN_MAY_BE_PIPELINED_P (i) = 0;
      RGN_PIPELINED_LOOP (i) = NULL_RTX;
      RGN_DO_NOT_SCHEDULE_P (i) = 0;
      if (USE_AUTOMATON_PIPELINE_INTERFACE)
	{
	  if (reload_completed && i != 0)
	    {
	      /* Searching for the note about software pipelined loop
		 start between basic blocks. */
	      get_block_head_tail (i - 1, &first_bb_insn, &last_bb_insn);
	      if (!in_pipelined_loop_p)
		{
		  for (insn = last_bb_insn;
		       insn != NULL && GET_CODE (insn) != NOTE;
		       insn = NEXT_INSN (insn))
		    ;
		  if (insn != NULL
		      && (NOTE_LINE_NUMBER (insn)
			  == NOTE_SOFTWARE_PIPELINED_LOOP_BEGIN))
		    in_pipelined_loop_p = 1;
		}
	    }
	  
	  RGN_DO_NOT_SCHEDULE_P (i) = in_pipelined_loop_p;
	  
	  if (reload_completed && !in_pipelined_loop_p)
	    {
	      /* The note may be also inside a basic block. */
	      get_block_head_tail (i, &first_bb_insn, &last_bb_insn);
	      for (insn = first_bb_insn; ; insn = NEXT_INSN (insn))
		{
		  if (GET_CODE (insn) == NOTE
		      && (NOTE_LINE_NUMBER (insn)
			  == NOTE_SOFTWARE_PIPELINED_LOOP_BEGIN))
		    {
		      in_pipelined_loop_p = 1;
		      break;
		    }
		  if (insn == last_bb_insn)
		    break;
		}
	    }
	  
	  if (reload_completed && in_pipelined_loop_p)
	    {
	      for (insn = last_bb_insn;
		   insn != NULL && GET_CODE (insn) != NOTE;
		   insn = NEXT_INSN (insn))
		;
	      if (insn != NULL
		  && (NOTE_LINE_NUMBER (insn)
		      == NOTE_SOFTWARE_PIPELINED_LOOP_END))
		in_pipelined_loop_p = 0;
	    }
	}
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
      CONTAINING_RGN (i) = i;
      BLOCK_TO_BB (i) = 0;
    }
  nr_regions = n_basic_blocks;
}

/* Update number of blocks and the estimate for number of insns
   in the region.  Return 1 if the region is "too large" for interblock
   scheduling (compile time considerations), otherwise return 0.  */

static int
too_large (block, num_bbs, num_insns)
     int block, *num_bbs, *num_insns;
{
  (*num_bbs)++;
  (*num_insns) += (INSN_LUID (BLOCK_END (block)) -
		   INSN_LUID (BLOCK_HEAD (block)));
  if ((*num_bbs > MAX_RGN_BLOCKS) || (*num_insns > MAX_RGN_INSNS))
    return 1;
  else
    return 0;
}

/* Update_loop_relations(blk, hdr): Check if the loop headed by max_hdr[blk]
   is still an inner loop.  Put in max_hdr[blk] the header of the most inner
   loop containing blk.  */
#define UPDATE_LOOP_RELATIONS(blk, hdr)                              \
{                                                                    \
  if (max_hdr[blk] == -1)                                            \
    max_hdr[blk] = hdr;                                              \
  else if (dfs_nr[max_hdr[blk]] > dfs_nr[hdr])                       \
         RESET_BIT (inner, hdr);                                     \
  else if (dfs_nr[max_hdr[blk]] < dfs_nr[hdr])                       \
         {                                                           \
            RESET_BIT (inner,max_hdr[blk]);			     \
            max_hdr[blk] = hdr;                                      \
         }                                                           \
}

/* Find regions for interblock scheduling.

   A region for scheduling can be:

     * A loop-free procedure, or

     * A reducible inner loop, or

     * A basic block not contained in any other region.

   ?!? In theory we could build other regions based on extended basic
   blocks or reverse extended basic blocks.  Is it worth the trouble?

   Loop blocks that form a region are put into the region's block list
   in topological order.

   This procedure stores its results into the following global (ick) variables

     * rgn_nr
     * rgn_table
     * rgn_bb_table
     * block_to_bb
     * containing region

   We use dominator relationships to avoid making regions out of non-reducible
   loops.

   This procedure needs to be converted to work on pred/succ lists instead
   of edge tables.  That would simplify it somewhat.  */

static void
find_rgns (edge_list, dom)
     struct edge_list *edge_list;
     sbitmap *dom;
{
  int *max_hdr, *dfs_nr, *stack, *degree;
  char no_loops = 1;
  int node, child, loop_head, i, head, tail;
  int count = 0, sp, idx = 0, current_edge = out_edges[0];
  int num_bbs, num_insns, unreachable;
  int too_large_failure;

  /* Note if an edge has been passed.  */
  sbitmap passed;

  /* Note if a block is a natural loop header.  */
  sbitmap header;

  /* Note if a block is an natural inner loop header.  */
  sbitmap inner;

  /* Note if a block is in the block queue.  */
  sbitmap in_queue;

  /* Note if a block is in the block queue.  */
  sbitmap in_stack;

  int num_edges = NUM_EDGES (edge_list);

  /* Perform a DFS traversal of the cfg.  Identify loop headers, inner loops
     and a mapping from block to its loop header (if the block is contained
     in a loop, else -1).

     Store results in HEADER, INNER, and MAX_HDR respectively, these will
     be used as inputs to the second traversal.

     STACK, SP and DFS_NR are only used during the first traversal.  */

  /* Allocate and initialize variables for the first traversal.  */
  max_hdr = (int *) xmalloc (n_basic_blocks * sizeof (int));
  dfs_nr = (int *) xcalloc (n_basic_blocks, sizeof (int));
  stack = (int *) xmalloc (nr_edges * sizeof (int));

  inner = sbitmap_alloc (n_basic_blocks);
  sbitmap_ones (inner);

  header = sbitmap_alloc (n_basic_blocks);
  sbitmap_zero (header);

  passed = sbitmap_alloc (nr_edges);
  sbitmap_zero (passed);

  in_queue = sbitmap_alloc (n_basic_blocks);
  sbitmap_zero (in_queue);

  in_stack = sbitmap_alloc (n_basic_blocks);
  sbitmap_zero (in_stack);

  for (i = 0; i < n_basic_blocks; i++)
    max_hdr[i] = -1;

  /* DFS traversal to find inner loops in the cfg.  */

  sp = -1;
  while (1)
    {
      if (current_edge == 0 || TEST_BIT (passed, current_edge))
	{
	  /* We have reached a leaf node or a node that was already
	     processed.  Pop edges off the stack until we find
	     an edge that has not yet been processed.  */
	  while (sp >= 0
		 && (current_edge == 0 || TEST_BIT (passed, current_edge)))
	    {
	      /* Pop entry off the stack.  */
	      current_edge = stack[sp--];
	      node = FROM_BLOCK (current_edge);
	      child = TO_BLOCK (current_edge);
	      RESET_BIT (in_stack, child);
	      if (max_hdr[child] >= 0 && TEST_BIT (in_stack, max_hdr[child]))
		UPDATE_LOOP_RELATIONS (node, max_hdr[child]);
	      current_edge = NEXT_OUT (current_edge);
	    }

	  /* See if have finished the DFS tree traversal.  */
	  if (sp < 0 && TEST_BIT (passed, current_edge))
	    break;

	  /* Nope, continue the traversal with the popped node.  */
	  continue;
	}

      /* Process a node.  */
      node = FROM_BLOCK (current_edge);
      child = TO_BLOCK (current_edge);
      SET_BIT (in_stack, node);
      dfs_nr[node] = ++count;

      /* If the successor is in the stack, then we've found a loop.
	 Mark the loop, if it is not a natural loop, then it will
	 be rejected during the second traversal.  */
      if (TEST_BIT (in_stack, child))
	{
	  no_loops = 0;
	  SET_BIT (header, child);
	  UPDATE_LOOP_RELATIONS (node, child);
	  SET_BIT (passed, current_edge);
	  current_edge = NEXT_OUT (current_edge);
	  continue;
	}

      /* If the child was already visited, then there is no need to visit
	 it again.  Just update the loop relationships and restart
	 with a new edge.  */
      if (dfs_nr[child])
	{
	  if (max_hdr[child] >= 0 && TEST_BIT (in_stack, max_hdr[child]))
	    UPDATE_LOOP_RELATIONS (node, max_hdr[child]);
	  SET_BIT (passed, current_edge);
	  current_edge = NEXT_OUT (current_edge);
	  continue;
	}

      /* Push an entry on the stack and continue DFS traversal.  */
      stack[++sp] = current_edge;
      SET_BIT (passed, current_edge);
      current_edge = OUT_EDGES (child);

      /* This is temporary until haifa is converted to use rth's new
	 cfg routines which have true entry/exit blocks and the
	 appropriate edges from/to those blocks.

	 Generally we update dfs_nr for a node when we process its
	 out edge.  However, if the node has no out edge then we will
	 not set dfs_nr for that node.  This can confuse the scheduler
	 into thinking that we have unreachable blocks, which in turn
	 disables cross block scheduling.

	 So, if we have a node with no out edges, go ahead and mark it
	 as reachable now.  */
      if (current_edge == 0)
	dfs_nr[child] = ++count;
    }

  /* Another check for unreachable blocks.  The earlier test in
     is_cfg_nonregular only finds unreachable blocks that do not
     form a loop.

     The DFS traversal will mark every block that is reachable from
     the entry node by placing a nonzero value in dfs_nr.  Thus if
     dfs_nr is zero for any block, then it must be unreachable.  */
  unreachable = 0;
  for (i = 0; i < n_basic_blocks; i++)
    if (dfs_nr[i] == 0)
      {
	unreachable = 1;
	break;
      }

  /* Gross.  To avoid wasting memory, the second pass uses the dfs_nr array
     to hold degree counts.  */
  degree = dfs_nr;

  for (i = 0; i < n_basic_blocks; i++)
    degree[i] = 0;
  for (i = 0; i < num_edges; i++)
    {
      edge e = INDEX_EDGE (edge_list, i);

      if (e->dest != EXIT_BLOCK_PTR)
	degree[e->dest->index]++;
    }

  /* Do not perform region scheduling if there are any unreachable
     blocks.  */
  if (!unreachable)
    {
      int *queue;

      if (no_loops)
	SET_BIT (header, 0);

      /* Second travsersal:find reducible inner loops and topologically sort
	 block of each region.  */

      queue = (int *) xmalloc (n_basic_blocks * sizeof (int));

      /* Find blocks which are inner loop headers.  We still have non-reducible
	 loops to consider at this point.  */
      for (i = 0; i < n_basic_blocks; i++)
	{
	  if (TEST_BIT (header, i) && TEST_BIT (inner, i))
	    {
	      edge e;
	      int j;
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
	      int bad_loop_p = 0;
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */

	      /* Now check that the loop is reducible.  We do this separate
		 from finding inner loops so that we do not find a reducible
		 loop which contains an inner non-reducible loop.

		 A simple way to find reducible/natural loops is to verify
		 that each block in the loop is dominated by the loop
		 header.

		 If there exists a block that is not dominated by the loop
		 header, then the block is reachable from outside the loop
		 and thus the loop is not a natural loop.  */
	      for (j = 0; j < n_basic_blocks; j++)
		{
		  /* First identify blocks in the loop, except for the loop
		     entry block.  */
		  if (i == max_hdr[j] && i != j)
		    {
		      /* Now verify that the block is dominated by the loop
			 header.  */
		      if (!TEST_BIT (dom[j], i))
			break;
		    }
		}

	      /* If we exited the loop early, then I is the header of
		 a non-reducible loop and we should quit processing it
		 now.  */
	      if (j != n_basic_blocks)
		continue;

	      /* I is a header of an inner loop, or block 0 in a subroutine
		 with no loops at all.  */
	      head = tail = -1;
	      too_large_failure = 0;
	      loop_head = max_hdr[i];

	      /* Decrease degree of all I's successors for topological
		 ordering.  */
	      for (e = BASIC_BLOCK (i)->succ; e; e = e->succ_next)
		if (e->dest != EXIT_BLOCK_PTR)
		  --degree[e->dest->index];

	      /* Estimate # insns, and count # blocks in the region.  */
	      num_bbs = 1;
	      num_insns = (INSN_LUID (BLOCK_END (i))
			   - INSN_LUID (BLOCK_HEAD (i)));

	      /* Find all loop latches (blocks with back edges to the loop
		 header) or all the leaf blocks in the cfg has no loops.

		 Place those blocks into the queue.  */
	      if (no_loops)
		{
		  for (j = 0; j < n_basic_blocks; j++)
		    /* Leaf nodes have only a single successor which must
		       be EXIT_BLOCK.  */
		    if (BASIC_BLOCK (j)->succ
			&& BASIC_BLOCK (j)->succ->dest == EXIT_BLOCK_PTR
			&& BASIC_BLOCK (j)->succ->succ_next == NULL)
		      {
			queue[++tail] = j;
			SET_BIT (in_queue, j);

			if (too_large (j, &num_bbs, &num_insns))
			  {
			    too_large_failure = 1;
			    break;
			  }
		      }
		}
	      else
		{
		  edge e;

		  for (e = BASIC_BLOCK (i)->pred; e; e = e->pred_next)
		    {
		      if (e->src == ENTRY_BLOCK_PTR)
			continue;

		      node = e->src->index;

		      if (max_hdr[node] == loop_head && node != i)
			{
			  /* This is a loop latch.  */
			  queue[++tail] = node;
			  SET_BIT (in_queue, node);

			  if (too_large (node, &num_bbs, &num_insns))
			    {
			      too_large_failure = 1;
			      break;
			    }
			}
		    }
		}

	      /* Now add all the blocks in the loop to the queue.

	     We know the loop is a natural loop; however the algorithm
	     above will not always mark certain blocks as being in the
	     loop.  Consider:
		node   children
		 a	  b,c
		 b	  c
		 c	  a,d
		 d	  b

	     The algorithm in the DFS traversal may not mark B & D as part
	     of the loop (ie they will not have max_hdr set to A).

	     We know they can not be loop latches (else they would have
	     had max_hdr set since they'd have a backedge to a dominator
	     block).  So we don't need them on the initial queue.

	     We know they are part of the loop because they are dominated
	     by the loop header and can be reached by a backwards walk of
	     the edges starting with nodes on the initial queue.

	     It is safe and desirable to include those nodes in the
	     loop/scheduling region.  To do so we would need to decrease
	     the degree of a node if it is the target of a backedge
	     within the loop itself as the node is placed in the queue.

	     We do not do this because I'm not sure that the actual
	     scheduling code will properly handle this case. ?!? */

	      while (head < tail && !too_large_failure)
		{
		  edge e;
		  child = queue[++head];

		  for (e = BASIC_BLOCK (child)->pred; e; e = e->pred_next)
		    {
		      node = e->src->index;

		      /* See discussion above about nodes not marked as in
			 this loop during the initial DFS traversal.  */
		      if (e->src == ENTRY_BLOCK_PTR
			  || max_hdr[node] != loop_head)
			{
			  tail = -1;
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
			  /* We mark loop as bad because rgn_bb_table
			     elements are not set up for all loop
			     basic blocks. ??? We could fix this and
			     also pipeline such loops.  */
			  bad_loop_p = 1;
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
			  break;
			}
		      else if (!TEST_BIT (in_queue, node) && node != i)
			{
			  queue[++tail] = node;
			  SET_BIT (in_queue, node);

			  if (too_large (node, &num_bbs, &num_insns))
			    {
			      too_large_failure = 1;
			      break;
			    }
			}
		    }
		}

/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
	      /* We ignore tail here because otherwise we could not
		 recognize one block loop.  */
	      if (!too_large_failure && !bad_loop_p)
#else
/* END CYGNUS LOCAL -- vmakarov/RCSP */
	      if (tail >= 0 && !too_large_failure)
/* CYGNUS LOCAL -- vmakarov/RCSP */
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
		{
		  /* Place the loop header into list of region blocks.  */
		  degree[i] = -1;
		  rgn_bb_table[idx] = i;
		  RGN_NR_BLOCKS (nr_regions) = num_bbs;
		  RGN_BLOCKS (nr_regions) = idx++;
		  CONTAINING_RGN (i) = nr_regions;
		  BLOCK_TO_BB (i) = count = 0;

		  /* Remove blocks from queue[] when their in degree
		     becomes zero.  Repeat until no blocks are left on the
		     list.  This produces a topological list of blocks in
		     the region.  */
		  while (tail >= 0)
		    {
		      if (head < 0)
			head = tail;
		      child = queue[head];
		      if (degree[child] == 0)
			{
			  edge e;

			  degree[child] = -1;
			  rgn_bb_table[idx++] = child;
			  BLOCK_TO_BB (child) = ++count;
			  CONTAINING_RGN (child) = nr_regions;
			  queue[head] = queue[tail--];

			  for (e = BASIC_BLOCK (child)->succ;
			       e;
			       e = e->succ_next)
			    if (e->dest != EXIT_BLOCK_PTR)
			      --degree[e->dest->index];
			}
		      else
			--head;
		    }
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
		  RGN_PIPELINED_LOOP (nr_regions) = NULL_RTX;
		  RGN_MAY_BE_PIPELINED_P (nr_regions) = 0;
		  RGN_DO_NOT_SCHEDULE_P (nr_regions) = 0;
		  
		  if (!no_loops && flag_rcsp
		      && USE_AUTOMATON_PIPELINE_INTERFACE)
		    {
		      rtx start_loop_body_insn, tail_loop_body_insn;
		      rtx loop_body_bound_insn;
		      
		      get_block_head_tail
			(rgn_bb_table[RGN_BLOCKS (nr_regions)
				     + RGN_NR_BLOCKS (nr_regions) - 1],
			 &start_loop_body_insn, &tail_loop_body_insn);
		      
		      loop_body_bound_insn = NEXT_INSN (tail_loop_body_insn);
		      
		      get_block_head_tail
			(rgn_bb_table[RGN_BLOCKS (nr_regions)],
			 &start_loop_body_insn, &tail_loop_body_insn);
		      
		      RGN_MAY_BE_PIPELINED_P (nr_regions)
			= may_be_pipelined_p (start_loop_body_insn,
					      loop_body_bound_insn);
		    }
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
		  ++nr_regions;
		}
	    }
	}
      free (queue);
    }

  /* Any block that did not end up in a region is placed into a region
     by itself.  */
  for (i = 0; i < n_basic_blocks; i++)
    if (degree[i] >= 0)
      {
	rgn_bb_table[idx] = i;
	RGN_NR_BLOCKS (nr_regions) = 1;
	RGN_BLOCKS (nr_regions) = idx++;
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
	RGN_MAY_BE_PIPELINED_P (nr_regions) = 0;
	RGN_PIPELINED_LOOP (nr_regions) = NULL_RTX;
	RGN_DO_NOT_SCHEDULE_P (i) = 0;
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
	CONTAINING_RGN (i) = nr_regions++;
	BLOCK_TO_BB (i) = 0;
      }

  free (max_hdr);
  free (dfs_nr);
  free (stack);
  free (passed);
  free (header);
  free (inner);
  free (in_queue);
  free (in_stack);
}

/* Functions for regions scheduling information.  */

/* Compute dominators, probability, and potential-split-edges of bb.
   Assume that these values were already computed for bb's predecessors.  */

static void
compute_dom_prob_ps (bb)
     int bb;
{
  int nxt_in_edge, fst_in_edge, pred;
  int fst_out_edge, nxt_out_edge, nr_out_edges, nr_rgn_out_edges;

  prob[bb] = 0.0;
  if (IS_RGN_ENTRY (bb))
    {
      BITSET_ADD (dom[bb], 0, bbset_size);
      prob[bb] = 1.0;
      return;
    }

  fst_in_edge = nxt_in_edge = IN_EDGES (BB_TO_BLOCK (bb));

  /* Intialize dom[bb] to '111..1'.  */
  BITSET_INVERT (dom[bb], bbset_size);

  do
    {
      pred = FROM_BLOCK (nxt_in_edge);
      BITSET_INTER (dom[bb], dom[BLOCK_TO_BB (pred)], bbset_size);

      BITSET_UNION (ancestor_edges[bb], ancestor_edges[BLOCK_TO_BB (pred)],
		    edgeset_size);

      BITSET_ADD (ancestor_edges[bb], EDGE_TO_BIT (nxt_in_edge), edgeset_size);

      nr_out_edges = 1;
      nr_rgn_out_edges = 0;
      fst_out_edge = OUT_EDGES (pred);
      nxt_out_edge = NEXT_OUT (fst_out_edge);
      BITSET_UNION (pot_split[bb], pot_split[BLOCK_TO_BB (pred)],
		    edgeset_size);

      BITSET_ADD (pot_split[bb], EDGE_TO_BIT (fst_out_edge), edgeset_size);

      /* The successor doesn't belong in the region?  */
      if (CONTAINING_RGN (TO_BLOCK (fst_out_edge)) !=
	  CONTAINING_RGN (BB_TO_BLOCK (bb)))
	++nr_rgn_out_edges;

      while (fst_out_edge != nxt_out_edge)
	{
	  ++nr_out_edges;
	  /* The successor doesn't belong in the region?  */
	  if (CONTAINING_RGN (TO_BLOCK (nxt_out_edge)) !=
	      CONTAINING_RGN (BB_TO_BLOCK (bb)))
	    ++nr_rgn_out_edges;
	  BITSET_ADD (pot_split[bb], EDGE_TO_BIT (nxt_out_edge), edgeset_size);
	  nxt_out_edge = NEXT_OUT (nxt_out_edge);

	}

      /* Now nr_rgn_out_edges is the number of region-exit edges from
         pred, and nr_out_edges will be the number of pred out edges
         not leaving the region.  */
      nr_out_edges -= nr_rgn_out_edges;
      if (nr_rgn_out_edges > 0)
	prob[bb] += 0.9 * prob[BLOCK_TO_BB (pred)] / nr_out_edges;
      else
	prob[bb] += prob[BLOCK_TO_BB (pred)] / nr_out_edges;
      nxt_in_edge = NEXT_IN (nxt_in_edge);
    }
  while (fst_in_edge != nxt_in_edge);

  BITSET_ADD (dom[bb], bb, bbset_size);
  BITSET_DIFFER (pot_split[bb], ancestor_edges[bb], edgeset_size);

  if (sched_verbose >= 2)
    fprintf (sched_dump, ";;  bb_prob(%d, %d) = %3d\n", bb, BB_TO_BLOCK (bb),
	     (int) (100.0 * prob[bb]));
}

/* Functions for target info.  */

/* Compute in BL the list of split-edges of bb_src relatively to bb_trg.
   Note that bb_trg dominates bb_src.  */

static void
split_edges (bb_src, bb_trg, bl)
     int bb_src;
     int bb_trg;
     edgelst *bl;
{
  int es = edgeset_size;
  edgeset src = (edgeset) xcalloc (es, sizeof (HOST_WIDE_INT));

  while (es--)
    src[es] = (pot_split[bb_src])[es];
  BITSET_DIFFER (src, pot_split[bb_trg], edgeset_size);
  extract_bitlst (src, edgeset_size, edgeset_bitsize, bl);
  free (src);
}

/* Find the valid candidate-source-blocks for the target block TRG, compute
   their probability, and check if they are speculative or not.
   For speculative sources, compute their update-blocks and split-blocks.  */

static void
compute_trg_info (trg)
     int trg;
{
  register candidate *sp;
  edgelst el;
  int check_block, update_idx;
  int i, j, k, fst_edge, nxt_edge;

  /* Define some of the fields for the target bb as well.  */
  sp = candidate_table + trg;
  sp->is_valid = 1;
  sp->is_speculative = 0;
  sp->src_prob = 100;

  for (i = trg + 1; i < current_nr_blocks; i++)
    {
      sp = candidate_table + i;

      sp->is_valid = IS_DOMINATED (i, trg);
      if (sp->is_valid)
	{
	  sp->src_prob = GET_SRC_PROB (i, trg);
	  sp->is_valid = (sp->src_prob >= MIN_PROBABILITY);
	}

      if (sp->is_valid)
	{
	  split_edges (i, trg, &el);
	  sp->is_speculative = (el.nr_members) ? 1 : 0;
	  if (sp->is_speculative && !flag_schedule_speculative)
	    sp->is_valid = 0;
	}

      if (sp->is_valid)
	{
	  char *update_blocks;

	  /* Compute split blocks and store them in bblst_table.
	     The TO block of every split edge is a split block.  */
	  sp->split_bbs.first_member = &bblst_table[bblst_last];
	  sp->split_bbs.nr_members = el.nr_members;
	  for (j = 0; j < el.nr_members; bblst_last++, j++)
	    bblst_table[bblst_last] =
	      TO_BLOCK (rgn_edges[el.first_member[j]]);
	  sp->update_bbs.first_member = &bblst_table[bblst_last];

	  /* Compute update blocks and store them in bblst_table.
	     For every split edge, look at the FROM block, and check
	     all out edges.  For each out edge that is not a split edge,
	     add the TO block to the update block list.  This list can end
	     up with a lot of duplicates.  We need to weed them out to avoid
	     overrunning the end of the bblst_table.  */
	  update_blocks = (char *) alloca (n_basic_blocks);
	  memset (update_blocks, 0, n_basic_blocks);

	  update_idx = 0;
	  for (j = 0; j < el.nr_members; j++)
	    {
	      check_block = FROM_BLOCK (rgn_edges[el.first_member[j]]);
	      fst_edge = nxt_edge = OUT_EDGES (check_block);
	      do
		{
		  if (! update_blocks[TO_BLOCK (nxt_edge)])
		    {
		      for (k = 0; k < el.nr_members; k++)
			if (EDGE_TO_BIT (nxt_edge) == el.first_member[k])
			  break;

		      if (k >= el.nr_members)
			{
			  bblst_table[bblst_last++] = TO_BLOCK (nxt_edge);
			  update_blocks[TO_BLOCK (nxt_edge)] = 1;
			  update_idx++;
			}
		    }

		  nxt_edge = NEXT_OUT (nxt_edge);
		}
	      while (fst_edge != nxt_edge);
	    }
	  sp->update_bbs.nr_members = update_idx;

	  /* Make sure we didn't overrun the end of bblst_table.  */
	  if (bblst_last > bblst_size)
	    abort ();
	}
      else
	{
	  sp->split_bbs.nr_members = sp->update_bbs.nr_members = 0;

	  sp->is_speculative = 0;
	  sp->src_prob = 0;
	}
    }
}

/* Print candidates info, for debugging purposes.  Callable from debugger.  */

void
debug_candidate (i)
     int i;
{
  if (!candidate_table[i].is_valid)
    return;

  if (candidate_table[i].is_speculative)
    {
      int j;
      fprintf (sched_dump, "src b %d bb %d speculative \n", BB_TO_BLOCK (i), i);

      fprintf (sched_dump, "split path: ");
      for (j = 0; j < candidate_table[i].split_bbs.nr_members; j++)
	{
	  int b = candidate_table[i].split_bbs.first_member[j];

	  fprintf (sched_dump, " %d ", b);
	}
      fprintf (sched_dump, "\n");

      fprintf (sched_dump, "update path: ");
      for (j = 0; j < candidate_table[i].update_bbs.nr_members; j++)
	{
	  int b = candidate_table[i].update_bbs.first_member[j];

	  fprintf (sched_dump, " %d ", b);
	}
      fprintf (sched_dump, "\n");
    }
  else
    {
      fprintf (sched_dump, " src %d equivalent\n", BB_TO_BLOCK (i));
    }
}

/* Print candidates info, for debugging purposes.  Callable from debugger.  */

void
debug_candidates (trg)
     int trg;
{
  int i;

  fprintf (sched_dump, "----------- candidate table: target: b=%d bb=%d ---\n",
	   BB_TO_BLOCK (trg), trg);
  for (i = trg + 1; i < current_nr_blocks; i++)
    debug_candidate (i);
}

/* Functions for speculative scheduing.  */

/* Return 0 if x is a set of a register alive in the beginning of one
   of the split-blocks of src, otherwise return 1.  */

static int
check_live_1 (src, x)
     int src;
     rtx x;
{
  register int i;
  register int regno;
  register rtx reg = SET_DEST (x);

  if (reg == 0)
    return 1;

  while (GET_CODE (reg) == SUBREG || GET_CODE (reg) == ZERO_EXTRACT
	 || GET_CODE (reg) == SIGN_EXTRACT
	 || GET_CODE (reg) == STRICT_LOW_PART)
    reg = XEXP (reg, 0);

  if (GET_CODE (reg) == PARALLEL
      && GET_MODE (reg) == BLKmode)
    {
      register int i;
      for (i = XVECLEN (reg, 0) - 1; i >= 0; i--)
	if (check_live_1 (src, XVECEXP (reg, 0, i)))
	  return 1;
      return 0;
    }

  if (GET_CODE (reg) != REG)
    return 1;

  regno = REGNO (reg);

  if (regno < FIRST_PSEUDO_REGISTER && global_regs[regno])
    {
      /* Global registers are assumed live.  */
      return 0;
    }
  else
    {
      if (regno < FIRST_PSEUDO_REGISTER)
	{
	  /* Check for hard registers.  */
	  int j = HARD_REGNO_NREGS (regno, GET_MODE (reg));
	  while (--j >= 0)
	    {
	      for (i = 0; i < candidate_table[src].split_bbs.nr_members; i++)
		{
		  int b = candidate_table[src].split_bbs.first_member[i];

		  if (REGNO_REG_SET_P (BASIC_BLOCK (b)->global_live_at_start,
				       regno + j))
		    {
		      return 0;
		    }
		}
	    }
	}
      else
	{
	  /* Check for psuedo registers.  */
	  for (i = 0; i < candidate_table[src].split_bbs.nr_members; i++)
	    {
	      int b = candidate_table[src].split_bbs.first_member[i];

	      if (REGNO_REG_SET_P (BASIC_BLOCK (b)->global_live_at_start, regno))
		{
		  return 0;
		}
	    }
	}
    }

  return 1;
}

/* If x is a set of a register R, mark that R is alive in the beginning
   of every update-block of src.  */

static void
update_live_1 (src, x)
     int src;
     rtx x;
{
  register int i;
  register int regno;
  register rtx reg = SET_DEST (x);

  if (reg == 0)
    return;

  while (GET_CODE (reg) == SUBREG || GET_CODE (reg) == ZERO_EXTRACT
	 || GET_CODE (reg) == SIGN_EXTRACT
	 || GET_CODE (reg) == STRICT_LOW_PART)
    reg = XEXP (reg, 0);

  if (GET_CODE (reg) == PARALLEL
      && GET_MODE (reg) == BLKmode)
    {
      register int i;
      for (i = XVECLEN (reg, 0) - 1; i >= 0; i--)
	update_live_1 (src, XVECEXP (reg, 0, i));
      return;
    }

  if (GET_CODE (reg) != REG)
    return;

  /* Global registers are always live, so the code below does not apply
     to them.  */

  regno = REGNO (reg);

  if (regno >= FIRST_PSEUDO_REGISTER || !global_regs[regno])
    {
      if (regno < FIRST_PSEUDO_REGISTER)
	{
	  int j = HARD_REGNO_NREGS (regno, GET_MODE (reg));
	  while (--j >= 0)
	    {
	      for (i = 0; i < candidate_table[src].update_bbs.nr_members; i++)
		{
		  int b = candidate_table[src].update_bbs.first_member[i];

		  SET_REGNO_REG_SET (BASIC_BLOCK (b)->global_live_at_start,
				     regno + j);
		}
	    }
	}
      else
	{
	  for (i = 0; i < candidate_table[src].update_bbs.nr_members; i++)
	    {
	      int b = candidate_table[src].update_bbs.first_member[i];

	      SET_REGNO_REG_SET (BASIC_BLOCK (b)->global_live_at_start, regno);
	    }
	}
    }
}

/* Return 1 if insn can be speculatively moved from block src to trg,
   otherwise return 0.  Called before first insertion of insn to
   ready-list or before the scheduling.  */

static int
check_live (insn, src)
     rtx insn;
     int src;
{
  /* Find the registers set by instruction.  */
  if (GET_CODE (PATTERN (insn)) == SET
      || GET_CODE (PATTERN (insn)) == CLOBBER)
    return check_live_1 (src, PATTERN (insn));
  else if (GET_CODE (PATTERN (insn)) == PARALLEL)
    {
      int j;
      for (j = XVECLEN (PATTERN (insn), 0) - 1; j >= 0; j--)
	if ((GET_CODE (XVECEXP (PATTERN (insn), 0, j)) == SET
	     || GET_CODE (XVECEXP (PATTERN (insn), 0, j)) == CLOBBER)
	    && !check_live_1 (src, XVECEXP (PATTERN (insn), 0, j)))
	  return 0;

      return 1;
    }

  return 1;
}

/* Update the live registers info after insn was moved speculatively from
   block src to trg.  */

static void
update_live (insn, src)
     rtx insn;
     int src;
{
  /* Find the registers set by instruction.  */
  if (GET_CODE (PATTERN (insn)) == SET
      || GET_CODE (PATTERN (insn)) == CLOBBER)
    update_live_1 (src, PATTERN (insn));
  else if (GET_CODE (PATTERN (insn)) == PARALLEL)
    {
      int j;
      for (j = XVECLEN (PATTERN (insn), 0) - 1; j >= 0; j--)
	if (GET_CODE (XVECEXP (PATTERN (insn), 0, j)) == SET
	    || GET_CODE (XVECEXP (PATTERN (insn), 0, j)) == CLOBBER)
	  update_live_1 (src, XVECEXP (PATTERN (insn), 0, j));
    }
}

/* Exception Free Loads:

   We define five classes of speculative loads: IFREE, IRISKY,
   PFREE, PRISKY, and MFREE.

   IFREE loads are loads that are proved to be exception-free, just
   by examining the load insn.  Examples for such loads are loads
   from TOC and loads of global data.

   IRISKY loads are loads that are proved to be exception-risky,
   just by examining the load insn.  Examples for such loads are
   volatile loads and loads from shared memory.

   PFREE loads are loads for which we can prove, by examining other
   insns, that they are exception-free.  Currently, this class consists
   of loads for which we are able to find a "similar load", either in
   the target block, or, if only one split-block exists, in that split
   block.  Load2 is similar to load1 if both have same single base
   register.  We identify only part of the similar loads, by finding
   an insn upon which both load1 and load2 have a DEF-USE dependence.

   PRISKY loads are loads for which we can prove, by examining other
   insns, that they are exception-risky.  Currently we have two proofs for
   such loads.  The first proof detects loads that are probably guarded by a
   test on the memory address.  This proof is based on the
   backward and forward data dependence information for the region.
   Let load-insn be the examined load.
   Load-insn is PRISKY iff ALL the following hold:

   - insn1 is not in the same block as load-insn
   - there is a DEF-USE dependence chain (insn1, ..., load-insn)
   - test-insn is either a compare or a branch, not in the same block
     as load-insn
   - load-insn is reachable from test-insn
   - there is a DEF-USE dependence chain (insn1, ..., test-insn)

   This proof might fail when the compare and the load are fed
   by an insn not in the region.  To solve this, we will add to this
   group all loads that have no input DEF-USE dependence.

   The second proof detects loads that are directly or indirectly
   fed by a speculative load.  This proof is affected by the
   scheduling process.  We will use the flag  fed_by_spec_load.
   Initially, all insns have this flag reset.  After a speculative
   motion of an insn, if insn is either a load, or marked as
   fed_by_spec_load, we will also mark as fed_by_spec_load every
   insn1 for which a DEF-USE dependence (insn, insn1) exists.  A
   load which is fed_by_spec_load is also PRISKY.

   MFREE (maybe-free) loads are all the remaining loads. They may be
   exception-free, but we cannot prove it.

   Now, all loads in IFREE and PFREE classes are considered
   exception-free, while all loads in IRISKY and PRISKY classes are
   considered exception-risky.  As for loads in the MFREE class,
   these are considered either exception-free or exception-risky,
   depending on whether we are pessimistic or optimistic.  We have
   to take the pessimistic approach to assure the safety of
   speculative scheduling, but we can take the optimistic approach
   by invoking the -fsched_spec_load_dangerous option.  */

enum INSN_TRAP_CLASS
{
  TRAP_FREE = 0, IFREE = 1, PFREE_CANDIDATE = 2,
  PRISKY_CANDIDATE = 3, IRISKY = 4, TRAP_RISKY = 5
};

#define WORST_CLASS(class1, class2) \
((class1 > class2) ? class1 : class2)

/* Non-zero if block bb_to is equal to, or reachable from block bb_from.  */
#define IS_REACHABLE(bb_from, bb_to)					\
(bb_from == bb_to                                                       \
   || IS_RGN_ENTRY (bb_from)						\
   || (bitset_member (ancestor_edges[bb_to],				\
		      EDGE_TO_BIT (IN_EDGES (BB_TO_BLOCK (bb_from))),	\
		      edgeset_size)))

/* Non-zero iff the address is comprised from at most 1 register.  */
#define CONST_BASED_ADDRESS_P(x)			\
  (GET_CODE (x) == REG					\
   || ((GET_CODE (x) == PLUS || GET_CODE (x) == MINUS   \
	|| (GET_CODE (x) == LO_SUM))	                \
       && (GET_CODE (XEXP (x, 0)) == CONST_INT		\
	   || GET_CODE (XEXP (x, 1)) == CONST_INT)))

/* Turns on the fed_by_spec_load flag for insns fed by load_insn.  */

static void
set_spec_fed (load_insn)
     rtx load_insn;
{
  rtx link;

  for (link = INSN_DEPEND (load_insn); link; link = XEXP (link, 1))
    if (GET_MODE (link) == VOIDmode)
      FED_BY_SPEC_LOAD (XEXP (link, 0)) = 1;
}				/* set_spec_fed */

/* On the path from the insn to load_insn_bb, find a conditional
branch depending on insn, that guards the speculative load.  */

static int
find_conditional_protection (insn, load_insn_bb)
     rtx insn;
     int load_insn_bb;
{
  rtx link;

  /* Iterate through DEF-USE forward dependences.  */
  for (link = INSN_DEPEND (insn); link; link = XEXP (link, 1))
    {
      rtx next = XEXP (link, 0);
      if ((CONTAINING_RGN (BLOCK_NUM (next)) ==
	   CONTAINING_RGN (BB_TO_BLOCK (load_insn_bb)))
	  && IS_REACHABLE (INSN_BB (next), load_insn_bb)
	  && load_insn_bb != INSN_BB (next)
	  && GET_MODE (link) == VOIDmode
	  && (GET_CODE (next) == JUMP_INSN
	      || find_conditional_protection (next, load_insn_bb)))
	return 1;
    }
  return 0;
}				/* find_conditional_protection */

/* Returns 1 if the same insn1 that participates in the computation
   of load_insn's address is feeding a conditional branch that is
   guarding on load_insn. This is true if we find a the two DEF-USE
   chains:
   insn1 -> ... -> conditional-branch
   insn1 -> ... -> load_insn,
   and if a flow path exist:
   insn1 -> ... -> conditional-branch -> ... -> load_insn,
   and if insn1 is on the path
   region-entry -> ... -> bb_trg -> ... load_insn.

   Locate insn1 by climbing on LOG_LINKS from load_insn.
   Locate the branch by following INSN_DEPEND from insn1.  */

static int
is_conditionally_protected (load_insn, bb_src, bb_trg)
     rtx load_insn;
     int bb_src, bb_trg;
{
  rtx link;

  for (link = LOG_LINKS (load_insn); link; link = XEXP (link, 1))
    {
      rtx insn1 = XEXP (link, 0);

      /* Must be a DEF-USE dependence upon non-branch.  */
      if (GET_MODE (link) != VOIDmode
	  || GET_CODE (insn1) == JUMP_INSN)
	continue;

      /* Must exist a path: region-entry -> ... -> bb_trg -> ... load_insn.  */
      if (INSN_BB (insn1) == bb_src
	  || (CONTAINING_RGN (BLOCK_NUM (insn1))
	      != CONTAINING_RGN (BB_TO_BLOCK (bb_src)))
	  || (!IS_REACHABLE (bb_trg, INSN_BB (insn1))
	      && !IS_REACHABLE (INSN_BB (insn1), bb_trg)))
	continue;

      /* Now search for the conditional-branch.  */
      if (find_conditional_protection (insn1, bb_src))
	return 1;

      /* Recursive step: search another insn1, "above" current insn1.  */
      return is_conditionally_protected (insn1, bb_src, bb_trg);
    }

  /* The chain does not exist.  */
  return 0;
}				/* is_conditionally_protected */

/* Returns 1 if a clue for "similar load" 'insn2' is found, and hence
   load_insn can move speculatively from bb_src to bb_trg.  All the
   following must hold:

   (1) both loads have 1 base register (PFREE_CANDIDATEs).
   (2) load_insn and load1 have a def-use dependence upon
   the same insn 'insn1'.
   (3) either load2 is in bb_trg, or:
   - there's only one split-block, and
   - load1 is on the escape path, and

   From all these we can conclude that the two loads access memory
   addresses that differ at most by a constant, and hence if moving
   load_insn would cause an exception, it would have been caused by
   load2 anyhow.  */

static int
is_pfree (load_insn, bb_src, bb_trg)
     rtx load_insn;
     int bb_src, bb_trg;
{
  rtx back_link;
  register candidate *candp = candidate_table + bb_src;

  if (candp->split_bbs.nr_members != 1)
    /* Must have exactly one escape block.  */
    return 0;

  for (back_link = LOG_LINKS (load_insn);
       back_link; back_link = XEXP (back_link, 1))
    {
      rtx insn1 = XEXP (back_link, 0);

      if (GET_MODE (back_link) == VOIDmode)
	{
	  /* Found a DEF-USE dependence (insn1, load_insn).  */
	  rtx fore_link;

	  for (fore_link = INSN_DEPEND (insn1);
	       fore_link; fore_link = XEXP (fore_link, 1))
	    {
	      rtx insn2 = XEXP (fore_link, 0);
	      if (GET_MODE (fore_link) == VOIDmode)
		{
		  /* Found a DEF-USE dependence (insn1, insn2).  */
		  if (haifa_classify_insn (insn2) != PFREE_CANDIDATE)
		    /* insn2 not guaranteed to be a 1 base reg load.  */
		    continue;

		  if (INSN_BB (insn2) == bb_trg)
		    /* insn2 is the similar load, in the target block.  */
		    return 1;

		  if (*(candp->split_bbs.first_member) == BLOCK_NUM (insn2))
		    /* insn2 is a similar load, in a split-block.  */
		    return 1;
		}
	    }
	}
    }

  /* Couldn't find a similar load.  */
  return 0;
}				/* is_pfree */

/* Returns a class that insn with GET_DEST(insn)=x may belong to,
   as found by analyzing insn's expression.  */

static int
may_trap_exp (x, is_store)
     rtx x;
     int is_store;
{
  enum rtx_code code;

  if (x == 0)
    return TRAP_FREE;
  code = GET_CODE (x);
  if (is_store)
    {
      if (code == MEM)
	return TRAP_RISKY;
      else
	return TRAP_FREE;
    }
  if (code == MEM)
    {
      /* The insn uses memory:  a volatile load.  */
      if (MEM_VOLATILE_P (x))
	return IRISKY;
      /* An exception-free load.  */
      if (!may_trap_p (x))
	return IFREE;
      /* A load with 1 base register, to be further checked.  */
      if (CONST_BASED_ADDRESS_P (XEXP (x, 0)))
	return PFREE_CANDIDATE;
      /* No info on the load, to be further checked.  */
      return PRISKY_CANDIDATE;
    }
  else
    {
      const char *fmt;
      int i, insn_class = TRAP_FREE;

      /* Neither store nor load, check if it may cause a trap.  */
      if (may_trap_p (x))
	return TRAP_RISKY;
      /* Recursive step: walk the insn...  */
      fmt = GET_RTX_FORMAT (code);
      for (i = GET_RTX_LENGTH (code) - 1; i >= 0; i--)
	{
	  if (fmt[i] == 'e')
	    {
	      int tmp_class = may_trap_exp (XEXP (x, i), is_store);
	      insn_class = WORST_CLASS (insn_class, tmp_class);
	    }
	  else if (fmt[i] == 'E')
	    {
	      int j;
	      for (j = 0; j < XVECLEN (x, i); j++)
		{
		  int tmp_class = may_trap_exp (XVECEXP (x, i, j), is_store);
		  insn_class = WORST_CLASS (insn_class, tmp_class);
		  if (insn_class == TRAP_RISKY || insn_class == IRISKY)
		    break;
		}
	    }
	  if (insn_class == TRAP_RISKY || insn_class == IRISKY)
	    break;
	}
      return insn_class;
    }
}

/* Classifies insn for the purpose of verifying that it can be
   moved speculatively, by examining it's patterns, returning:
   TRAP_RISKY: store, or risky non-load insn (e.g. division by variable).
   TRAP_FREE: non-load insn.
   IFREE: load from a globaly safe location.
   IRISKY: volatile load.
   PFREE_CANDIDATE, PRISKY_CANDIDATE: load that need to be checked for
   being either PFREE or PRISKY.  */

static int
haifa_classify_insn (insn)
     rtx insn;
{
  rtx pat = PATTERN (insn);
  int tmp_class = TRAP_FREE;
  int insn_class = TRAP_FREE;
  enum rtx_code code;

  if (GET_CODE (pat) == PARALLEL)
    {
      int i, len = XVECLEN (pat, 0);

      for (i = len - 1; i >= 0; i--)
	{
	  code = GET_CODE (XVECEXP (pat, 0, i));
	  switch (code)
	    {
	    case CLOBBER:
	      /* Test if it is a 'store'.  */
	      tmp_class = may_trap_exp (XEXP (XVECEXP (pat, 0, i), 0), 1);
	      break;
	    case SET:
	      /* Test if it is a store.  */
	      tmp_class = may_trap_exp (SET_DEST (XVECEXP (pat, 0, i)), 1);
	      if (tmp_class == TRAP_RISKY)
		break;
	      /* Test if it is a load.  */
	      tmp_class =
		WORST_CLASS (tmp_class,
			     may_trap_exp (SET_SRC (XVECEXP (pat, 0, i)), 0));
	      break;
	    case COND_EXEC:
	    case TRAP_IF:
	      tmp_class = TRAP_RISKY;
	      break;
	    default:;
	    }
	  insn_class = WORST_CLASS (insn_class, tmp_class);
	  if (insn_class == TRAP_RISKY || insn_class == IRISKY)
	    break;
	}
    }
  else
    {
      code = GET_CODE (pat);
      switch (code)
	{
	case CLOBBER:
	  /* Test if it is a 'store'.  */
	  tmp_class = may_trap_exp (XEXP (pat, 0), 1);
	  break;
	case SET:
	  /* Test if it is a store.  */
	  tmp_class = may_trap_exp (SET_DEST (pat), 1);
	  if (tmp_class == TRAP_RISKY)
	    break;
	  /* Test if it is a load.  */
	  tmp_class =
	    WORST_CLASS (tmp_class,
			 may_trap_exp (SET_SRC (pat), 0));
	  break;
	case COND_EXEC:
	case TRAP_IF:
	  tmp_class = TRAP_RISKY;
	  break;
	default:;
	}
      insn_class = tmp_class;
    }

  return insn_class;
}

/* Return 1 if load_insn is prisky (i.e. if load_insn is fed by
   a load moved speculatively, or if load_insn is protected by
   a compare on load_insn's address).  */

static int
is_prisky (load_insn, bb_src, bb_trg)
     rtx load_insn;
     int bb_src, bb_trg;
{
  if (FED_BY_SPEC_LOAD (load_insn))
    return 1;

  if (LOG_LINKS (load_insn) == NULL)
    /* Dependence may 'hide' out of the region.  */
    return 1;

  if (is_conditionally_protected (load_insn, bb_src, bb_trg))
    return 1;

  return 0;
}

/* Insn is a candidate to be moved speculatively from bb_src to bb_trg.
   Return 1 if insn is exception-free (and the motion is valid)
   and 0 otherwise.  */

static int
is_exception_free (insn, bb_src, bb_trg)
     rtx insn;
     int bb_src, bb_trg;
{
  int insn_class = haifa_classify_insn (insn);

  /* Handle non-load insns.  */
  switch (insn_class)
    {
    case TRAP_FREE:
      return 1;
    case TRAP_RISKY:
      return 0;
    default:;
    }

  /* Handle loads.  */
  if (!flag_schedule_speculative_load)
    return 0;
  IS_LOAD_INSN (insn) = 1;
  switch (insn_class)
    {
    case IFREE:
      return (1);
    case IRISKY:
      return 0;
    case PFREE_CANDIDATE:
      if (is_pfree (insn, bb_src, bb_trg))
	return 1;
      /* Don't 'break' here: PFREE-candidate is also PRISKY-candidate.  */
    case PRISKY_CANDIDATE:
      if (!flag_schedule_speculative_load_dangerous
	  || is_prisky (insn, bb_src, bb_trg))
	return 0;
      break;
    default:;
    }

  return flag_schedule_speculative_load_dangerous;
}

/* The number of insns from the current block scheduled so far.  */
static int sched_target_n_insns;
/* The number of insns from the current block to be scheduled in total.  */
static int target_n_insns;
/* The number of insns from the entire region scheduled so far.  */
static int sched_n_insns;

/* Implementations of the sched_info functions for region scheduling.  */
static void init_ready_list PARAMS ((struct ready_list *));
static int can_schedule_ready_p PARAMS ((rtx));
static int new_ready PARAMS ((rtx));
static int schedule_more_p PARAMS ((void));
static const char *rgn_print_insn PARAMS ((rtx, int));
static int rgn_rank PARAMS ((rtx, rtx));

/* Return nonzero if there are more insns that should be scheduled.  */

static int
schedule_more_p ()
{
  return sched_target_n_insns < target_n_insns;
}

/* Add all insns that are initially ready to the ready list READY.  Called
   once before scheduling a set of insns.  */

static void
init_ready_list (ready)
     struct ready_list *ready;
{
  rtx prev_head = current_sched_info->prev_head;
  rtx next_tail = current_sched_info->next_tail;
  int bb_src;
  rtx insn;

  target_n_insns = 0;
  sched_target_n_insns = 0;
  sched_n_insns = 0;

  /* Print debugging information.  */
  if (sched_verbose >= 5)
    debug_dependencies ();

  /* Prepare current target block info.  */
  if (current_nr_blocks > 1)
    {
      candidate_table = (candidate *) xmalloc (current_nr_blocks
					       * sizeof (candidate));

      bblst_last = 0;
      /* bblst_table holds split blocks and update blocks for each block after
	 the current one in the region.  split blocks and update blocks are
	 the TO blocks of region edges, so there can be at most rgn_nr_edges
	 of them.  */
      bblst_size = (current_nr_blocks - target_bb) * rgn_nr_edges;
      bblst_table = (int *) xmalloc (bblst_size * sizeof (int));

      bitlst_table_last = 0;
      bitlst_table_size = rgn_nr_edges;
      bitlst_table = (int *) xmalloc (rgn_nr_edges * sizeof (int));

      compute_trg_info (target_bb);
    }

  /* Initialize ready list with all 'ready' insns in target block.
     Count number of insns in the target block being scheduled.  */
  for (insn = NEXT_INSN (prev_head); insn != next_tail; insn = NEXT_INSN (insn))
    {
      rtx next;

      if (! INSN_P (insn))
	continue;
      next = NEXT_INSN (insn);

      if (INSN_DEP_COUNT (insn) == 0
	  && (SCHED_GROUP_P (next) == 0 || ! INSN_P (next)))
	ready_add (ready, insn);
      if (!(SCHED_GROUP_P (insn)))
	target_n_insns++;
    }

  /* Add to ready list all 'ready' insns in valid source blocks.
     For speculative insns, check-live, exception-free, and
     issue-delay.  */
  for (bb_src = target_bb + 1; bb_src < current_nr_blocks; bb_src++)
    if (IS_VALID (bb_src))
      {
	rtx src_head;
	rtx src_next_tail;
	rtx tail, head;

	get_block_head_tail (BB_TO_BLOCK (bb_src), &head, &tail);
	src_next_tail = NEXT_INSN (tail);
	src_head = head;

	for (insn = src_head; insn != src_next_tail; insn = NEXT_INSN (insn))
	  {
/* CYGNUS LOCAL -- vmakarov/dfa */
	    int move_p = 0;
/* END CYGNUS LOCAL -- vmakarov/dfa */
	    if (! INSN_P (insn))
	      continue;

/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
	    if (USE_AUTOMATON_PIPELINE_INTERFACE)
	      recog_memoized (insn);
	    
	    if (USE_AUTOMATON_PIPELINE_INTERFACE && !CANT_MOVE (insn)
		&& (!IS_SPECULATIVE_INSN (insn)
		    || (INSN_CODE (insn) >= 0
			&& min_insn_conflict_delay (curr_state, insn, insn) <= 3
			&& check_live (insn, bb_src)
			&& is_exception_free (insn, bb_src, target_bb))))
	      move_p = 0;
#endif

#if OLD_PIPELINE_INTERFACE
	    if (!USE_AUTOMATON_PIPELINE_INTERFACE)
	      {
/* END CYGNUS LOCAL -- vmakarov/dfa */
	    if (!CANT_MOVE (insn)
		&& (!IS_SPECULATIVE_INSN (insn)
		    || (insn_issue_delay (insn) <= 3
			&& check_live (insn, bb_src)
			&& is_exception_free (insn, bb_src, target_bb))))
/* CYGNUS LOCAL -- vmakarov/dfa */
	        move_p = 0;
	      }
#endif
	    if (move_p)
/* END CYGNUS LOCAL -- vmakarov/dfa */
	      {
		rtx next;

		/* Note that we havn't squirrled away the notes for
		   blocks other than the current.  So if this is a
		   speculative insn, NEXT might otherwise be a note.  */
		next = next_nonnote_insn (insn);
		if (INSN_DEP_COUNT (insn) == 0
		    && (! next
			|| SCHED_GROUP_P (next) == 0
			|| ! INSN_P (next)))
		  ready_add (ready, insn);
	      }
	  }
      }
}

/* Called after taking INSN from the ready list.  Returns nonzero if this
   insn can be scheduled, nonzero if we should silently discard it.  */

static int
can_schedule_ready_p (insn)
     rtx insn;
{
  /* An interblock motion?  */
  if (INSN_BB (insn) != target_bb)
    {
      rtx temp;
      basic_block b1;

      if (IS_SPECULATIVE_INSN (insn))
	{
	  if (!check_live (insn, INSN_BB (insn)))
	    return 0;
	  update_live (insn, INSN_BB (insn));

	  /* For speculative load, mark insns fed by it.  */
	  if (IS_LOAD_INSN (insn) || FED_BY_SPEC_LOAD (insn))
	    set_spec_fed (insn);

	  nr_spec++;
	}
      nr_inter++;

      /* Find the beginning of the scheduling group.  */
      /* ??? Ought to update basic block here, but later bits of
	 schedule_block assumes the original insn block is
	 still intact.  */

      temp = insn;
      while (SCHED_GROUP_P (temp))
	temp = PREV_INSN (temp);

      /* Update source block boundaries.   */
      b1 = BLOCK_FOR_INSN (temp);
      if (temp == b1->head && insn == b1->end)
	{
	  /* We moved all the insns in the basic block.
	     Emit a note after the last insn and update the
	     begin/end boundaries to point to the note.  */
	  rtx note = emit_note_after (NOTE_INSN_DELETED, insn);
	  b1->head = note;
	  b1->end = note;
	}
      else if (insn == b1->end)
	{
	  /* We took insns from the end of the basic block,
	     so update the end of block boundary so that it
	     points to the first insn we did not move.  */
	  b1->end = PREV_INSN (temp);
	}
      else if (temp == b1->head)
	{
	  /* We took insns from the start of the basic block,
	     so update the start of block boundary so that
	     it points to the first insn we did not move.  */
	  b1->head = NEXT_INSN (insn);
	}
    }
  else
    {
      /* In block motion.  */
      sched_target_n_insns++;
    }
  sched_n_insns++;

  return 1;
}

/* Called after INSN has all its dependencies resolved.  Return nonzero
   if it should be moved to the ready list or the queue, or zero if we
   should silently discard it.  */
static int
new_ready (next)
     rtx next;
{
  /* For speculative insns, before inserting to ready/queue,
     check live, exception-free, and issue-delay.  */
/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
  if (USE_AUTOMATON_PIPELINE_INTERFACE)
    {
      recog_memoized (next);
      
      if (INSN_BB (next) != target_bb
	  && (!IS_VALID (INSN_BB (next))
	      || CANT_MOVE (next)
	      || (IS_SPECULATIVE_INSN (next)
		  && (INSN_CODE (next) < 0
		      || min_insn_conflict_delay (curr_state, next, next) > 3
		      || !check_live (next, INSN_BB (next))
		      || !is_exception_free (next, INSN_BB (next),
					     target_bb)))))
	return 0;
    }
#endif /* #if AUTOMATON_PIPELINE_INTERFACE */

#if OLD_PIPELINE_INTERFACE
  if (!USE_AUTOMATON_PIPELINE_INTERFACE)
    {
/* END CYGNUS LOCAL -- vmakarov/dfa */
  if (INSN_BB (next) != target_bb
      && (!IS_VALID (INSN_BB (next))
	  || CANT_MOVE (next)
	  || (IS_SPECULATIVE_INSN (next)
	      && (insn_issue_delay (next) > 3
		  || !check_live (next, INSN_BB (next))
		  || !is_exception_free (next, INSN_BB (next), target_bb)))))
    return 0;
/* CYGNUS LOCAL -- vmakarov/dfa */
	    }
#endif /* #if OLD_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/dfa */
  return 1;
}

/* Return a string that contains the insn uid and optionally anything else
   necessary to identify this insn in an output.  It's valid to use a
   static buffer for this.  The ALIGNED parameter should cause the string
   to be formatted so that multiple output lines will line up nicely.  */

static const char *
rgn_print_insn (insn, aligned)
     rtx insn;
     int aligned;
{
  static char tmp[80];

  if (aligned)
    sprintf (tmp, "b%3d: i%4d", INSN_BB (insn), INSN_UID (insn));
  else
    {
      if (current_nr_blocks > 1 && INSN_BB (insn) != target_bb)
	sprintf (tmp, "%d/b%d", INSN_UID (insn), INSN_BB (insn));
      else
	sprintf (tmp, "%d", INSN_UID (insn));
    }
  return tmp;
}

/* Compare priority of two insns.  Return a positive number if the second
   insn is to be preferred for scheduling, and a negative one if the first
   is to be preferred.  Zero if they are equally good.  */

static int
rgn_rank (insn1, insn2)
     rtx insn1, insn2;
{
  /* Some comparison make sense in interblock scheduling only.  */
  if (INSN_BB (insn1) != INSN_BB (insn2))
    {
      int spec_val, prob_val;

      /* Prefer an inblock motion on an interblock motion.  */
      if ((INSN_BB (insn2) == target_bb) && (INSN_BB (insn1) != target_bb))
	return 1;
      if ((INSN_BB (insn1) == target_bb) && (INSN_BB (insn2) != target_bb))
	return -1;

      /* Prefer a useful motion on a speculative one.  */
      spec_val = IS_SPECULATIVE_INSN (insn1) - IS_SPECULATIVE_INSN (insn2);
      if (spec_val)
	return spec_val;

      /* Prefer a more probable (speculative) insn.  */
      prob_val = INSN_PROBABILITY (insn2) - INSN_PROBABILITY (insn1);
      if (prob_val)
	return prob_val;
    }
  return 0;
}

/* Used in schedule_insns to initialize current_sched_info for scheduling
   regions (or single basic blocks).  */

static struct sched_info region_sched_info =
{
  init_ready_list,
  can_schedule_ready_p,
  schedule_more_p,
  new_ready,
  rgn_rank,
  rgn_print_insn,

  NULL, NULL,
  NULL, NULL,
  0
};

/* Add dependences so that branches are scheduled to run last in their
   block.  */

static void
add_branch_dependences (head, tail)
     rtx head, tail;
{
  rtx insn, last;

  /* For all branches, calls, uses, clobbers, and cc0 setters, force them
     to remain in order at the end of the block by adding dependencies and
     giving the last a high priority.  There may be notes present, and
     prev_head may also be a note.

     Branches must obviously remain at the end.  Calls should remain at the
     end since moving them results in worse register allocation.  Uses remain
     at the end to ensure proper register allocation.  cc0 setters remaim
     at the end because they can't be moved away from their cc0 user.  */
  insn = tail;
  last = 0;
  while (GET_CODE (insn) == CALL_INSN
	 || GET_CODE (insn) == JUMP_INSN
	 || (GET_CODE (insn) == INSN
	     && (GET_CODE (PATTERN (insn)) == USE
		 || GET_CODE (PATTERN (insn)) == CLOBBER
#ifdef HAVE_cc0
		 || sets_cc0_p (PATTERN (insn))
#endif
	     ))
	 || GET_CODE (insn) == NOTE)
    {
      if (GET_CODE (insn) != NOTE)
	{
	  if (last != 0
	      && !find_insn_list (insn, LOG_LINKS (last)))
	    {
	      add_dependence (last, insn, REG_DEP_ANTI);
	      INSN_REF_COUNT (insn)++;
	    }

	  CANT_MOVE (insn) = 1;

	  last = insn;
	  /* Skip over insns that are part of a group.
	     Make each insn explicitly depend on the previous insn.
	     This ensures that only the group header will ever enter
	     the ready queue (and, when scheduled, will automatically
	     schedule the SCHED_GROUP_P block).  */
	  while (SCHED_GROUP_P (insn))
	    {
	      rtx temp = prev_nonnote_insn (insn);
	      add_dependence (insn, temp, REG_DEP_ANTI);
	      insn = temp;
	    }
	}

      /* Don't overrun the bounds of the basic block.  */
      if (insn == head)
	break;

      insn = PREV_INSN (insn);
    }

  /* Make sure these insns are scheduled last in their block.  */
  insn = last;
  if (insn != 0)
    while (insn != head)
      {
	insn = prev_nonnote_insn (insn);

	if (INSN_REF_COUNT (insn) != 0)
	  continue;

	add_dependence (last, insn, REG_DEP_ANTI);
	INSN_REF_COUNT (insn) = 1;

	/* Skip over insns that are part of a group.  */
	while (SCHED_GROUP_P (insn))
	  insn = prev_nonnote_insn (insn);
      }
}

/* Data structures for the computation of data dependences in a regions.  We
   keep one `deps' structure for every basic block.  Before analyzing the
   data dependences for a bb, its variables are initialized as a function of
   the variables of its predecessors.  When the analysis for a bb completes,
   we save the contents to the corresponding bb_deps[bb] variable.  */

static struct deps *bb_deps;

/* After computing the dependencies for block BB, propagate the dependencies
   found in TMP_DEPS to the successors of the block.  MAX_REG is the number
   of registers.  */
static void
propagate_deps (bb, tmp_deps, max_reg)
     int bb;
     struct deps *tmp_deps;
     int max_reg;
{
  int b = BB_TO_BLOCK (bb);
  int e, first_edge;
  int reg;
  rtx link_insn, link_mem;
  rtx u;

  /* These lists should point to the right place, for correct
     freeing later.  */
  bb_deps[bb].pending_read_insns = tmp_deps->pending_read_insns;
  bb_deps[bb].pending_read_mems = tmp_deps->pending_read_mems;
  bb_deps[bb].pending_write_insns = tmp_deps->pending_write_insns;
  bb_deps[bb].pending_write_mems = tmp_deps->pending_write_mems;

  /* bb's structures are inherited by its successors.  */
  first_edge = e = OUT_EDGES (b);
  if (e <= 0)
    return;

  do
    {
      rtx x;
      int b_succ = TO_BLOCK (e);
      int bb_succ = BLOCK_TO_BB (b_succ);
      struct deps *succ_deps = bb_deps + bb_succ;

      /* Only bbs "below" bb, in the same region, are interesting.  */
      if (CONTAINING_RGN (b) != CONTAINING_RGN (b_succ)
	  || bb_succ <= bb)
	{
	  e = NEXT_OUT (e);
	  continue;
	}

      for (reg = 0; reg < max_reg; reg++)
	{
	  /* reg-last-uses lists are inherited by bb_succ.  */
	  for (u = tmp_deps->reg_last_uses[reg]; u; u = XEXP (u, 1))
	    {
	      if (find_insn_list (XEXP (u, 0),
				  succ_deps->reg_last_uses[reg]))
		continue;

	      succ_deps->reg_last_uses[reg]
		= alloc_INSN_LIST (XEXP (u, 0),
				   succ_deps->reg_last_uses[reg]);
	    }

	  /* reg-last-defs lists are inherited by bb_succ.  */
	  for (u = tmp_deps->reg_last_sets[reg]; u; u = XEXP (u, 1))
	    {
	      if (find_insn_list (XEXP (u, 0),
				  succ_deps->reg_last_sets[reg]))
		continue;

	      succ_deps->reg_last_sets[reg]
		= alloc_INSN_LIST (XEXP (u, 0),
				   succ_deps->reg_last_sets[reg]);
	    }

	  for (u = tmp_deps->reg_last_clobbers[reg]; u; u = XEXP (u, 1))
	    {
	      if (find_insn_list (XEXP (u, 0),
				  succ_deps->reg_last_clobbers[reg]))
		continue;

	      succ_deps->reg_last_clobbers[reg]
		= alloc_INSN_LIST (XEXP (u, 0),
				   succ_deps->reg_last_clobbers[reg]);
	    }
	}

      /* Mem read/write lists are inherited by bb_succ.  */
      link_insn = tmp_deps->pending_read_insns;
      link_mem = tmp_deps->pending_read_mems;
      while (link_insn)
	{
	  if (!(find_insn_mem_list (XEXP (link_insn, 0),
				    XEXP (link_mem, 0),
				    succ_deps->pending_read_insns,
				    succ_deps->pending_read_mems)))
	    add_insn_mem_dependence (succ_deps, &succ_deps->pending_read_insns,
				     &succ_deps->pending_read_mems,
				     XEXP (link_insn, 0), XEXP (link_mem, 0));
	  link_insn = XEXP (link_insn, 1);
	  link_mem = XEXP (link_mem, 1);
	}

      link_insn = tmp_deps->pending_write_insns;
      link_mem = tmp_deps->pending_write_mems;
      while (link_insn)
	{
	  if (!(find_insn_mem_list (XEXP (link_insn, 0),
				    XEXP (link_mem, 0),
				    succ_deps->pending_write_insns,
				    succ_deps->pending_write_mems)))
	    add_insn_mem_dependence (succ_deps,
				     &succ_deps->pending_write_insns,
				     &succ_deps->pending_write_mems,
				     XEXP (link_insn, 0), XEXP (link_mem, 0));

	  link_insn = XEXP (link_insn, 1);
	  link_mem = XEXP (link_mem, 1);
	}

      /* last_function_call is inherited by bb_succ.  */
      for (u = tmp_deps->last_function_call; u; u = XEXP (u, 1))
	{
	  if (find_insn_list (XEXP (u, 0),
			      succ_deps->last_function_call))
	    continue;

	  succ_deps->last_function_call
	    = alloc_INSN_LIST (XEXP (u, 0),
			       succ_deps->last_function_call);
	}

      /* last_pending_memory_flush is inherited by bb_succ.  */
      for (u = tmp_deps->last_pending_memory_flush; u; u = XEXP (u, 1))
	{
	  if (find_insn_list (XEXP (u, 0),
			      succ_deps->last_pending_memory_flush))
	    continue;

	  succ_deps->last_pending_memory_flush
	    = alloc_INSN_LIST (XEXP (u, 0),
			       succ_deps->last_pending_memory_flush);
	}

      /* sched_before_next_call is inherited by bb_succ.  */
      x = LOG_LINKS (tmp_deps->sched_before_next_call);
      for (; x; x = XEXP (x, 1))
	add_dependence (succ_deps->sched_before_next_call,
			XEXP (x, 0), REG_DEP_ANTI);

      e = NEXT_OUT (e);
    }
  while (e != first_edge);
}

/* Compute backward dependences inside bb.  In a multiple blocks region:
   (1) a bb is analyzed after its predecessors, and (2) the lists in
   effect at the end of bb (after analyzing for bb) are inherited by
   bb's successrs.

   Specifically for reg-reg data dependences, the block insns are
   scanned by sched_analyze () top-to-bottom.  Two lists are
   maintained by sched_analyze (): reg_last_sets[] for register DEFs,
   and reg_last_uses[] for register USEs.

   When analysis is completed for bb, we update for its successors:
   ;  - DEFS[succ] = Union (DEFS [succ], DEFS [bb])
   ;  - USES[succ] = Union (USES [succ], DEFS [bb])

   The mechanism for computing mem-mem data dependence is very
   similar, and the result is interblock dependences in the region.  */

static void
compute_block_backward_dependences (bb)
     int bb;
{
  rtx head, tail;
  int max_reg = max_reg_num ();
  struct deps tmp_deps;

  tmp_deps = bb_deps[bb];

  /* Do the analysis for this block.  */
  get_block_head_tail (BB_TO_BLOCK (bb), &head, &tail);
  sched_analyze (&tmp_deps, head, tail);
  add_branch_dependences (head, tail);

  if (current_nr_blocks > 1)
    propagate_deps (bb, &tmp_deps, max_reg);

  /* Free up the INSN_LISTs.  */
  free_deps (&tmp_deps);

  /* Assert that we won't need bb_reg_last_* for this block anymore.  */
  free (bb_deps[bb].reg_last_uses);
  free (bb_deps[bb].reg_last_sets);
  free (bb_deps[bb].reg_last_clobbers);
  bb_deps[bb].reg_last_uses = 0;
  bb_deps[bb].reg_last_sets = 0;
  bb_deps[bb].reg_last_clobbers = 0;
}
/* Remove all INSN_LISTs and EXPR_LISTs from the pending lists and add
   them to the unused_*_list variables, so that they can be reused.  */

static void
free_pending_lists ()
{
  int bb;

  for (bb = 0; bb < current_nr_blocks; bb++)
    {
      free_INSN_LIST_list (&bb_deps[bb].pending_read_insns);
      free_INSN_LIST_list (&bb_deps[bb].pending_write_insns);
      free_EXPR_LIST_list (&bb_deps[bb].pending_read_mems);
      free_EXPR_LIST_list (&bb_deps[bb].pending_write_mems);
    }
}

/* Print dependences for debugging, callable from debugger.  */

void
debug_dependencies ()
{
  int bb;

  fprintf (sched_dump, ";;   --------------- forward dependences: ------------ \n");
  for (bb = 0; bb < current_nr_blocks; bb++)
    {
      if (1)
	{
	  rtx head, tail;
	  rtx next_tail;
	  rtx insn;

	  get_block_head_tail (BB_TO_BLOCK (bb), &head, &tail);
	  next_tail = NEXT_INSN (tail);
	  fprintf (sched_dump, "\n;;   --- Region Dependences --- b %d bb %d \n",
		   BB_TO_BLOCK (bb), bb);

/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
	  if (USE_AUTOMATON_PIPELINE_INTERFACE)
	    {
	      fprintf (sched_dump, ";;   %7s%6s%6s%6s%6s%6s%14s\n",
		       "insn", "code", "bb", "dep", "prio", "cost",
		       "reservation");
	      fprintf (sched_dump, ";;   %7s%6s%6s%6s%6s%6s%14s\n",
		       "----", "----", "--", "---", "----", "----",
		       "-----------");
	    }
#endif /* #if AUTOMATON_PIPELINE_INTERFACE */

#if OLD_PIPELINE_INTERFACE
	  if (!USE_AUTOMATON_PIPELINE_INTERFACE)
	    {
/* END CYGNUS LOCAL -- vmakarov/dfa */
	  fprintf (sched_dump, ";;   %7s%6s%6s%6s%6s%6s%11s%6s\n",
	  "insn", "code", "bb", "dep", "prio", "cost", "blockage", "units");
	  fprintf (sched_dump, ";;   %7s%6s%6s%6s%6s%6s%11s%6s\n",
	  "----", "----", "--", "---", "----", "----", "--------", "-----");
/* CYGNUS LOCAL -- vmakarov/dfa */
	    }
#endif /* #if OLD_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/dfa */
	  for (insn = head; insn != next_tail; insn = NEXT_INSN (insn))
	    {
	      rtx link;
/* CYGNUS LOCAL -- vmakarov/dfa */
#if OLD_PIPELINE_INTERFACE
/* END CYGNUS LOCAL -- vmakarov/dfa */
	      int unit, range;
/* CYGNUS LOCAL -- vmakarov/dfa */
#endif /* #if OLD_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/dfa */

	      if (! INSN_P (insn))
		{
		  int n;
		  fprintf (sched_dump, ";;   %6d ", INSN_UID (insn));
		  if (GET_CODE (insn) == NOTE)
		    {
		      n = NOTE_LINE_NUMBER (insn);
		      if (n < 0)
			fprintf (sched_dump, "%s\n", GET_NOTE_INSN_NAME (n));
		      else
			fprintf (sched_dump, "line %d, file %s\n", n,
				 NOTE_SOURCE_FILE (insn));
		    }
		  else
		    fprintf (sched_dump, " {%s}\n", GET_RTX_NAME (GET_CODE (insn)));
		  continue;
		}

/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
	      if (USE_AUTOMATON_PIPELINE_INTERFACE)
		{
		  fprintf (sched_dump,
			   ";;   %s%5d%6d%6d%6d%6d%6d   ",
			   (SCHED_GROUP_P (insn) ? "+" : " "),
			   INSN_UID (insn),
			   INSN_CODE (insn),
			   INSN_BB (insn),
			   INSN_DEP_COUNT (insn),
			   INSN_PRIORITY (insn),
			   insn_cost (insn, 0, 0));
		  
		  recog_memoized (insn);
		  if (INSN_CODE (insn) < 0)
		    fprintf (sched_dump, "nothing");
		  else
		    print_reservation (sched_dump, insn);
		}
#endif /* #if AUTOMATON_PIPELINE_INTERFACE */

#if OLD_PIPELINE_INTERFACE
	      if (!USE_AUTOMATON_PIPELINE_INTERFACE)
		{
/* END CYGNUS LOCAL -- vmakarov/dfa */
	      unit = insn_unit (insn);
	      range = (unit < 0
		 || function_units[unit].blockage_range_function == 0) ? 0 :
		function_units[unit].blockage_range_function (insn);
	      fprintf (sched_dump,
		       ";;   %s%5d%6d%6d%6d%6d%6d  %3d -%3d   ",
		       (SCHED_GROUP_P (insn) ? "+" : " "),
		       INSN_UID (insn),
		       INSN_CODE (insn),
		       INSN_BB (insn),
		       INSN_DEP_COUNT (insn),
		       INSN_PRIORITY (insn),
		       insn_cost (insn, 0, 0),
		       (int) MIN_BLOCKAGE_COST (range),
		       (int) MAX_BLOCKAGE_COST (range));
	      insn_print_units (insn);
/* CYGNUS LOCAL -- vmakarov/dfa */
		}
#endif /* #if OLD_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/dfa */
	      fprintf (sched_dump, "\t: ");
	      for (link = INSN_DEPEND (insn); link; link = XEXP (link, 1))
		fprintf (sched_dump, "%d ", INSN_UID (XEXP (link, 0)));
	      fprintf (sched_dump, "\n");
	    }
	}
    }
  fprintf (sched_dump, "\n");
}

/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
/* The function removes marks about start of new cycle made in the first
   instruction scheduling.  Although regmove may remove them too.  */

static void
remove_new_cpu_cycle_marks (bb)
     int bb;
{
  rtx next_tail;
  rtx tail;
  rtx head;
  rtx insn;

  get_block_head_tail (BB_TO_BLOCK (bb), &head, &tail);

  next_tail = NEXT_INSN (tail);
  
  for (insn = head; insn != next_tail; insn = NEXT_INSN (insn))
    if (INSN_P (insn) && GET_MODE (insn) == TImode)
      PUT_MODE (insn, VOIDmode);
}
#endif
/* END CYGNUS LOCAL -- vmakarov/dfa */
/* Schedule a region.  A region is either an inner loop, a loop-free
   subroutine, or a single basic block.  Each bb in the region is
   scheduled after its flow predecessors.  */

static void
schedule_region (rgn)
     int rgn;
{
  int bb;
  int rgn_n_insns = 0;
  int sched_rgn_n_insns = 0;

  /* Set variables for the current region.  */
  current_nr_blocks = RGN_NR_BLOCKS (rgn);
  current_blocks = RGN_BLOCKS (rgn);

  init_deps_global ();

  /* Initializations for region data dependence analyisis.  */
/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
  if (reload_completed && USE_AUTOMATON_PIPELINE_INTERFACE)
    for (bb = 0; bb < current_nr_blocks; bb++)
      remove_new_cpu_cycle_marks (bb);
#endif
/* END CYGNUS LOCAL -- vmakarov/dfa */
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
  if (flag_rcsp && RGN_MAY_BE_PIPELINED_P (rgn))
    {
      rtx start_loop_body_insn, tail_loop_body_insn;
      rtx prev_loop_body_insn, loop_body_bound_insn;

      if (reload_completed)
	abort ();
      
      get_block_head_tail (rgn_bb_table[RGN_BLOCKS (rgn)
				       + RGN_NR_BLOCKS (rgn) - 1],
			   &start_loop_body_insn, &tail_loop_body_insn);
      
      loop_body_bound_insn = NEXT_INSN (tail_loop_body_insn);
      if (loop_body_bound_insn != NULL
	  && GET_CODE (loop_body_bound_insn) == BARRIER)
	loop_body_bound_insn = NEXT_INSN (loop_body_bound_insn);
      
      get_block_head_tail (rgn_bb_table[RGN_BLOCKS (rgn)],
			   &start_loop_body_insn, &tail_loop_body_insn);
      
      prev_loop_body_insn = PREV_INSN (start_loop_body_insn);
      start_RCSP_interface (start_loop_body_insn, loop_body_bound_insn, 5);
      
      RGN_PIPELINED_LOOP (rgn) = call_RCSP (sched_dump);
      
      finish_RCSP_interface ();
      
      if (RGN_PIPELINED_LOOP (rgn) != NULL_RTX)
	{
	  finish_deps_global ();
	  return;
	}
    }
  
  only_evaluate_scheduled_block_p = RGN_DO_NOT_SCHEDULE_P (rgn);
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
  bb_deps = (struct deps *) xmalloc (sizeof (struct deps) * current_nr_blocks);
  for (bb = 0; bb < current_nr_blocks; bb++)
    init_deps (bb_deps + bb);

  /* Compute LOG_LINKS.  */
  for (bb = 0; bb < current_nr_blocks; bb++)
    compute_block_backward_dependences (bb);

  /* Compute INSN_DEPEND.  */
  for (bb = current_nr_blocks - 1; bb >= 0; bb--)
    {
      rtx head, tail;
      get_block_head_tail (BB_TO_BLOCK (bb), &head, &tail);

      compute_forward_dependences (head, tail);
    }

  /* Set priorities.  */
  for (bb = 0; bb < current_nr_blocks; bb++)
    rgn_n_insns += set_priorities (BB_TO_BLOCK (bb));

  /* Compute interblock info: probabilities, split-edges, dominators, etc.  */
  if (current_nr_blocks > 1)
    {
      int i;

      prob = (float *) xmalloc ((current_nr_blocks) * sizeof (float));

      bbset_size = current_nr_blocks / HOST_BITS_PER_WIDE_INT + 1;
      dom = (bbset *) xmalloc (current_nr_blocks * sizeof (bbset));
      for (i = 0; i < current_nr_blocks; i++)
	dom[i] = (bbset) xcalloc (bbset_size, sizeof (HOST_WIDE_INT));

      /* Edge to bit.  */
      rgn_nr_edges = 0;
      edge_to_bit = (int *) xmalloc (nr_edges * sizeof (int));
      for (i = 1; i < nr_edges; i++)
	if (CONTAINING_RGN (FROM_BLOCK (i)) == rgn)
	  EDGE_TO_BIT (i) = rgn_nr_edges++;
      rgn_edges = (int *) xmalloc (rgn_nr_edges * sizeof (int));

      rgn_nr_edges = 0;
      for (i = 1; i < nr_edges; i++)
	if (CONTAINING_RGN (FROM_BLOCK (i)) == (rgn))
	  rgn_edges[rgn_nr_edges++] = i;

      /* Split edges.  */
      edgeset_size = rgn_nr_edges / HOST_BITS_PER_WIDE_INT + 1;
      edgeset_bitsize = rgn_nr_edges;
      pot_split = (edgeset *) xmalloc (current_nr_blocks * sizeof (edgeset));
      ancestor_edges
	= (edgeset *) xmalloc (current_nr_blocks * sizeof (edgeset));
      for (i = 0; i < current_nr_blocks; i++)
	{
	  pot_split[i] =
	    (edgeset) xcalloc (edgeset_size, sizeof (HOST_WIDE_INT));
	  ancestor_edges[i] =
	    (edgeset) xcalloc (edgeset_size, sizeof (HOST_WIDE_INT));
	}

      /* Compute probabilities, dominators, split_edges.  */
      for (bb = 0; bb < current_nr_blocks; bb++)
	compute_dom_prob_ps (bb);
    }

  /* Now we can schedule all blocks.  */
  for (bb = 0; bb < current_nr_blocks; bb++)
    {
      rtx head, tail;
      int b = BB_TO_BLOCK (bb);

      get_block_head_tail (b, &head, &tail);

      if (no_real_insns_p (head, tail))
	continue;

      current_sched_info->prev_head = PREV_INSN (head);
      current_sched_info->next_tail = NEXT_INSN (tail);

      if (write_symbols != NO_DEBUG)
	{
	  save_line_notes (b);
	  rm_line_notes (b);
	}

      /* rm_other_notes only removes notes which are _inside_ the
	 block---that is, it won't remove notes before the first real insn
 	 or after the last real insn of the block.  So if the first insn
	 has a REG_SAVE_NOTE which would otherwise be emitted before the
	 insn, it is redundant with the note before the start of the
	 block, and so we have to take it out.

	 FIXME: Probably the same thing should be done with REG_SAVE_NOTEs
	 referencing NOTE_INSN_SETJMP at the end of the block.  */
      if (INSN_P (head))
	{
	  rtx note;

	  for (note = REG_NOTES (head); note; note = XEXP (note, 1))
	    if (REG_NOTE_KIND (note) == REG_SAVE_NOTE)
	      {
		if (INTVAL (XEXP (note, 0)) != NOTE_INSN_SETJMP)
		  {
		    remove_note (head, note);
		    note = XEXP (note, 1);
		    remove_note (head, note);
		  }
		else
		  note = XEXP (note, 1);
	      }
	}

      /* Remove remaining note insns from the block, save them in
	 note_list.  These notes are restored at the end of
	 schedule_block ().  */
      rm_other_notes (head, tail);

      target_bb = bb;

      current_sched_info->queue_must_finish_empty
	= current_nr_blocks > 1 && !flag_schedule_interblock;

      schedule_block (b, rgn_n_insns);
      sched_rgn_n_insns += sched_n_insns;

      /* Update target block boundaries.  */
      if (head == BLOCK_HEAD (b))
	BLOCK_HEAD (b) = current_sched_info->head;
      if (tail == BLOCK_END (b))
	BLOCK_END (b) = current_sched_info->tail;

      /* Clean up.  */
      if (current_nr_blocks > 1)
	{
	  free (candidate_table);
	  free (bblst_table);
	  free (bitlst_table);
	}
    }

  /* Sanity check: verify that all region insns were scheduled.  */
  if (sched_rgn_n_insns != rgn_n_insns)
    abort ();

  /* Restore line notes.  */
  if (write_symbols != NO_DEBUG)
    {
      for (bb = 0; bb < current_nr_blocks; bb++)
	restore_line_notes (BB_TO_BLOCK (bb));
    }

  /* Done with this region.  */
  free_pending_lists ();

  finish_deps_global ();

  free (bb_deps);

  if (current_nr_blocks > 1)
    {
      int i;

      free (prob);
      for (i = 0; i < current_nr_blocks; ++i)
	{
	  free (dom[i]);
	  free (pot_split[i]);
	  free (ancestor_edges[i]);
	}
      free (dom);
      free (edge_to_bit);
      free (rgn_edges);
      free (pot_split);
      free (ancestor_edges);
    }
}

/* Indexed by region, holds the number of death notes found in that region.
   Used for consistency checks.  */
static int *deaths_in_region;

/* Initialize data structures for region scheduling.  */

static void
init_regions ()
{
  sbitmap blocks;
  int rgn;

  nr_regions = 0;
  rgn_table = (region *) xmalloc ((n_basic_blocks) * sizeof (region));
  rgn_bb_table = (int *) xmalloc ((n_basic_blocks) * sizeof (int));
  block_to_bb = (int *) xmalloc ((n_basic_blocks) * sizeof (int));
  containing_rgn = (int *) xmalloc ((n_basic_blocks) * sizeof (int));

  blocks = sbitmap_alloc (n_basic_blocks);

  /* Compute regions for scheduling.  */
  if (reload_completed
      || n_basic_blocks == 1
      || !flag_schedule_interblock)
    {
      find_single_block_region ();
    }
  else
    {
      /* Verify that a 'good' control flow graph can be built.  */
      if (is_cfg_nonregular ())
	{
	  find_single_block_region ();
	}
      else
	{
	  sbitmap *dom;
	  struct edge_list *edge_list;

	  dom = sbitmap_vector_alloc (n_basic_blocks, n_basic_blocks);

	  /* The scheduler runs after flow; therefore, we can't blindly call
	     back into find_basic_blocks since doing so could invalidate the
	     info in global_live_at_start.

	     Consider a block consisting entirely of dead stores; after life
	     analysis it would be a block of NOTE_INSN_DELETED notes.  If
	     we call find_basic_blocks again, then the block would be removed
	     entirely and invalidate our the register live information.

	     We could (should?) recompute register live information.  Doing
	     so may even be beneficial.  */
	  edge_list = create_edge_list ();

	  /* Compute the dominators and post dominators.  */
	  calculate_dominance_info (NULL, dom, CDI_DOMINATORS);

	  /* build_control_flow will return nonzero if it detects unreachable
	     blocks or any other irregularity with the cfg which prevents
	     cross block scheduling.  */
	  if (build_control_flow (edge_list) != 0)
	    find_single_block_region ();
	  else
	    find_rgns (edge_list, dom);

	  if (sched_verbose >= 3)
	    debug_regions ();

	  /* We are done with flow's edge list.  */
	  free_edge_list (edge_list);

	  /* For now.  This will move as more and more of haifa is converted
	     to using the cfg code in flow.c.  */
	  free (dom);
	}
    }

  deaths_in_region = (int *) xmalloc (sizeof (int) * nr_regions);

  /* Remove all death notes from the subroutine.  */
  for (rgn = 0; rgn < nr_regions; rgn++)
    {
      int b;

      sbitmap_zero (blocks);
      for (b = RGN_NR_BLOCKS (rgn) - 1; b >= 0; --b)
	SET_BIT (blocks, rgn_bb_table[RGN_BLOCKS (rgn) + b]);

      deaths_in_region[rgn] = count_or_remove_death_notes (blocks, 1);
    }

  sbitmap_free (blocks);
}

/* The one entry point in this file.  DUMP_FILE is the dump file for
   this pass.  */

void
schedule_insns (dump_file)
     FILE *dump_file;
{
  sbitmap large_region_blocks, blocks;
  int rgn;
  int any_large_regions;
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
  int pipelined_loop_p = 0;
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */

  /* Taking care of this degenerate case makes the rest of
     this code simpler.  */
  if (n_basic_blocks == 0)
    return;

  nr_inter = 0;
  nr_spec = 0;

  sched_init (dump_file);

  init_regions ();

  current_sched_info = &region_sched_info;
  
  /* Schedule every region in the subroutine.  */
  for (rgn = 0; rgn < nr_regions; rgn++)
    schedule_region (rgn);

/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
  if (!reload_completed)
    for (rgn = 0; rgn < nr_regions; rgn++)
      if (RGN_MAY_BE_PIPELINED_P (rgn) && RGN_PIPELINED_LOOP (rgn) != NULL_RTX)
	{
	  rtx insn, start_loop_body_insn, tail_loop_body_insn;
	  rtx prev_loop_body_insn, loop_body_bound_insn, loop_label;

	  get_block_head_tail (rgn_bb_table[RGN_BLOCKS (rgn)
					   + RGN_NR_BLOCKS (rgn) - 1],
			       &start_loop_body_insn, &tail_loop_body_insn);
	  
	  loop_body_bound_insn = NEXT_INSN (tail_loop_body_insn);
	  if (loop_body_bound_insn != NULL
	      && GET_CODE (loop_body_bound_insn) == BARRIER)
	    loop_body_bound_insn = NEXT_INSN (loop_body_bound_insn);
	  
	  get_block_head_tail (rgn_bb_table[RGN_BLOCKS (rgn)],
			       &start_loop_body_insn, &tail_loop_body_insn);
	  
	  prev_loop_body_insn = PREV_INSN (start_loop_body_insn);
	  
	  loop_label = prev_nonnote_insn (prev_loop_body_insn);
	  
	  if (loop_label == NULL_RTX || GET_CODE (loop_label) != CODE_LABEL)
	    abort ();
	  
	  /* We don't remove LOOP NOTES because the loop may have no
             such notes, e.g. loop is formed by gotos in C.  */
	  for (insn = loop_label;
	       insn != loop_body_bound_insn;
	       insn = NEXT_INSN (insn))
	    if (GET_CODE (insn) == JUMP_INSN && JUMP_LABEL (insn) != NULL_RTX)
	      --LABEL_NUSES (JUMP_LABEL (insn));
	  
	  NEXT_INSN (prev_loop_body_insn) = loop_body_bound_insn;
	  PREV_INSN (loop_body_bound_insn) = prev_loop_body_insn;
	  emit_insns_after (RGN_PIPELINED_LOOP (rgn), prev_loop_body_insn);

	  for (insn = NEXT_INSN (prev_loop_body_insn);
	       insn != loop_body_bound_insn;
	       insn = NEXT_INSN (insn))
	    if (GET_CODE (insn) == JUMP_INSN && JUMP_LABEL (insn) != NULL_RTX)
	      LABEL_NUSES (JUMP_LABEL (insn))++;

	  pipelined_loop_p = 1;
	}
    
  if (!pipelined_loop_p)
    {
#endif      
/* END CYGNUS LOCAL -- vmakarov/RCSP */
  /* Update life analysis for the subroutine.  Do single block regions
     first so that we can verify that live_at_start didn't change.  Then
     do all other blocks.   */
  /* ??? There is an outside possibility that update_life_info, or more
     to the point propagate_block, could get called with non-zero flags
     more than once for one basic block.  This would be kinda bad if it
     were to happen, since REG_INFO would be accumulated twice for the
     block, and we'd have twice the REG_DEAD notes.

     I'm fairly certain that this _shouldn't_ happen, since I don't think
     that live_at_start should change at region heads.  Not sure what the
     best way to test for this kind of thing...  */

  allocate_reg_life_data ();
  compute_bb_for_insn (get_max_uid ());

  any_large_regions = 0;
  large_region_blocks = sbitmap_alloc (n_basic_blocks);
  sbitmap_ones (large_region_blocks);

  blocks = sbitmap_alloc (n_basic_blocks);

  for (rgn = 0; rgn < nr_regions; rgn++)
    if (RGN_NR_BLOCKS (rgn) > 1)
      any_large_regions = 1;
    else
      {
	sbitmap_zero (blocks);
	SET_BIT (blocks, rgn_bb_table[RGN_BLOCKS (rgn)]);
	RESET_BIT (large_region_blocks, rgn_bb_table[RGN_BLOCKS (rgn)]);

	/* Don't update reg info after reload, since that affects
	   regs_ever_live, which should not change after reload.  */
	update_life_info (blocks, UPDATE_LIFE_LOCAL,
			  (reload_completed ? PROP_DEATH_NOTES
			   : PROP_DEATH_NOTES | PROP_REG_INFO));

#ifndef HAVE_conditional_execution
	/* ??? REG_DEAD notes only exist for unconditional deaths.  We need
	   a count of the conditional plus unconditional deaths for this to
	   work out.  */
	/* In the single block case, the count of registers that died should
	   not have changed during the schedule.  */
	if (count_or_remove_death_notes (blocks, 0) != deaths_in_region[rgn])
	  abort ();
#endif
      }

  if (any_large_regions)
    {
      update_life_info (large_region_blocks, UPDATE_LIFE_GLOBAL,
			PROP_DEATH_NOTES | PROP_REG_INFO);
    }
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
    }
#endif /* #if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/RCSP */

  /* Reposition the prologue and epilogue notes in case we moved the
     prologue/epilogue insns.  */
  if (reload_completed)
    reposition_prologue_and_epilogue_notes (get_insns ());

  /* Delete redundant line notes.  */
  if (write_symbols != NO_DEBUG)
    rm_redundant_line_notes ();

  if (sched_verbose)
    {
      if (reload_completed == 0 && flag_schedule_interblock)
	{
	  fprintf (sched_dump,
		   "\n;; Procedure interblock/speculative motions == %d/%d \n",
		   nr_inter, nr_spec);
	}
      else
	{
	  if (nr_inter > 0)
	    abort ();
	}
      fprintf (sched_dump, "\n\n");
    }

  /* Clean up.  */
  free (rgn_table);
  free (rgn_bb_table);
  free (block_to_bb);
  free (containing_rgn);

  sched_finish ();

  if (edge_table)
    {
      free (edge_table);
      edge_table = NULL;
    }

  if (in_edges)
    {
      free (in_edges);
      in_edges = NULL;
    }
  if (out_edges)
    {
      free (out_edges);
      out_edges = NULL;
    }

  sbitmap_free (blocks);
  sbitmap_free (large_region_blocks);

  free (deaths_in_region);
/* CYGNUS LOCAL -- vmakarov/dfa */
#if AUTOMATON_PIPELINE_INTERFACE
  free (curr_state);
  dfa_finish ();
#endif /* if AUTOMATON_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/dfa */
/* CYGNUS LOCAL -- vmakarov/RCSP */
#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE
  if (pipelined_loop_p)
    {
      struct loops loops;

      /* Pipelining loops changes CFG too much.  We need to restore
         flow anaysis data.  */
      find_basic_blocks (get_insns (), max_reg_num (),
			 (sched_verbose <= 1 ? NULL : sched_dump));
      
      flow_loops_find (& loops, LOOP_TREE);

      if (flag_guess_branch_prob)
	estimate_probability (&loops);
      
      flow_loops_free (&loops);
      
      life_analysis (get_insns (), (sched_verbose <= 1 ? NULL : sched_dump),
		     PROP_DEATH_NOTES | PROP_REG_INFO);
    }
#endif
/* END CYGNUS LOCAL -- vmakarov/RCSP */
}

/* CYGNUS LOCAL -- vmakarov/RCSP */


#if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE

/* This page contains of forward declarations of RCSP software
   pipelining.  */

typedef struct insn *insn_t;
typedef struct loc *location_t;
typedef state_t cpu_state_t;

/* The following describes type of writting.  */
enum write_kind
{
  /* The following is obligatory writing.  */
  USUAL_WRITE = 0,
  /* The following is nonobligatory writing.  */
  POSSIBLE_WRITE,
  /* The following is write barrier, i.e. any writes can not be moved
     through the barrier.  */
  WRITE_BARRIER
};

typedef enum write_kind write_kind_t;

static void * rcsp_malloc  PARAMS ((size_t));
static void * rcsp_realloc PARAMS ((void *, size_t, size_t));
static void   rcsp_free    PARAMS ((void *));

struct packet;
static int nonbranch_p PARAMS ((insn_t));
static int uncond_branch_p PARAMS ((insn_t));
static int cond_branch_p PARAMS ((insn_t));
static void convert_cond_branch PARAMS ((insn_t));
static insn_t create_uncond_branch PARAMS ((insn_t));
static int dest_packet_requires_jump_p PARAMS ((struct packet *));
static int insn_eq_p PARAMS ((insn_t, insn_t));
static unsigned insn_hash PARAMS ((insn_t));
static const char *insn_name PARAMS ((insn_t, int));
static int insn_dummy_p PARAMS ((insn_t));
static insn_t insn_copy PARAMS ((insn_t));
static int insn_dep_time PARAMS ((insn_t, insn_t, int));
static int insn_risky_speculative_move_p PARAMS ((insn_t, struct packet *));
static void cpu_state_reset PARAMS ((cpu_state_t));
static int cpu_state_issue PARAMS ((insn_t, cpu_state_t));
static void cpu_state_advance_cycle PARAMS ((cpu_state_t));
static int cpu_state_size PARAMS ((void));
static void cpu_state_union PARAMS ((cpu_state_t, cpu_state_t));
static int loc_eq_p PARAMS ((location_t, location_t));
static unsigned loc_hash PARAMS ((location_t));
static location_t loc_sub PARAMS ((location_t, int));
static location_t loc_super PARAMS ((location_t, int));
static location_t getloc PARAMS ((insn_t, int, int *, write_kind_t *));


/* This page contains forward declarations of functions needed for the
   interface of software pipeliner.  */
static void add_vector_element PARAMS ((varray_type *, void *, int *));
static struct loc_link *new_loc_link PARAMS ((struct loc *,
					      struct loc_link *));
static struct loc *new_loc PARAMS ((rtx, int, int, struct loc *));
static struct loc *new_insn_loc PARAMS ((insn_t, rtx, int, int));
static void free_loc PARAMS ((struct loc *));
static void free_insn_locs PARAMS ((struct loc *));
static void process_reads PARAMS ((rtx, insn_t));
static void process_writes PARAMS ((rtx, insn_t));
static void process_insn PARAMS ((insn_t));
static struct insn *new_insn PARAMS ((rtx, char *, int));
static unsigned jump_packet_hash PARAMS ((const void *));
static int jump_packet_eq PARAMS ((const void *, const void *));
static void jump_packet_free PARAMS ((void *));

static void insert_jump_packet PARAMS ((rtx, struct packet *, int));
static struct jump_packet *find_jump_packet PARAMS ((rtx));
static void traverse_jumps_table PARAMS ((int (*) PARAMS ((void **, void *)),
					  void *));
static void initiate_jumps_table PARAMS ((void));
static void finish_jumps_table PARAMS ((void));
static void update_regs_living_after_loop PARAMS ((rtx));
static int rtx_sets_live_reg PARAMS ((rtx));
static int insn_sets_live_reg_p PARAMS ((rtx));
static int second_jumps_pass PARAMS ((void **, void *));
static void connect_packets PARAMS ((struct packet *, struct packet **,
				     struct packet **));
static void connect_jump_packets PARAMS ((struct packet *,
					  struct jump_packet *));
static void redirect_jump_packets PARAMS ((struct jump_packet *, rtx));
static void clear_exit_packet_flag PARAMS ((struct packet *));
static struct packet *unroll PARAMS ((rtx, rtx, rtx *));
static void set_loc_inheritance PARAMS ((struct loc *));
static int nonjump_p PARAMS ((rtx));
static int cond_jump_p PARAMS ((rtx));
static int uncond_jump_p PARAMS ((rtx));
static void add_jump_labels PARAMS ((struct packet *));
static void emit_notes_list PARAMS ((rtx));
static struct packet *emit_packet PARAMS ((struct packet *));


/* -----All the folowing code is abstract data `software pipeliner'.  ----  */

/* Idea of this software pipeline algorithm is taken from RCSP
   (resource-constrained software pipelining).  You can find RCSP
   description e.g. in

      Resource-constrained Software Pipelining by A. Aiken,
      A. Nicolau, and S. Novack IEEE Transactions on Parallel and
      Distributed Systems, 6(12), pp. 1248--1270, December 1995.

   The algorithm is trying to schedule loop insns (operations)
   by searching repeating patterns of loop insns (operations).

   This implementation assumes that all current and future
   reservations of processor functional units are coded by data block
   of fixed length.  This is good assumption if we use pipeline
   hazards recognizer implemented as deterministic finite automaton.

   Actually, usage of backtracking could find optimally software
   pipelined loop.  The core of algorithm also could be used for
   inter-block scheduling with possibility of copying insns too.  So it
   is very powerful method for possible future improvements of the
   scheduler.

   What could we make the next for improving RCSP?

   1. Better aliasing memory locations.
   
   2. Register renaming for registers which are temporary on each
      iteration.

   3. Loop unrolling with epilogue loop for loops whose number of
      iterations is known before starting loop.  This could permit to
      move risky insns from one iteration to another.

   4. Better recognition of risky insns and/or transformation of them
      into corresponding non exceptional insns.

   5. Modification of algorithm for making more compact pipelined
      loops and /or more effective ones (see comments for merging
      frontier states).  Additional heuristics in choosing insn for
      scheduling (critical path length, register pressure, basic block
      movement, probability for speculative moving).  Implementation
      of more accurate merging cpu states.

   6. Adding jump insns during scheduling and conversion of conditional
      branches.  More one conditional branches in packet.

   7. Speedup the pipeliner.  We could add new iteration and update
      DDG only when it is necessary.  We could specially process
      simple cases for moving operation in basic block.

   1-4 are relative to the pipeliner interface.  5-7 are for the
   pipeliner itself.  */

#include <stdio.h>
#include <string.h>
#include "hashtab.h"

#define FALSE 0
#define TRUE 1

/* The following macro value defines unrolling the loop.  There will
   be (RCSP_UNROLL_FACTOR * maximal iteration look ahead + 1) loop bodies.
   The value should be more 1.  */
#define RCSP_UNROLL_FACTOR 2

/* The following macro value defines maximal size of successfuly
   pipelined loop in original loop size.  The value should be more
   1.  */
#define RCSP_EXPANSION_LOOP_FACTOR 8

typedef struct op *op_t;
typedef struct dep *dep_t;
typedef struct packet *packet_t;
typedef struct avail_list *avail_list_t;
typedef struct locref *locref_t;
typedef htab_t locrefset_t;
typedef htab_t packet_set_t;

enum dep_mode
{
  DEF_USE = 0, /* True dependency.  */
  DEF_DEF,     /* Output dependency.  */
  USE_DEF      /* Anti dependency.  */
};

typedef enum dep_mode dep_mode_t;

/* The following enumeration decribes how one packet is connected to
   another packet in CFG.  */
enum achievable_mode
{
  IGNORE_ACHIEVABILITY = 0,
  UNACHIEVABLE,
  BASIC_BLOCK_PATH,	 /* Basic block connection.  */
  OTHER_PATH		 /* All other cases.  */
};

typedef enum achievable_mode achievable_mode_t;

/* The following is a packet of operations.  All operations in a
   packet start simultaneously on the same processor cycle.

   There are three types of packets:

   a) formed packet.  They refer only for formed packets through
      member `pred' and only for formed/or frontier packets through
      `succ'.  Such packets are not changed any more.

   b) frontier packet whose forming started but has not finished yet.
      They refer only for formed packets through `pred' and only for
      interior packets through `succ'.  Directed CFGs starting with
      frontier packets are always acyclic.

   c) interior packets which simply contains original ops (insns).
      One interior packet corresponds to at most one op (insn).  They
      refer only for frontier or/and interior packets through `pred'
      and only for interior packets through `succ'.  Directed CFGs
      starting with interior packets are always acyclic.  Only
      interior packets can be in list `available'.  Finally, such
      packets will be removed.

   When software pipelining starts, we have only interior packets
   which form control flow directed acyclic graph (DA CFG).  When
   software pipelining finishes, we have only formed states.  */

struct packet
{
  /* All allocated packets are chained by this member.  */
  struct packet *all_packets_chain;
  /* Unique packet number.  */
  int number;
  /* The following is TRUE if the packet is the last packet on the
     exit from the loop.  It may be more one such packet when there
     are different exits from the loop.  Stop packets contain only
     dummy operations.  */
  char stop_packet_p;
  /* The following is TRUE if the packet is actually executed only
     once in the loop (it is on an exit branch from the loop).  Stop
     packet has also value TRUE.  */
  char exit_packet_p;
  /* The following member values are TRUE if the packet is
     correspondingly frontier packet and formed packet.  */
  char frontier_packet_p;
  char formed_packet_p;
  /* The following is defined only for stop packets.  It is TRUE for
     stop packet which corresponds the first insn after the loop.  */
  int main_exit_p;
  /* The following member is used as temporary variable during
     evaluation of packet achievability.  */
  achievable_mode_t achievable_result;
  /* Allocated length of the two next arrays.  */
  int succ_length, pred_length;
  /* Successors (control flow from the packet) and predecessors
   (control flow to the packet).  The end of arrays is NULL
   element.  */
  packet_t *succ, *pred;
  /* The following members are used to traverse packets for different
     goals (see comments for `last_packet_visit_number...').  */
  int visit_number1, visit_number2, visit_number3;
  int visit_number4, visit_number5;
  /* The following values refer for list of ops (the first and the
     last list elements) in the packet in the same order as they were
     added to packet during scheduling.  */
  op_t first_op, last_op;
  /* The minimal iteration number of the ops (excluding ops in exit
     packets -- remember that ops from exit packets are never moved
     into the loop) which are in packet and in its available_list.
     The original value is -1.  */
  int min_it_n;
  /* The following temporary member is used for transformation of CFG
     and DDG.  Its value is a new copy of given packet.  */
  packet_t copy;
  /* The following is defined only for formed packets and for the
     current frontier packet.  This is available list for the packet.
     Available list contains all operations which are movable into the
     packet and depended on only issued operations.  Lists for formed
     packets also contain operations achievable from the packet (not
     only available).  The list contains only operations with the
     iteration lookahead.  The original article is completely wrong
     that only available ops are sufficient to reflect scheduler
     states which are used to searching for repetitive op patterns.
     ??? Storing achievable ops is too expensive.  Can we decrease
     this list.  */
  avail_list_t avail_list;
  /* The following is a reference for structure which represents
     functional units state.  The structure is defined only for
     formed and frontier packets.  */
  cpu_state_t cpu_state;
  /* The following is defined only for frontier packets.  */
  packet_t next_frontier_packet;
  /* The following member is used to link packets (e.g. for forming
     list of copied packets).  */
  packet_t next_packet;
  /* The following member is relative time when ops in packet is
     issued.  The value is negative for interior packets.  */
  int issue_time;
};

/* The following describes list of packets.  */
struct avail_list
{
  /* All allocated els are chained by this member.  */
  struct avail_list *all_avail_list_els_chain;
  /* The following is operation which could be moved into frontier
     packet to which the available list belongs.  */
  op_t op;
  /* The prev, next list element.  */
  avail_list_t prev_avail, next_avail;
  /* The following is frontier or formed packet to which given
     available list belongs.  */
  packet_t packet;
  /* The following member is used for sets "available".  Ops in the
     set are of only interior packets.  The member value is time until
     all data used in the operation (insn) are accessible.  */
  int delay;
  /* The following is cost of moving the op to the packet to which the
     available list belongs.  */
  int cost;
  /* The following is TRUE if we should make changes in CFG before
     moving OP into the packet.  */
  int inconsistence_p;
  /* The following member value is FALSE for achievable but available
     ops.  */
  int avail_p;
};

/* The following describes an operation (insn).  */
struct op
{
  /* All allocated ops are chained by this member.  */
  struct op *all_ops_chain;
  /* Internal op representation.  */
  insn_t insn;
  /* Iteration number of the insn.  */
  int it_num;
  /* Packet to which op belongs.  */
  packet_t packet;
  /* The member values are packets which are targets of conditional
     branch op when the condition is correspondingly true or false.
     If op is not a branch the values are NULL.  If op is
     unconditional branch false_target is always NULL and true_target
     is always not NULL.  */
  packet_t true_target;
  packet_t false_target;
  /* Input and output data dependency arcs .*/
  dep_t out_deps;
  dep_t in_deps;
  /* The next operation in the packet.  */
  op_t next_op;
  /* The following member is used to form list of operations to which
     an input dependency has added.  If the value is not NULL then op
     is in the list.  */
  op_t next_added_in_deps_op;
  /* The following member value is not NULL only for added
     unconditional branch.  Its value is conditional branch op which
     resulted in adding this unconditional branch.  */
  op_t initiator_op;
  /* The following values are TRUE if op has been scheduled and if all
     its input ops have been started.  Scheduled op can not be
     ready.  */
  char scheduled_p, ready_p;
  /* The following values are the previous and the next ready op with
     the same iteration.  These values are defined if scheduled_p is
     FALSE.  */
  op_t prev_ready_op, next_ready_op;
  /* The following flag is TRUE for operations which define and use
     intersected locations.  Such operations are never considered as
     available operation requiring making CFG move consistent.
     Without such flag we will be not able to determine correctly that
     operation can be moved because we do not store dependencies
     inside operations, e.g. when operation modifies a location.  */
  char use_def_p;
 /* The following member may be FALSE only for conditional branch when
    insn makes jump onto false_target and true_target is achieved by
    fall through.  */
  char true_jump_p;
  /*** The following members are defined only for conditional branch: ***/
  /* The following member is TRUE if insn can be converted, i.e. jump
     and fall through (and as consequence condition of jump) can be
     exchanged.  */
  char can_be_converted_p;
  /* The member value is TRUE when we can not convert conditional
     branch any more because there are subsequent insns in the packet
     achieved by fall through or we fixed it when we added CFG
     consistent jumps.  The initial value is FALSE.  */
  char target_fixed_p;
  /* The following member value is TRUE if we already processed the
     conditional branch to add unconditional jumps.  Remember that in
     initial CFG we have no unconditional jumps at all.  Initial value
     is FALSE.  */
  char processed_4_consistent_jumps_p;
};

/* The following describes DDG arc.  */
struct dep
{
  /* All allocated deps are chained by this member.  */
  struct dep *all_deps_chain;
  /* Mode of data dependency.  */
  dep_mode_t dep_mode;
  /* Delay on the data dependency.  */
  int latency;
  /* Source and target of the data dependency.  */
  op_t from, to;
  /* This is dependency on the following location in op `to'.  */
  location_t to_loc;
  /* The previous and next output dependency of `from' with the same
     to_loc.  */
  dep_t prev_out_loc_dep, next_out_loc_dep;
  /* The prev/next in/out DDG arcs for operations `to' and `from'.  */
  dep_t prev_in, prev_out;
  dep_t next_in, next_out;
};

/* The following describes a location reference in interior packet.
   There may be several usages of one location (but they are deleted
   when the location definition is occurred).  */
struct locref
{
  /* Flag of that this is reference for the location usage.  */
  char usage_p;
  /* The following is defined only for location definition.  Value
     true means that only ops defining (but using) the loc will be
     depended on this op.  */
  char write_barrier_p;
  location_t loc;
  /* Operation in which the loc is mentioned.  */
  op_t op;
  /* The next locref with the loc defined or used in OP->PACKET.  */
  locref_t next_locref;
};

/* The following variable value is used maximal look ahead (window) in
   iterations.  */
static int iteration_look_ahead;

/* The following variable value is current time.  */
static int curr_time;

/* Define this to debug RCSP.  */
#define RCSP_DEBUG

#ifdef RCSP_DEBUG
static int rcsp_debug_level = 3;
#endif

/* Define this if you want fast evaluation of achievability of one
   packet from another one.  The cache size is (sizeof (int) + 1) * (2
   ** round_up (log2 (packets_number))) ** 2.  So the cache may be
   huge in many cases (e.g. 5MB for 1K packets).  */
#define USE_ACHIEVABLE_CACHE

static void free_op PARAMS ((op_t));
static op_t get_free_op PARAMS ((void));
static op_t op_create PARAMS ((insn_t, int));
static void op_delete PARAMS ((op_t));
static op_t op_copy PARAMS ((op_t));
static void op_become_ready PARAMS ((op_t));
static void op_become_not_ready PARAMS ((op_t));
static void op_become_scheduled PARAMS ((op_t));
static void start_ops PARAMS ((void));
static void finish_ops PARAMS ((void));

static void avail_list_el_free PARAMS ((avail_list_t));
static avail_list_t avail_list_el_create PARAMS ((packet_t, op_t, int));
static void avail_list_free PARAMS ((avail_list_t));
static int avail_list_length PARAMS ((avail_list_t));
static avail_list_t avail_list_sort PARAMS ((avail_list_t,
					     int (*) (avail_list_t *,
						      avail_list_t *)));
static void start_avail_lists PARAMS ((void));
static void finish_avail_lists PARAMS ((void));


static void free_packet PARAMS ((packet_t));
static packet_t get_free_packet PARAMS ((void));
static void expand_packet_succ_pred PARAMS ((packet_t, int, int));
#ifdef USE_ACHIEVABLE_CACHE
static void expand_achievable_cache PARAMS ((void));
#endif
static packet_t packet_create PARAMS ((int));
static void packet_delete PARAMS ((packet_t));
static void packet_append_op PARAMS ((packet_t, op_t, int));
static packet_t packet_copy PARAMS ((packet_t));
static packet_t *packet_find_in_arr PARAMS ((packet_t *, packet_t));
static void packet_del_from_arr PARAMS ((packet_t *, packet_t));
static void packet_add_to_arr PARAMS ((packet_t, packet_t, int));
static void packet_connect PARAMS ((packet_t, packet_t, int, int));
static op_t packet_find_branch PARAMS ((packet_t, packet_t, int));
static void packet_change_dest PARAMS ((packet_t, packet_t, packet_t));
static void packet_change_src PARAMS ((packet_t, packet_t, packet_t));
static void start_packets PARAMS ((void));
static void finish_packets PARAMS ((void));

static void free_dep PARAMS ((dep_t));
static dep_t get_free_dep PARAMS ((void));
static void dep_delete PARAMS ((dep_t));
static dep_t dep_out_find PARAMS ((op_t, location_t));
static void dep_create PARAMS ((op_t, op_t, location_t, dep_mode_t, int));
static int in_deps_cmp PARAMS ((const void *, const void *));
static void sort_op_in_deps PARAMS ((op_t));
static void dep_sort_added_in_deps_op PARAMS ((void));
static unsigned dep_hash PARAMS ((const void *));
static int dep_eq PARAMS ((const void *, const void *));
static void start_deps PARAMS ((void));
static void finish_deps PARAMS ((void));

static unsigned schedule_state_hash PARAMS ((const void *));
static int schedule_state_op_eq PARAMS ((op_t, op_t, int *, int *));
static int schedule_state_eq PARAMS ((const void *, const void *));
static packet_set_t packet_set_create PARAMS ((void));
static void packet_set_delete PARAMS ((packet_set_t));
static packet_t packet_set_insert PARAMS ((packet_set_t, packet_t));

static void start_pipeline_data PARAMS ((void));
static void finish_pipeline_data PARAMS ((void));

static void free_locref PARAMS ((locref_t));
static locref_t get_free_locref PARAMS ((void));
static unsigned locref_hash PARAMS ((const void *));
static int locref_eq PARAMS ((const void *, const void *));
static locref_t locref_find PARAMS ((location_t, op_t));
static void locref_add PARAMS ((op_t, location_t, int, int));
static void locref_remove_all PARAMS ((location_t, op_t));
static void start_locrefs PARAMS ((void));
static int free_htab_locrefs PARAMS ((void **, void *));
static void finish_locrefs PARAMS ((void));

static packet_t create_dummy_packet PARAMS ((void));
static void add_locref PARAMS ((op_t, location_t, int, write_kind_t));
static int process_locrefs PARAMS ((op_t, location_t, int, int,
				    op_t, location_t));
static void process_packet_op_locs PARAMS ((packet_t, op_t, location_t,
					    int, write_kind_t));
static void process_op_locs PARAMS ((op_t));
static packet_t link_packets PARAMS ((packet_t, packet_t));
static void build_deps PARAMS ((packet_t));

static packet_t frontier_packet_list_head_remove PARAMS ((void));
static int frontier_packet_list_length PARAMS ((void));
static void add_frontier_packet PARAMS ((packet_t));
static int frontier_packet_sort_cmp PARAMS ((const void *, const void *));
static void frontier_packet_list_sort_and_merge PARAMS ((FILE *,
							 const char *));
static void start_frontier_packet_list PARAMS ((void));
static void finish_frontier_packet_list PARAMS ((void));

static achievable_mode_t achievable_1 PARAMS ((packet_t, packet_t));
static achievable_mode_t achievable PARAMS ((packet_t, packet_t));
static int dom_p_1 PARAMS ((packet_t, packet_t, packet_t));
static int dom_p PARAMS ((packet_t, packet_t));
static int postdom_p_1 PARAMS ((packet_t, packet_t));
static int postdom_p PARAMS ((packet_t, packet_t));

static int kill_loc_move_p PARAMS ((op_t, packet_t, op_t, location_t));
static int kill_move_p PARAMS ((op_t, packet_t));

static void mark_src_packets PARAMS ((op_t, packet_t));
static int mark_covered_1 PARAMS ((op_t, packet_t, int *));
static int mark_covered PARAMS ((op_t, packet_t));

static int traverse_CFG_for_consistent_cost PARAMS ((op_t, packet_t));
static int move_cost PARAMS ((op_t, packet_t, int *));

static int get_op_ready_time PARAMS ((op_t, packet_t));

static void copy_dependencies_4_consistent_move PARAMS ((packet_t));
static void remove_obsolete_dependencies_4_move_consistence PARAMS((packet_t));
static void modify_packet_CFG PARAMS ((packet_t));
static void make_move_consistent PARAMS ((op_t, packet_t));

static void update_loc_deps PARAMS ((op_t, op_t, location_t, int));
static void update_deps PARAMS ((op_t));
static op_t create_simple_jump PARAMS ((packet_t, packet_t, op_t));
static void create_avail_list_el PARAMS ((packet_t, op_t, int, int, int, int));
static int needs_jump_p PARAMS ((packet_t, packet_t));
static void add_consistent_jumps PARAMS ((packet_t));
static void update_ready_avail_ops PARAMS ((op_t, int));
static void schedule_op PARAMS ((op_t, packet_t));
static int avail_op_cmp PARAMS ((avail_list_t *, avail_list_t *));
static void add_next_packets PARAMS ((packet_t));
static avail_list_t get_best_candidate_avail_op PARAMS ((packet_t, int *));
static void make_packet_avail_list PARAMS ((packet_t));
#ifndef WRONG_ORIGINAL_RCSP
static void add_achievable_ops PARAMS ((packet_t, packet_t));
#endif
static void schedule_packet PARAMS ((packet_t, FILE *f, const char *));
static void add_independent_ready_ops PARAMS ((packet_t));
static void mark_achievable_packets PARAMS ((packet_t));
static packet_t link_unachievable_packets PARAMS ((packet_t, packet_t));
static void redirect_packet PARAMS ((packet_t, packet_t, packet_t,
				     FILE *f, const char *));
static void add_jumps_4_cond_branches PARAMS ((packet_t));
static void add_jumps_on_back_edges PARAMS ((packet_t));
static void add_the_rest_jumps PARAMS ((packet_t));
static void add_exit_jump PARAMS ((packet_t, int));
static void add_jumps PARAMS ((packet_t));

#ifdef RCSP_DEBUG
static void check_CFG PARAMS ((packet_t));
static void check_DDG PARAMS ((packet_t));
static void check_graphs PARAMS ((packet_t));
static void print_ready_ops PARAMS ((FILE *f, const char *));
static void print_avail_list PARAMS ((FILE *f, avail_list_t,
				      const char *, int));
static const char *get_packet_name PARAMS ((packet_t, int));
static void print_packet PARAMS ((FILE *, packet_t, const char *prefix, int));
static void print_packets_1 PARAMS ((FILE *, packet_t,
				     const char *prefix, int));
static void print_packets PARAMS ((FILE *, packet_t, const char *prefix, int));
#endif /* #ifdef RCSP_DEBUG */

static packet_t pipeline PARAMS ((packet_t, FILE *, const char *));



#ifdef USE_ACHIEVABLE_CACHE

/* CFG_shift is more than the following value we don't use the cache.  */
#define MAX_CFG_SHIFT 10

/* The following variables are used to faster evaluation of packets
   achievability.  */
static int CFG_tick;         /* Increased every time when CFG is changed.    */
static int *CFG_validation;  /* Used for validation of the next array value.  */
static char *CFG_achievable; /* Cache of achievable values.                  */
static int CFG_shift;        /* Used for fast access to the previous arrays.  */

/* This macro is used to evaluate achievability of the packet TO from
   packet FROM when the cache may be used.  */
#define ACHIEVABLE(from, to)						      \
  (CFG_shift <= MAX_CFG_SHIFT						      \
   &&									      \
   CFG_validation [((from)->number << CFG_shift) + (to)->number] == CFG_tick  \
   ? (achievable_mode_t) CFG_achievable [((from)->number << CFG_shift)	      \
					+ (to)->number]			      \
   : achievable (from, to))

#else

#define ACHIEVABLE(from, to) achievable (from, to)

#endif /* #ifdef USE_ACHIEVABLE_CACHE */

/* This is dummy packet which has no predecessors and contains one
   dummy operation.  It is used to store dummy definitions of
   locations used somewhere in the CFG and not defined somewhere upper
   in the CFG.  */
static packet_t top_packet;

/* The following variable contains current number of CFG traversing
   for different goals:

   1. Original building DDG (including sorting input dependencies),
      evaluating achievability of packets in CFG and post-dominator
      relation in CFG, removing unachievable packets, and output of
      pipelined insns.

   2. Traversing graph for searching for set `covered'.

   3. Evaluating cost of making CFG move consistent.

   4. Adding elements to available list of new packets and outputting
      additional jumps on back edges.

   5. Checking CFG and DDG, printing packets, and outputting
      additional jumps.  */

static int last_packet_visit_number1;
static int last_packet_visit_number2;
static int last_packet_visit_number3;
static int last_packet_visit_number4;
static int last_packet_visit_number5;

/* The following variable value is length (in bytes) of structure
   which represents cpu functional units state.  */
static int cpu_state_length;

/* The following variable value is a reference to temporary structure
   which represents cpu functional units state.  */
static cpu_state_t temp_cpu_state;



/* This page contains functions for work with operations (insns).  */

/* The following variable value is reference for array of ready ops.
   Each array element with index N starts list of ready ops with
   iteration number N.  It should only be read from outside the
   page.  */
static op_t *ready_ops;

/* The following variable value is size of array ready_ops (number of
   loop body iterations after unrolling).  It should only be read from
   outside the page.  */
static int ready_ops_size;

/* The following variavle value is number of ops which are already
   scheduled.  It should only be read from outside the page.  */
static int scheduled_ops_number;

/* The following variable refers for unused ops.  */
static op_t free_ops;

/* The following variable refers for all allocated ops.  */
static op_t all_ops;

/* The following variable value is maximal index of ready_ops for
   which ready_ops [i] == NULL for all i < MIN_READY_OP_IT_NUM.  */
static int min_ready_op_it_num;

/* The following function frees OP.  */
static void
free_op (op)
     op_t op;
{
  op->next_op = free_ops;
  free_ops = op;
}

/* The following function returns free op.  */
static op_t
get_free_op ()
{
  op_t result;

  if (free_ops)
    {
      result = free_ops;
      free_ops = free_ops->next_op;
    }
  else
    {
      result = rcsp_malloc (sizeof (struct op));
      result->all_ops_chain = all_ops;
      all_ops = result;
    }
  return result;
}

/* The following function creates and returns new operation with given
   characteristics.  */
static op_t
op_create (insn, iter)
     insn_t insn;
     int iter;
{
  op_t result;

  result = get_free_op ();
  result->insn = insn;
  result->it_num = iter;
  result->packet = NULL;
  result->true_target = result->false_target = NULL;
  result->in_deps = result->out_deps = NULL;
  result->next_op = NULL;
  result->next_added_in_deps_op = NULL;
  result->initiator_op = NULL;
  result->scheduled_p = FALSE;
  result->ready_p = FALSE;
  result->next_ready_op = result->prev_ready_op = NULL;
  result->can_be_converted_p = result->processed_4_consistent_jumps_p
    = result->use_def_p = FALSE;
  result->true_jump_p = result->target_fixed_p = TRUE;
  return result;
}

/* The following function deletes operation.  If OP is ready then it
   is removed from ready list.  */
static void
op_delete (op)
     op_t op;
{
  if (op->ready_p)
    {
      /* Remove from ready list.  */
      if (op->prev_ready_op == NULL)
	ready_ops [op->it_num] = op->next_ready_op;
      else
	op->prev_ready_op->next_ready_op = op->next_ready_op;
      if (op->next_ready_op != NULL)
	op->next_ready_op->prev_ready_op = op->prev_ready_op;
    }
  if (op->scheduled_p)
    scheduled_ops_number--;
  free_op (op);
}

/* The following function makes copy of OP (op should be not
   scheduled).  If OP is ready, then the copy will be also ready.  The
   copy does not belong to a packet.  */
static op_t
op_copy (op)
     op_t op;
{
  op_t result;

  if (op->scheduled_p)
    abort ();
  result = get_free_op ();
  result->insn = insn_copy (op->insn);
  result->it_num = op->it_num;
  result->true_target = op->true_target;
  result->false_target = op->false_target;
  result->packet = NULL;
  result->in_deps = result->out_deps = NULL;
  result->next_op = NULL;
  result->next_added_in_deps_op = NULL;
  result->initiator_op = op->initiator_op;
  result->scheduled_p = FALSE;
  result->ready_p = op->ready_p;
  result->prev_ready_op = NULL;
  if (!op->ready_p)
    result->next_ready_op = NULL;
  else
    {
      result->next_ready_op = ready_ops [op->it_num];
      if (result->next_ready_op != NULL)
	result->next_ready_op->prev_ready_op = result;
      ready_ops [op->it_num] = result;
    }
  result->can_be_converted_p = op->can_be_converted_p;
  result->target_fixed_p = op->target_fixed_p;
  result->use_def_p = op->use_def_p;
  result->true_jump_p = op->true_jump_p;
  result->processed_4_consistent_jumps_p = op->processed_4_consistent_jumps_p;
  return result;
}

/* The following function fixes that OP is ready.  */
static void
op_become_ready (op)
     op_t op;
{
  if (op->scheduled_p || op->ready_p)
    abort ();
  op->ready_p = TRUE;
  /* Add OP to ready list.  */
  op->prev_ready_op = NULL;
  op->next_ready_op = ready_ops [op->it_num];
  if (op->next_ready_op != NULL)
    op->next_ready_op->prev_ready_op = op;
  ready_ops [op->it_num] = op;
  if (min_ready_op_it_num > op->it_num)
    min_ready_op_it_num = op->it_num;
}

/* The following function fixes that OP is not ready (it should be
   ready before the function call).  */
static void
op_become_not_ready (op)
     op_t op;
{
  if (op->scheduled_p || !op->ready_p)
    abort ();
  op->ready_p = FALSE;
  /* Remove from ready list.  */
  if (op->prev_ready_op == NULL)
    ready_ops [op->it_num] = op->next_ready_op;
  else
    op->prev_ready_op->next_ready_op = op->next_ready_op;
  if (op->next_ready_op != NULL)
    op->next_ready_op->prev_ready_op = op->prev_ready_op;
  if (min_ready_op_it_num == op->it_num)
    for (min_ready_op_it_num = op->it_num;
	 min_ready_op_it_num < ready_ops_size - 1
	   && ready_ops [min_ready_op_it_num] == NULL;
	 min_ready_op_it_num++)
      ;
}

/* The following function fixes that ready OP has been scheduled.  */

static void
op_become_scheduled (op)
     op_t op;
{
  op_become_not_ready (op);
  op->scheduled_p = TRUE;
  scheduled_ops_number++;
}

/* The function initiates work with ops.  */

static void
start_ops ()
{
  scheduled_ops_number = 0;
  min_ready_op_it_num = 0;
  ready_ops_size = iteration_look_ahead * RCSP_UNROLL_FACTOR + 1;
  ready_ops = rcsp_malloc (ready_ops_size * sizeof (op_t));
  memset ((char *) ready_ops, 0, ready_ops_size * sizeof (op_t));
  all_ops = free_ops = NULL;
}

/* The function finishes all work with ops.  */

static void
finish_ops ()
{
  op_t curr_op, next_op;

  for (curr_op = all_ops; curr_op != NULL; curr_op = next_op)
    {
      next_op = curr_op->all_ops_chain;
      rcsp_free (curr_op);
    }
  if (ready_ops != NULL)
    rcsp_free (ready_ops);
}



/* The page contains functions for work with avail lists.  Frontier
   and formed packets have such lists.  The list refers for operations
   whose all op sources of input dependencies have started (but may be
   input data are not ready).  */

/* The following variable refers for unused elements of avail
   lists.  */
static avail_list_t free_avail_list_els;

/* The following variable refers for all allocated elements of avail
   lists.  */
static avail_list_t all_avail_list_els;

/* The following variables refer for array of pointers to available
   list elements and its length in the pointers.  This array is used to
   sort available lists.  */
static avail_list_t *avail_list_array;
static int avail_list_array_length;

/* The function frees avail_list element EL.  */
static void
avail_list_el_free (el)
     avail_list_t el;
{
  el->next_avail = free_avail_list_els;
  free_avail_list_els = el;
}

/* The following function returns free avail_list element with member
   values PACKET, OP, and AVAIL_P.  */
static avail_list_t
avail_list_el_create (packet, op, avail_p)
     packet_t packet;
     op_t op;
     int avail_p;
{
  avail_list_t result;

  if (free_avail_list_els)
    {
      result = free_avail_list_els;
      free_avail_list_els = free_avail_list_els->next_avail;
    }
  else
    {
      result = rcsp_malloc (sizeof (struct avail_list));
      result->all_avail_list_els_chain = all_avail_list_els;
      all_avail_list_els = result;
    }
  result->packet = packet;
  result->op = op;
  result->avail_p = avail_p;
  result->prev_avail = result->next_avail = NULL;
  return result;
}

/* The following function frees all elements of AVAIL_LIST.  */
static void
avail_list_free (avail_list)
     avail_list_t avail_list;
{
  avail_list_t curr_avail_list_el, next_avail_list_el;

  for (curr_avail_list_el = avail_list;
       curr_avail_list_el != NULL;
       curr_avail_list_el = next_avail_list_el)
    {
      next_avail_list_el = curr_avail_list_el->next_avail;
      avail_list_el_free (curr_avail_list_el);
    }
}

/* The function returns length of AVAIL_LIST.  */
static int
avail_list_length (avail_list)
     avail_list_t avail_list;
{
  avail_list_t curr;
  int length = 0;

  for (curr = avail_list; curr != NULL; curr = curr->next_avail)
    length++;
  return length;
}

/* The following function sorts AVAIL_LIST and returns the result.
   Comparison function (-1, 0, 1) is CMP_FUNC.  */
static avail_list_t
avail_list_sort (avail_list, cmp_func)
     avail_list_t avail_list;
     int (*cmp_func) (avail_list_t *, avail_list_t *);
{
  int length = avail_list_length (avail_list);
  int i;
  avail_list_t curr;
  int (*compar)(const void *, const void *)
    = (int (*)(const void *, const void *)) cmp_func;

  if (length <= 1)
    return avail_list;
  if (avail_list_array == NULL)
    {
      avail_list_array = rcsp_malloc (sizeof (avail_list_t) * length);
      avail_list_array_length = length;
    }
  else if (avail_list_array_length < length)
    {
      avail_list_array
	= rcsp_realloc (avail_list_array,
			sizeof (avail_list_t) * avail_list_array_length,
			sizeof (avail_list_t) * length);
      avail_list_array_length = length;
    }
  for (i = 0, curr = avail_list; curr != NULL; i++, curr = curr->next_avail)
    avail_list_array [i] = curr;
  if (i != length)
    abort ();
  qsort (avail_list_array, length, sizeof (avail_list_t), compar);
  for (i = 0; i < length - 1; i++)
    {
      avail_list_array [i]->next_avail = avail_list_array [i + 1];
      avail_list_array [i + 1]->prev_avail = avail_list_array [i];
    }
  avail_list_array [0]->prev_avail = NULL;
  avail_list_array [length - 1]->next_avail = NULL;
  return avail_list_array [0];
}

/* The function initiates work with avail lists.  */
static void
start_avail_lists ()
{
  all_avail_list_els = free_avail_list_els = NULL;
  avail_list_array = NULL;
}

/* The function finishes all work with avail lists.  */
static void
finish_avail_lists ()
{
  avail_list_t curr_el, next_el;

  for (curr_el = all_avail_list_els; curr_el != NULL; curr_el = next_el)
    {
      next_el = curr_el->all_avail_list_els_chain;
      rcsp_free (curr_el);
    }
  if (avail_list_array != NULL)
    rcsp_free (avail_list_array);
}



/* The page contains functions for work with packets.  */

/* The following variable refers for unused packets.  */
static packet_t free_packets;

/* The following variable refers for all allocated packets.  */
static packet_t all_packets;

/* The following variable is number of packets created and copied.  */
static int packets_number;

/* The following function frees PACKET.  */
static void
free_packet (packet)
     packet_t packet;
{
  packet->next_packet = free_packets;
  free_packets = packet;
}

/* The macro value is initial length of arrays `succ' and `pred' of
   packet.  */
#define INIT_SUCC_PRED_LENGTH 5

/* The following function returns free packet.  */
static packet_t
get_free_packet ()
{
  packet_t result;

  if (free_packets != NULL)
    {
      result = free_packets;
      free_packets = free_packets->next_packet;
    }
  else
    {
      result = rcsp_malloc (sizeof (struct packet));
      result->all_packets_chain = all_packets;
      all_packets = result;
      result->succ_length = INIT_SUCC_PRED_LENGTH;
      result->succ = rcsp_malloc (sizeof (packet_t) * INIT_SUCC_PRED_LENGTH);
      result->pred_length = INIT_SUCC_PRED_LENGTH;
      result->pred = rcsp_malloc (sizeof (packet_t) * INIT_SUCC_PRED_LENGTH);
      result->cpu_state = rcsp_malloc (cpu_state_length);
    }
  return result;
}

/* The following function checks that length of array `succ' (if
   SUCC_P) or `pred' of PACKET not less than LENGTH.  If it is not
   true, the array is expanded.  */
static void
expand_packet_succ_pred (packet, succ_p, length)
     packet_t packet;
     int succ_p;
     int length;
{
  int old_length;

  if (succ_p)
    {
      if (length > packet->succ_length)
	{
	  old_length = packet->succ_length;
	  packet->succ_length += packet->succ_length / 2 + 1;
          if (packet->succ_length < length)
            packet->succ_length = length;
          packet->succ
	    = rcsp_realloc (packet->succ, sizeof (packet_t) * old_length,
			    sizeof (packet_t) * packet->succ_length);
	}
    }
  else
    {
      if (length > packet->pred_length)
	{
	  old_length = packet->pred_length;
	  packet->pred_length += packet->pred_length / 2 + 1;
          if (packet->pred_length < length)
            packet->pred_length = length;
	  packet->pred
	    = rcsp_realloc (packet->pred, sizeof (packet_t) * old_length,
			    sizeof (packet_t) * packet->pred_length);
	}
    }
}

#ifdef USE_ACHIEVABLE_CACHE

/* The following function checks ACHIEVABLE_CACHE size and expands it
   if it is necessary.  */
static void
expand_achievable_cache ()
{
  int i, old_cache_size, cache_size;

  if (packets_number >= (1 << CFG_shift))
    {
      old_cache_size = (1 << CFG_shift) * (1 << CFG_shift);
      while (packets_number >= (1 << CFG_shift))
	CFG_shift++;
      if (CFG_shift <= MAX_CFG_SHIFT)
	{
	  cache_size = (1 << CFG_shift) * (1 << CFG_shift);
	  CFG_validation
	    = rcsp_realloc (CFG_validation, old_cache_size * sizeof (int),
			    cache_size * sizeof (int));
	  for (i = 0; i < cache_size; i++)
	    CFG_validation [i] = 0;
	  CFG_achievable
	    = rcsp_realloc (CFG_achievable, old_cache_size * sizeof (char),
			    cache_size * sizeof (char));
	}
    }
}

#endif

/* The following function creates packet with ISSUE_TIME and default
   member values.  */
static packet_t
packet_create (issue_time)
     int issue_time;
{
  packet_t p = get_free_packet ();
  
  p->number = ++packets_number;
  p->stop_packet_p = FALSE;
  p->exit_packet_p = FALSE;
  p->frontier_packet_p = FALSE;
  p->formed_packet_p = FALSE;
  p->main_exit_p = FALSE;
  *p->succ = NULL;
  *p->pred = NULL;
  p->visit_number1 = p->visit_number2
    = p->visit_number3 = p->visit_number4 = p->visit_number5 = -1;
  p->first_op = p->last_op = NULL;
  p->min_it_n = -1;
  p->copy = NULL;
  p->next_frontier_packet = NULL;
  p->avail_list = NULL;
  p->next_packet = NULL;
  p->issue_time = issue_time;
#ifdef USE_ACHIEVABLE_CACHE
  expand_achievable_cache ();
#endif
  return p;
}

/* The following function deletes packet.  The packet should not refer
   for a CFG packet or should be not referred from a CFG packet.  */
static void
packet_delete (packet)
     packet_t packet;
{
  op_t curr_op, next_op;

  if (*packet->succ != NULL || *packet->pred != NULL)
    abort ();
  avail_list_free (packet->avail_list);
  for (curr_op = packet->first_op; curr_op != NULL; curr_op = next_op)
    {
      next_op = curr_op->next_op;
      op_delete (curr_op);
    }
  free_packet (packet);
}

/* The following function appends OP to PACKET at the beginning of the
   operation list (if HEAD_P) or at the end of it.  */
static void
packet_append_op (packet, op, head_p)
     packet_t packet;
     op_t op;
     int head_p;
{
  op->packet = packet;
  if (packet->first_op == NULL)
    {
      packet->first_op = op;
      packet->last_op = op;
      op->next_op = NULL;
    }
  else if (head_p)
    {
      op->next_op = packet->first_op;
      packet->first_op = op;
    }
  else
    {
      op->next_op = NULL;
      packet->last_op->next_op = op;
      packet->last_op = op;
    }
}

/* The following function makes copy of interior packet and all its
   operations and clears all in/out dependencies.  */
static packet_t
packet_copy (packet)
     packet_t packet;
{
  packet_t p = get_free_packet ();
  op_t curr_op, copy;
  int min_it_n;

  if (packet->avail_list != NULL
      || packet->frontier_packet_p || packet->formed_packet_p
      || packet->stop_packet_p)
    /* This is not interior packet.  */
    abort ();
  p->number = ++packets_number;
  expand_packet_succ_pred (p, TRUE, packet->succ_length);
  expand_packet_succ_pred (p, FALSE, packet->pred_length);
  memcpy (p->succ, packet->succ, packet->succ_length * sizeof (packet_t));
  memcpy (p->pred, packet->pred, packet->pred_length * sizeof (packet_t));
#ifdef USE_ACHIEVABLE_CACHE
  CFG_tick++;
#endif
  p->visit_number1 = p->visit_number2 = p->visit_number3 = -1;
  p->first_op = p->last_op = NULL;
  min_it_n = -1;
  for (curr_op = packet->first_op;
       curr_op != NULL;
       curr_op = curr_op->next_op)
    {
      if (min_it_n < 0 || min_it_n > curr_op->it_num)
	min_it_n = curr_op->it_num;
      copy = op_copy (curr_op);
      packet_append_op (p, copy, FALSE);
    }
  p->stop_packet_p = FALSE;
  p->exit_packet_p = packet->exit_packet_p;
  p->main_exit_p = FALSE;
  p->min_it_n = min_it_n;
  p->copy = NULL;
  p->next_frontier_packet = NULL;
  p->frontier_packet_p = FALSE;
  p->formed_packet_p = FALSE;
  p->avail_list = NULL;
  p->next_packet = NULL;
  p->issue_time = packet->issue_time;
#ifdef USE_ACHIEVABLE_CACHE
  expand_achievable_cache ();
#endif
  return p;
}

/* The following function returns place of WHAT in array ARR.  If
   there is no such packet, the function return place of array end
   marker (containing NULL).  */
static packet_t *
packet_find_in_arr (arr, what)
     packet_t *arr;
     packet_t what;
{
  if (arr == NULL)
    abort ();
  for (; *arr != NULL; arr++)
    if (*arr == what)
      break;
  return arr;
}

/* The following function removes WHAT from array ARR (by shifting all
   following array elements).  If there is no such packet, the
   function does nothing.  */
static void
packet_del_from_arr (arr, what)
     packet_t *arr;
     packet_t what;
{
  packet_t *place;

  if (arr == NULL)
    abort ();
  if (what == NULL)
    return;
  place = packet_find_in_arr (arr, what);
  if (*place != NULL)
    {
#ifdef USE_ACHIEVABLE_CACHE
      CFG_tick++;
#endif
      for (; *place != NULL; place++)
	*place = place [1];
    }
}

/* The function adds SUCC_PRED at end of array `succ' (if SUCC_P) or
   `pred' of PACKET.  If there is already such packet in the array,
   the function does nothing.  */
static void
packet_add_to_arr (packet, succ_pred, succ_p)
     packet_t packet, succ_pred;
     int succ_p;
{
  packet_t *arr, *place;

  if (succ_pred == NULL)
    return;
  arr = succ_p ? packet->succ : packet->pred;
  place = packet_find_in_arr (arr, succ_pred);
  if (*place != NULL)
    return;
#ifdef USE_ACHIEVABLE_CACHE
  CFG_tick++;
#endif
  expand_packet_succ_pred (packet, succ_p, place - arr + 2);
  if (succ_p)
    {
      packet->succ [place - arr] = succ_pred;
      packet->succ [place - arr + 1] = NULL;
    }
  else
    {
      packet->pred [place - arr] = succ_pred;
      packet->pred [place - arr + 1] = NULL;
    }
}

/* The function connects TO and FROM modifying correspondingly their
   array `succ' and `pred'.  When POSSIBLE_BRANCH_P the function tries
   to set up {true,false}_target too.  In this case JUMP_P means that
   we should set up target corresponding to jump and corresponding to
   fall through otherwise.  The function is safe even if there is
   already such link between the packets.  */
static void
packet_connect (from, to, possible_branch_p, jump_p)
     packet_t from, to;
     int possible_branch_p, jump_p;
{
  op_t curr_op;
  packet_t *place;
  int new_p;

  if (from == NULL || to == NULL)
    abort ();
  place = packet_find_in_arr (from->succ, to);
  new_p = FALSE;
  if (*place == NULL)
    {
      packet_add_to_arr (from, to, TRUE);
      new_p = TRUE;
    }
  packet_add_to_arr (to, from, FALSE);
  if (new_p && possible_branch_p)
    {
      /* Set up targets for ops.  */
      for (curr_op = from->first_op;
	   curr_op != NULL;
	   curr_op = curr_op = curr_op->next_op)
	if (uncond_branch_p (curr_op->insn))
	  {
	    /* There is only one unconditional jump in a packet. ???
               What should be here if we permit several branches in
               packet.  */
	    curr_op->true_target = to;
	    break;
	  }
	else if (cond_branch_p (curr_op->insn))
	  {
	    if (curr_op->true_target == NULL
		&& ((curr_op->true_jump_p && jump_p)
		    || (!curr_op->true_jump_p && !jump_p)))
	      {
		curr_op->true_target = to;
		break;
	      }
	    else if (curr_op->false_target == NULL
		     && ((!curr_op->true_jump_p && jump_p)
			 || (curr_op->true_jump_p && !jump_p)))
	      {
		curr_op->false_target = to;
		break;
	      }
	  }
    }
}

/* The following function returns operation in PACKET which is branch
   (with jump if JUMP_P) onto TARGET, otherwise NULL.  */
static op_t
packet_find_branch (packet, target, jump_p)
     packet_t packet, target;
     int jump_p;
{
  op_t curr_op;

  if (target == NULL)
    abort ();
  for (curr_op = packet->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    if ((curr_op->false_target == target && (!jump_p || !curr_op->true_jump_p))
	|| (curr_op->true_target == target
	    && (!jump_p || curr_op->true_jump_p)))
      return curr_op;
  return NULL;
}

/* The function changes possible link FROM->OLD_TO onto FROM->NEW_TO.
   The function also can be used for repair inconsistent values `succ'
   and `pred' for copied packets.  If NEW_TO is NULL link (FROM,
   OLD_TO) is simply removed.  */
static void
packet_change_dest (from, old_to, new_to)
     packet_t from, old_to, new_to;
{
  packet_t *place;
  op_t branch;

  if (old_to == new_to || old_to == NULL)
    abort ();
  branch = packet_find_branch (from, old_to, FALSE);
  if (branch != NULL)
    {
      if (branch->true_target == old_to)
	branch->true_target = new_to;
      else if (branch->false_target == old_to)
	branch->false_target = new_to;
      else
	abort ();
    }
  if (new_to == NULL || *packet_find_in_arr (from->succ, new_to) != NULL)
    packet_del_from_arr (from->succ, old_to);
  else
    {
      place = packet_find_in_arr (from->succ, old_to);
      if (*place == NULL)
	packet_add_to_arr (from, new_to, TRUE);
      else
	{
	  *place = new_to;
#ifdef USE_ACHIEVABLE_CACHE
	  CFG_tick++;
#endif
	}
    }
  packet_del_from_arr (old_to->pred, from);
  if (new_to != NULL)
    packet_add_to_arr (new_to, from, FALSE);
}

/* The function changes possible link OLD_FROM->TO onto NEW_FROM->TO.
   The function also can be used for repair inconsistent values `succ'
   and `pred' for copied packets.  If NEW_ROM is NULL link (TO,
   OLD_FROM) is simply removed.  */
static void
packet_change_src (to, old_from, new_from)
     packet_t to, old_from, new_from;
{
  packet_t *place;
  op_t branch;

  if (old_from == new_from || old_from == NULL)
    abort ();
  branch = packet_find_branch (old_from, to, FALSE);
  if (branch != NULL)
    {
      if (branch->true_target == to)
	branch->true_target = NULL;
      else if (branch->false_target == to)
	branch->false_target = NULL;
      else
	abort ();
    }
  if (new_from == NULL || *packet_find_in_arr (to->pred, new_from) != NULL)
    packet_del_from_arr (to->pred, old_from);
  else
    {
      place = packet_find_in_arr (to->pred, old_from);
      if (*place == NULL)
	packet_add_to_arr (to, new_from, FALSE);
      else
	{
	  *place = new_from;
#ifdef USE_ACHIEVABLE_CACHE
	  CFG_tick++;
#endif
	}
    }
  packet_del_from_arr (old_from->succ, to);
  if (new_from != NULL)
    packet_add_to_arr (new_from, to, TRUE);
}

/* The function initiates work with packets.  */
static void
start_packets ()
{
#ifdef USE_ACHIEVABLE_CACHE
  int i, cache_size;
#endif

  all_packets = free_packets = NULL;
  packets_number = 0;
#ifdef USE_ACHIEVABLE_CACHE
  CFG_tick = 0;
  CFG_shift = 8;
  cache_size = (1 << CFG_shift) * (1 << CFG_shift);
  CFG_validation = rcsp_malloc (cache_size * sizeof (int));
  for (i = 0; i < cache_size; i++)
    CFG_validation [i] = 0;
  CFG_achievable = rcsp_malloc (cache_size * sizeof (char));
#endif
}

/* The function finishes all work with packets.  */
static void
finish_packets ()
{
  packet_t curr_packet, next_packet;

  for (curr_packet = all_packets;
       curr_packet != NULL;
       curr_packet = next_packet)
    {
      next_packet = curr_packet->all_packets_chain;
      rcsp_free (curr_packet->succ);
      rcsp_free (curr_packet->pred);
      rcsp_free (curr_packet);
    }
#ifdef USE_ACHIEVABLE_CACHE
  rcsp_free (CFG_achievable);
  rcsp_free (CFG_validation);
#endif
}



/* The page contains functions for work with dependencies.  */

/* The following variable refers for unused dependencies.  */
static dep_t unused_deps;

/* The following variable refers for all allocated dependencies.  */
static dep_t all_deps;

/* The following table is used to find dependency by pair (from, loc)
   quickly.  */
static htab_t deps_table;

/* The following variable values are dep array and its length.  The
   array is used to sort list of input deps of operation.  */
static dep_t *in_deps_array;
static int in_deps_array_length;

/* The following variables forms list of ops to which new dependencies
   have been added.  */
static op_t first_added_in_deps_op, last_added_in_deps_op;

/* The following function frees dependency DEP.  */
static void
free_dep (dep)
     dep_t dep;
{
  dep->next_out = unused_deps;
  unused_deps = dep;
}

/* The following function returns free dep.  */
static dep_t
get_free_dep ()
{
  dep_t result;

  if (unused_deps != NULL)
    {
      result = unused_deps;
      unused_deps = unused_deps->next_out;
    }
  else
    {
      result = rcsp_malloc (sizeof (struct dep));
      result->all_deps_chain = all_deps;
      all_deps = result;
    }
  return result;
}

/* The following function deletes dependency DEP.  */
static void
dep_delete (dep)
     dep_t dep;
{
  dep_t prev = dep->prev_in;
  dep_t next = dep->next_in;
  void **slot;

  /* Remove dep from double link chain of input dependencies for
     dep->from.  */
  if (prev == NULL)
    dep->to->in_deps = next;
  else
    prev->next_in = next;
  if (next != NULL)
    next->prev_in = prev;
  /* The same for output dependencies of dep->to.  */
  prev = dep->prev_out;
  next = dep->next_out;
  if (prev == NULL)
    dep->from->out_deps = next;
  else
    prev->next_out = next;
  if (next != NULL)
    next->prev_out = prev;
  /* Remove dep from loc dependencies.  */
  prev = dep->prev_out_loc_dep;
  next = dep->next_out_loc_dep;
  if (prev != NULL)
    prev->next_out_loc_dep = next;
  if (next != NULL)
    next->prev_out_loc_dep = prev;
  slot = htab_find_slot (deps_table, dep, FALSE);
  if (slot == NULL)
    abort ();
  else if (*(dep_t *) slot == dep)
    {
      if (prev != NULL)
	abort ();
      if (next == NULL)
	htab_clear_slot (deps_table, slot);
      else
	*(dep_t *) slot = next;
    }
  free_dep (dep);
}

/* The following function returns the first output dependency (marked
   with LOC) from OP.  */
static dep_t
dep_out_find (op, loc)
     op_t op;
     location_t loc;
{
#if 0
  dep_t curr_dep;

  for (curr_dep = op->out_deps;
       curr_dep != NULL;
       curr_dep = curr_dep->next_out)
    if (loc_eq_p (curr_dep->to_loc, loc))
      return curr_dep;
  return NULL;
#else
  struct dep dep;

  dep.to_loc = loc;
  dep.from = op;
  return (dep_t) htab_find (deps_table, &dep);
#endif
}

/* The following function creates dependency with given
   characteristics.  */
static void
dep_create (from, to, loc, dep_mode, latency)
     op_t from, to;
     location_t loc;
     dep_mode_t dep_mode;
     int latency;
{
  dep_t first_loc_dep;
  void **slot;
  dep_t curr_loc_dep;
  dep_t dep = get_free_dep ();
  
  dep->dep_mode = dep_mode;
  dep->latency = latency;
  dep->from = from;
  dep->to_loc = loc;
  dep->to = to;
  /* Insert dep in loc_dep list.  */
  slot = htab_find_slot (deps_table, dep, TRUE);
  first_loc_dep = *(dep_t *) slot;
  if (first_loc_dep == NULL)
    {
      dep->prev_out_loc_dep = dep->next_out_loc_dep = NULL;
      *(dep_t *) slot = dep;
    }
  else
    {
      for (curr_loc_dep = first_loc_dep;
	   curr_loc_dep != NULL;
	   curr_loc_dep = curr_loc_dep->next_out_loc_dep)
	if (curr_loc_dep->to == to && curr_loc_dep->dep_mode == dep_mode
	    && curr_loc_dep->latency == latency)
	  break;
      if (curr_loc_dep != NULL)
	{
	  /* There is already such dep.  */
	  if (!loc_eq_p (curr_loc_dep->to_loc, loc)
	      || curr_loc_dep->from != from)
	    abort ();
	  free_dep (dep);
	  return;
	}
      dep->prev_out_loc_dep = first_loc_dep;
      dep->next_out_loc_dep = first_loc_dep->next_out_loc_dep;
      if (first_loc_dep->next_out_loc_dep != NULL)
        first_loc_dep->next_out_loc_dep->prev_out_loc_dep = dep;
      first_loc_dep->next_out_loc_dep = dep;
    }
  if (to->next_added_in_deps_op == NULL && to != last_added_in_deps_op)
    {
      /* TO is not in the list.  Insert TO into the list of operations
	 to which input dependencies have been added.  */
      if (first_added_in_deps_op == NULL)
	first_added_in_deps_op = to;
      else
	last_added_in_deps_op->next_added_in_deps_op = to;
      last_added_in_deps_op = to;
    }
  /* Insert dep in output dep list.  */
  dep->prev_out = NULL;
  if (from->out_deps == NULL)
    dep->next_out = NULL;
  else
    {
      dep->next_out = from->out_deps;
      from->out_deps->prev_out = dep;
    }
  from->out_deps = dep;
  /* Insert dep in input dep list.  */
  dep->prev_in = NULL;
  if (to->in_deps == NULL)
    dep->next_in = NULL;
  else
    {
      dep->next_in = to->in_deps;
      to->in_deps->prev_in = dep;
    }
  to->in_deps = dep;
}

/* The following function is used as comparison function to sort input
   dependencies list of operation.  */
static int
in_deps_cmp (dep1, dep2)
     const void *dep1;
     const void *dep2;
{
  dep_t d1 = *(dep_t *) dep1;
  dep_t d2 = *(dep_t *) dep2;
  
  if (d1->to_loc < d2->to_loc)
    return -1;
  else if (d1->to_loc > d2->to_loc)
    return 1;
  else
    return d1->from->packet->number - d2->from->packet->number;
}

/* The following function sorts input dependencies list of OP.  The
   goal of sorting is to place input dependencies of OP with the same
   to_loc nearby.  */
static void
sort_op_in_deps (op)
     op_t op;
{
  int length = 0;
  int i;
  dep_t curr_dep;

  for (curr_dep = op->in_deps;
       curr_dep != NULL;
       curr_dep = curr_dep->next_in)
    length++;
  if (length <= 1)
    return;
  if (in_deps_array == NULL)
    {
      in_deps_array = rcsp_malloc (sizeof (dep_t) * length);
      in_deps_array_length = length;
    }
  else if (length > in_deps_array_length)
    {
      in_deps_array = rcsp_realloc (in_deps_array,
				    sizeof (dep_t) * in_deps_array_length,
				    sizeof (dep_t) * length);
      in_deps_array_length = length;
    }
  for (i = 0, curr_dep = op->in_deps;
       curr_dep != NULL;
       curr_dep = curr_dep->next_in, i++)
    in_deps_array [i] = curr_dep;
  qsort (in_deps_array, length, sizeof (dep_t), in_deps_cmp);
  for (i = 1; i < length; i++)
    {
      in_deps_array [i - 1]->next_in = in_deps_array [i];
      in_deps_array [i]->prev_in = in_deps_array [i - 1];
    }
  in_deps_array [0]->prev_in = in_deps_array [length - 1]->next_in = NULL;
  op->in_deps = in_deps_array [0];
}

/* The following function sorts input dependencies for all OP to which
   new input dependencies added.  After this such ops are forgot.  */
static void
dep_sort_added_in_deps_op ()
{
  op_t curr_op, next_op;

  for (curr_op = first_added_in_deps_op; curr_op != NULL; curr_op = next_op)
    {
      sort_op_in_deps (curr_op);
      next_op = curr_op->next_added_in_deps_op;
      curr_op->next_added_in_deps_op = NULL;
    }
  first_added_in_deps_op = last_added_in_deps_op = NULL;
}

/* Hash function for DEP.  */
static unsigned
dep_hash (dep)
     const void *dep;
{
  dep_t d = (dep_t) dep;
  return loc_hash (d->to_loc) + ((unsigned) d->from << 8);
}

/* The following function returns TRUE if DEP1 has the same key as
   DEP2.  */
static int
dep_eq (dep1, dep2)
     const void *dep1, *dep2;
{
  dep_t d1 = (dep_t) dep1;
  dep_t d2 = (dep_t) dep2;

  return d1->from == d2->from && loc_eq_p (d1->to_loc, d2->to_loc);
}


/* The function initiates work with deps.  */
static void
start_deps ()
{
  all_deps = unused_deps = NULL;
  deps_table = htab_create (1000, dep_hash, dep_eq, NULL);
  in_deps_array = NULL;
  first_added_in_deps_op = last_added_in_deps_op = NULL;
}

/* The function finishes all work with deps.  */
static void
finish_deps ()
{
  dep_t curr_dep, next_dep;

  for (curr_dep = all_deps;
       curr_dep != NULL;
       curr_dep = next_dep)
    {
      next_dep = curr_dep->all_deps_chain;
      rcsp_free (curr_dep);
    }
  htab_delete (deps_table);
  if (in_deps_array != NULL)
    rcsp_free (in_deps_array);
}



/* The page contains functions for work with software pipeliner
   generalized states.  When we found the same state that means that
   we found the same repetitive pattern in software pipelining, and
   instead of packet being formed we use already formed packet
   corresponding to the found software pipeliner generalized state.  */

/* Hash function for SCHEDULE_STATE.  */
static unsigned
schedule_state_hash (packet)
     const void *packet;
{
  packet_t p = (packet_t) packet;
  unsigned result = 0;
  avail_list_t avail;
  op_t curr_op;
  int i;

  for (i = 0; i < cpu_state_length; i++)
    result
      += ((unsigned char *) p->cpu_state) [i] << (i % sizeof (unsigned) * 8);
  for (curr_op = p->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    if (curr_op->initiator_op != NULL)
      result += insn_hash (curr_op->initiator_op->insn);
    else
      {
	result += insn_hash (curr_op->insn);
	if (cond_branch_p (curr_op->insn))
	  result += !curr_op->true_jump_p;
      }
  for (avail = p->avail_list; avail != NULL; avail = avail->next_avail)
    {
      /* To speed it up, Take only available operations into account.  */
      if (!avail->avail_p)
	break;
      /* Ignore iteration number.  */
      if (avail->op->initiator_op != NULL)
	result += (insn_hash (avail->op->initiator_op->insn) + avail->delay);
      else
	result += (insn_hash (avail->op->insn) + avail->delay);
    }
  return result;
}

/* The following function returns TRUE if OP1 is equal to OP2 with the
   point of view of scheduled state equality. *UNDEFINED_DIFF_P and
   *DIFF are in/out parameters used to comparison of the operation
   iteration numbers.  */
static int
schedule_state_op_eq (op1, op2, undefined_diff_p, diff)
     op_t op1, op2;
     int *undefined_diff_p, *diff;
{
  if (op1->initiator_op != op2->initiator_op)
    return FALSE;
  else if (op1->initiator_op != NULL)
    {
      if (!insn_eq_p (op1->initiator_op->insn, op2->initiator_op->insn))
	return FALSE;
      else if (op1->initiator_op->packet->exit_packet_p)
	{
	  if (!op2->initiator_op->packet->exit_packet_p)
	    abort ();
	  /* We require equality for iteration numbers.  */
	  if (op1->initiator_op->it_num != op2->initiator_op->it_num)
	    return FALSE;
	}
      else if (*undefined_diff_p)
	{
	  *diff = (op1->initiator_op->it_num - op2->initiator_op->it_num);
	  *undefined_diff_p = FALSE;
	}
      else if (op1->initiator_op->it_num - op2->initiator_op->it_num != *diff)
	return FALSE;
    }
  else if (!insn_eq_p (op1->insn, op2->insn)
	   || (!insn_dummy_p (op1->insn) && cond_branch_p (op1->insn)
	       && !op1->true_jump_p != !op2->true_jump_p))
    return FALSE;
  else if (op1->packet->exit_packet_p && op2->packet->exit_packet_p)
    {
      /* We require equality for iteration numbers.  */
      if (op1->it_num != op2->it_num)
	return FALSE;
    }
  else if (op1->packet->exit_packet_p || op2->packet->exit_packet_p)
    /* Insns from loop and exit branch are never equal.  */
    return FALSE;
  else if (*undefined_diff_p)
    {
      *diff = op1->it_num - op2->it_num;
      *undefined_diff_p = FALSE;
    }
  else if (op1->it_num - op2->it_num != *diff)
    return FALSE;
  return TRUE;
}

/* The function returns TRUE if SCHEDULE_STATE1 is equal to
   SCHEDULE_STATE2 with the point of view of the pipeliner.  It means
   that they have the same CPU_STATE and sets "available" with
   possible the same differences in the packet operations.  */
static int
schedule_state_eq (packet1, packet2)
     const void *packet1, *packet2;
{
  packet_t p1 = (packet_t) packet1;
  packet_t p2 = (packet_t) packet2;
  avail_list_t avail1, avail2;
  op_t curr_op1, curr_op2;
  int diff = 0;
  int undefined_diff_p;

  if (memcmp (p1->cpu_state, p2->cpu_state, cpu_state_length) != 0)
    return FALSE;
  undefined_diff_p = TRUE;
  for (curr_op1 = p1->first_op, curr_op2 = p2->first_op;
       curr_op1 != NULL && curr_op2 != NULL;
       curr_op1 = curr_op1->next_op, curr_op2 = curr_op2->next_op)
    if (!schedule_state_op_eq (curr_op1, curr_op2, &undefined_diff_p, &diff))
      return FALSE;
  if (curr_op1 != curr_op2)
    return FALSE;
  for (avail1 = p1->avail_list, avail2 = p2->avail_list;
       avail1 != NULL && avail2 != NULL;
       avail1 = avail1->next_avail, avail2 = avail2->next_avail)
    {
      curr_op1 = avail1->op;
      curr_op2 = avail2->op;
      if (curr_op1 == NULL || curr_op2 == NULL)
        abort ();
      if (avail1->avail_p != avail2->avail_p
	  || !schedule_state_op_eq (curr_op1, curr_op2,
				    &undefined_diff_p, &diff))
	return FALSE;
      /* Take delays into account for non-exit packets.  */ 
      if ((!curr_op1->packet->exit_packet_p
	   || !curr_op2->packet->exit_packet_p)
	  && avail1->avail_p && avail1->delay != avail2->delay)
	return FALSE;
    }
  if (avail1 != avail2)
    return FALSE;
  if (p1->exit_packet_p && p2->exit_packet_p)
    {
      /* Available lists are not sufficient for exit packets.  We
         could check all achievable packets may be ignoring iteration
         numbers. ??? But now we do not do it.  The question is it
         worth to do it, especially if we implement temporary register
         renaming.  Because in this case we can not ignore iteration
         numbers.  */
      return FALSE;
    }
  return TRUE;
}

/* The following function creates empty scheduler states set.  */
static packet_set_t
packet_set_create ()
{
  return htab_create (1000, schedule_state_hash, schedule_state_eq, NULL);
}

/* The following function deletes PACKET_SET.  */
static void
packet_set_delete (packet_set)
     packet_set_t packet_set;
{
  htab_delete (packet_set);
}

/* The function inserts PACKET into PACKET_SET if it is not there.
   The function returns the packet which is in PACKET_SET.  */
static packet_t
packet_set_insert (packet_set, packet)
     packet_set_t packet_set;
     packet_t packet;
{
  void **slot;

  slot = htab_find_slot (packet_set, packet, TRUE);
  if (*slot == NULL)
    *slot = (void *) packet;
  return (packet_t) *slot;
}



/* The page contains functions to start and finish work with all
   abstract data needed for pipelining.  */

/* Start work with abstract data needed for pipelining.  */
static void
start_pipeline_data ()
{
  cpu_state_length = cpu_state_size ();
  temp_cpu_state = rcsp_malloc (cpu_state_length);
  start_ops ();
  start_avail_lists ();
  start_packets ();
  start_deps ();
  start_frontier_packet_list ();
}

/* Finish work with abstract data needed for pipelining.  */
static void
finish_pipeline_data ()
{
  finish_frontier_packet_list ();
  finish_deps ();
  finish_packets ();
  finish_avail_lists ();
  finish_ops ();
  rcsp_free (temp_cpu_state);
}



/* The page contains functions for building original DDG.  These
   functions work with locrefs used/defined by operations.  We keep
   set of the last active locrefs used/defined by operations.  */

/* The following set contains recent locrefs defined or used in ops.
   The set is used for faster building original DDG.  The key is
   location and operation.  */
static locrefset_t last_op_refs;

/* The following variable refers for unused locrefs.  */
static locref_t free_locrefs;

/* The following function frees LOCREF.  */
static void
free_locref (locref)
     locref_t locref;
{
  locref->next_locref = free_locrefs;
  free_locrefs = locref;
}

/* The following function returns free locref.  */
static locref_t
get_free_locref ()
{
  locref_t result;

  if (free_locrefs)
    {
      result = free_locrefs;
      free_locrefs = free_locrefs->next_locref;
    }
  else
    result = rcsp_malloc (sizeof (struct locref));
  return result;
}

/* The following function returns hash key of LOCREF.  */
static unsigned
locref_hash (locref)
     const void *locref;
{
  locref_t l = (locref_t) locref;

  return loc_hash (l->loc) + ((unsigned) l->op << 8);
}

/* The following function returns TRUE if locref1 has the same key as
   LOCRE2.  */
static int
locref_eq (locref1, locref2)
     const void *locref1, *locref2;
{
  locref_t l1 = (locref_t) locref1;
  locref_t l2 = (locref_t) locref2;

  return l1->op == l2->op && loc_eq_p (l1->loc, l2->loc);
}

/* The following function returns list of active locrefs with LOC
   used/defined in OP.  */
static locref_t
locref_find (loc, op)
     location_t loc;
     op_t op;
{
  struct locref locref;

  locref.loc = loc;
  locref.op = op;
  return (locref_t) htab_find (last_op_refs, &locref);
}

/* The following function adds locref with LOC used (if USAGE_P) or
   defined in OP to the list of active locrefs.  */
static void
locref_add (op, loc, usage_p, write_barrier_p)
     op_t op;
     location_t loc;
     int usage_p;
     int write_barrier_p;
{
  void **slot;
  locref_t last_locref;
  locref_t l = get_free_locref ();
  
  l->op = op;
  l->loc = loc;
  l->usage_p = usage_p;
  l->write_barrier_p = write_barrier_p;
  l->next_locref = NULL;
  slot = htab_find_slot (last_op_refs, l, TRUE);
  if (*slot == NULL)
    *slot = l;
  else
    {
      for (last_locref = (locref_t) *slot;
	   last_locref->next_locref != NULL; 
	   last_locref = last_locref->next_locref)
	;
      last_locref->next_locref = l;
    }
}

/* The following function removes all active locrefs with LOC
   used/defined in OP.  */
static void
locref_remove_all (loc, op)
     location_t loc;
     op_t op;
{
  struct locref locref;
  locref_t curr_locref, next_locref;
  void **slot;
  
  locref.loc = loc;
  locref.op = op;
  slot = htab_find_slot (last_op_refs, &locref, FALSE);
  if (slot != NULL)
    {
      for (curr_locref = (locref_t) *slot;
	   curr_locref != NULL; 
	   curr_locref = next_locref)
	{
	  next_locref = curr_locref->next_locref;
	  free_locref (curr_locref);
	}
      htab_clear_slot (last_op_refs, slot);
    }
}

/* The function initiates work with locrefs.  */
static void
start_locrefs ()
{
  free_locrefs = NULL;
  last_op_refs = htab_create (1000, locref_hash, locref_eq, NULL);
}

/* The following function is used to remove locrefs staring with
   LOCREF which are in the hash table.  */
static int
free_htab_locrefs (locref_ref, param)
     void **locref_ref;
     void *param ATTRIBUTE_UNUSED;
{
  locref_t curr_locref, next_locref;

  for (curr_locref = (locref_t) *locref_ref;
       curr_locref != NULL; 
       curr_locref = next_locref)
    {
      next_locref = curr_locref->next_locref;
      free_locref (curr_locref);
    }
  return TRUE;
}

/* The function finishes all work with locrefs.  */
static void
finish_locrefs ()
{
  locref_t curr_locref, next_locref;

  htab_traverse (last_op_refs, free_htab_locrefs, NULL);
  for (curr_locref = free_locrefs;
       curr_locref != NULL; 
       curr_locref = next_locref)
    {
      next_locref = curr_locref->next_locref;
      rcsp_free (curr_locref);
    }
  htab_delete (last_op_refs);
}



/* The page contains functions of initial building DDG.  */

/* The following creates dummy packet (see comments for
   `top_packet').  */
static packet_t
create_dummy_packet ()
{
  op_t dummy_op;
  packet_t dummy_packet;
  
  dummy_packet = packet_create (0);
  dummy_op = op_create (NULL, 0);
  packet_append_op (dummy_packet, dummy_op, TRUE);
  return dummy_packet;
}

/* The following function adds reference for location LOC used (if
   USAGE_P) or defined in operation OP.  If LOC is defined then all
   locations in LOC defined or used in OP are removed.  The function
   also sets up flag `use_def_p' if it is necessary.  */
static void
add_locref (op, loc, usage_p, write_kind)
     op_t op;
     location_t loc;
     int usage_p;
     write_kind_t write_kind;
{
  location_t curr_loc;
  int i;

  if (!usage_p)
    {
      if (write_kind != WRITE_BARRIER && locref_find (loc, op) != NULL)
	op->use_def_p = TRUE;
      if (write_kind == USUAL_WRITE)
	locref_remove_all (loc, op);
      for (i = 0; ; i++)
	{
	  curr_loc = loc_sub (loc, i == 0);
	  if (curr_loc == NULL)
	    break;
	  if (write_kind != WRITE_BARRIER
	      && locref_find (curr_loc, op) != NULL)
	    op->use_def_p = TRUE;
	  if (write_kind == USUAL_WRITE)
	    locref_remove_all (curr_loc, op);
	}
    }
  locref_add (op, loc, usage_p, write_kind == WRITE_BARRIER);
}

/* The following function processes location LOC (which is a super or
   sub location of OP_LOC) to find and to add (if it is found) new
   dependency from PRED_OP to OP on OP_LOC used (if USAGE_P) or
   defined in OP.  */
static int
process_locrefs (pred_op, loc, usage_p, write_barrier_p, op, op_loc)
     op_t pred_op;
     location_t loc;
     int usage_p, write_barrier_p;
     op_t op;
     location_t op_loc;
{
  int cont_p = TRUE;
  locref_t locref, curr_locref;

  locref = locref_find (loc, pred_op);
  for (curr_locref = locref;
       curr_locref != NULL;
       curr_locref = curr_locref->next_locref)
    if (curr_locref->op != op)
      {
	if (!usage_p)
	  {
	    if (!curr_locref->usage_p)
	      {
		/* Add "def-def" dep.  */
		dep_create (curr_locref->op, op, op_loc, DEF_DEF,
			    insn_dep_time (curr_locref->op->insn, op->insn,
					   DEF_DEF));
		cont_p = FALSE;
	      }
	    else if (!write_barrier_p)
	      /* Add "use-def" dep.  */
	      dep_create (curr_locref->op, op, op_loc, USE_DEF,
			  insn_dep_time (curr_locref->op->insn, op->insn,
					 USE_DEF));
	  }
	else if (!curr_locref->usage_p && !curr_locref->write_barrier_p)
	  {
	    /* Add "def-use" dep.  */
	    dep_create (curr_locref->op, op, op_loc, DEF_USE,
			insn_dep_time (curr_locref->op->insn, op->insn,
				       DEF_USE));
	    cont_p = FALSE;
	  }
      }
  return cont_p;
}

/* This recursive function processes pred packets starting with PACKET
   to add deps on LOC used (if USAGE_P) or defined in OP from another
   op in PACKET.  */
static void
process_packet_op_locs (packet, op, loc, usage_p, write_kind)
     packet_t packet;
     op_t op;
     location_t loc;
     int usage_p;
     write_kind_t write_kind;
{
  int pred_number;
  int i;
  int cont_p;
  location_t curr_loc;

  if (packet->visit_number1 == last_packet_visit_number1)
    return;
  packet->visit_number1 = last_packet_visit_number1;
  if (packet == top_packet)
    {
      /* Include dummy definition of loc in the top packet and add
	 dependency arc.  We define dummy definition even for
	 definition too.  It is needed for our algorithm of making
	 interblock operation moving.  */
      add_locref (top_packet->first_op, loc, FALSE, USUAL_WRITE);
      dep_create (top_packet->first_op, op, loc,
		  (usage_p ? DEF_USE : DEF_DEF), 0);
      return;
    }

  cont_p = process_locrefs (packet->first_op, loc,
			    usage_p, write_kind == WRITE_BARRIER, op, loc);
  if (cont_p)
    for (i = 0; ; i++)
      {
	curr_loc = loc_sub (loc, i == 0);
	if (curr_loc == NULL)
	  break;
	process_locrefs (packet->first_op, curr_loc,
			 usage_p, write_kind == WRITE_BARRIER, op, loc);
      }

  if (cont_p)
    for (pred_number = 0;
	 packet->pred [pred_number] != NULL;
	 pred_number++)
      process_packet_op_locs (packet->pred [pred_number], op, loc,
			      usage_p, write_kind);
}

/* The following function processes operation OP for building DDG.  */
static void
process_op_locs (op)
     op_t op;
{
  int i, usage_p;
  write_kind_t write_kind;
  location_t loc;

  for (i = 0; ; i++)
    {
      loc = getloc (op->insn, i, &usage_p, &write_kind);
      if (loc == NULL)
	break;
      last_packet_visit_number1++;
      process_packet_op_locs (op->packet, op, loc, usage_p, write_kind);
      add_locref (op, loc, usage_p, write_kind);
    }
}

/* This recursive function processes (depth-first CFG traversing)
   PACKET and all its successors to insert them into the packet list
   whose the first element is FIRST_PACKET.  The function returns the
   first element of the list after processing all the packets.  */
static packet_t
link_packets (packet, first_packet)
     packet_t packet, first_packet;
{
  int succ_number;

  if (packet->visit_number1 == last_packet_visit_number1)
    return first_packet;
  packet->visit_number1 = last_packet_visit_number1;
  for (succ_number = 0;
       packet->succ [succ_number] != NULL;
       succ_number++)
    first_packet = link_packets (packet->succ [succ_number], first_packet);
  packet->next_packet = first_packet;
  first_packet = packet;
  return first_packet;
}

/* The following function builds DDG for CFG with root START.  */
static void
build_deps (start)
      packet_t start;
{
  packet_t curr_packet, next_packet;

  start_locrefs ();
  top_packet = create_dummy_packet ();
  packet_connect (top_packet, start, FALSE, FALSE);
  last_packet_visit_number1++;
  link_packets (start, NULL);
  for (curr_packet = start;
       curr_packet != NULL;
       curr_packet = next_packet)
    {
      next_packet = curr_packet->next_packet;
      curr_packet->next_packet = NULL;
      if (curr_packet->first_op != NULL
	  && curr_packet->first_op->next_op != NULL)
	/* At most one insn should be in the packet now.  */
	abort ();
      if (curr_packet->first_op != NULL
	  && !insn_dummy_p (curr_packet->first_op->insn))
	process_op_locs (curr_packet->first_op);
    }
  finish_locrefs ();
  dep_sort_added_in_deps_op ();
}



/* This page contains functions for work with frontier packet list.  */

/* The following variable refers for the frontier packet list head.
   It should be only read from outside the page.  */
static packet_t first_frontier_packet;

/* The following variables refer for array of pointers to packets and
   its length in the pointers.  This array is used to sort the frontier
   packet list.  */
static packet_t *frontier_packet_array;
static int frontier_packet_array_length;

/* The following function returns the frontier packet list head and
   removes it from the list.  */
static packet_t
frontier_packet_list_head_remove ()
{
  packet_t result = first_frontier_packet;
  
  if (first_frontier_packet == NULL)
    abort ();
  first_frontier_packet = first_frontier_packet->next_frontier_packet;
  if (!result->frontier_packet_p)
    abort ();
  result->frontier_packet_p = FALSE;
  return result;
}

/* The following function returns the frontier packet list length.  */
static int
frontier_packet_list_length ()
{
  packet_t curr_packet;
  int length;

  for (length = 0, curr_packet = first_frontier_packet;
       curr_packet != NULL;
       curr_packet = curr_packet->next_frontier_packet)
    length++;
  return length;
}

/* The following function adds PACKET to list of frontier packets.  */
static void
add_frontier_packet (packet)
     packet_t packet;
{
  if (packet->frontier_packet_p || packet->formed_packet_p)
    abort ();
  packet->frontier_packet_p = TRUE;
  packet->next_frontier_packet = first_frontier_packet;
  first_frontier_packet = packet;
}

/* The following function is used as comparison function to sort
   frontier packet list to choose the first frontier packet for
   scheduling.  The first we schedule packets whose successors are not
   point of CFG merge.  Between such packets we choose packet whose
   minimal op iteration number is minimal.  Then we choose the first
   created frontier packet.  Packets with the same merge point will be
   adjacent.  Remember that frontier packets has only one
   successor.  */
static int
frontier_packet_sort_cmp (packet1, packet2)
     const void *packet1;
     const void *packet2;
{
  packet_t p1 = *(packet_t *) packet1;
  packet_t p2 = *(packet_t *) packet2;
  
  if (p1->succ [0]->pred [1] == NULL)
    {
      if (p2->succ [0]->pred [1] != NULL)
	return -1; /* choose packet whose successor is not CFG merge point.  */
      else if (p1->min_it_n != p2->min_it_n)
	return p1->min_it_n - p2->min_it_n; /* chose the packet with
					       minimal iteration number.  */
      else
	return p1->number - p2->number; /* chose the first created packet */
    }
  else if (p2->succ [0]->pred [1] == NULL)
    return 1; /* choose packet whose successor is not CFG merge point.  */
  else
    /* Choose packet according to the merge point.  */
    return p1->succ [0]->number - p2->succ [0]->number;
}

/* The following function sorts the frontier packet list and merges
   frontier packets with the same successor which is CFG merge point
   (in other words which has more one predecessor).  FILE and PREFIX
   are used for debugging output.  ??? Merging is necessary if we have
   constraint mentioned in get_best_candidate_avail_op.  Otherwise
   merging probably results in more compact and less effective
   pipelined loop.  Investigate this.  */
static void
frontier_packet_list_sort_and_merge (f, prefix)
     FILE *f;
     const char *prefix;
{
  int i;
  packet_t curr_packet, pred_packet;
  int length = frontier_packet_list_length ();
  int decr_time;

  if (first_frontier_packet == NULL)
    abort ();
  if (frontier_packet_array == NULL)
    {
      frontier_packet_array = rcsp_malloc (sizeof (packet_t) * length);
      frontier_packet_array_length = length;
    }
  else if (frontier_packet_array_length < length)
    {
      frontier_packet_array
	= rcsp_realloc (frontier_packet_array,
			sizeof (packet_t) * frontier_packet_array_length,
			sizeof (packet_t) * length);
      frontier_packet_array_length = length;
    }
  for (i = 0, curr_packet = first_frontier_packet;
       curr_packet != NULL;
       i++, curr_packet = curr_packet->next_frontier_packet)
    {
      if (curr_packet->succ [0] == NULL || curr_packet->succ [1] != NULL)
	abort ();
      frontier_packet_array [i] = curr_packet;
    }
  if (i != length)
    abort ();
  qsort (frontier_packet_array, length, sizeof (packet_t),
	 frontier_packet_sort_cmp);
  pred_packet = NULL;
  for (i = 0; i < length; i++)
    {
      curr_packet = frontier_packet_array [i];
      if (pred_packet != NULL)
	pred_packet->next_frontier_packet = curr_packet;
      if (pred_packet != NULL
	  && curr_packet->first_op == NULL && pred_packet->first_op == NULL
	  && curr_packet->succ [0]->pred [1] != NULL
	  && pred_packet->succ[0]->number == curr_packet->succ[0]->number)
	/* Merge two frontier packets (CURR_PACKET and PRED_PACKET)
	   into PRED_PACKET: */
	{
#ifdef RCSP_DEBUG
	  if (rcsp_debug_level >= 2)
	    fprintf (f, "\n%s\n%s merging packets %3d and %3d into %3d\n",
		     prefix, prefix, pred_packet->number, curr_packet->number,
		     pred_packet->number);
#endif
	  /* Changing CFG: Redirect CURR_PACKET predecessors onto
             PRED_PACKET.  */
	  while (curr_packet->pred[0] != NULL)
	    /* Remember: packet_change_dest removes the link from pred
	       by shifting array `pred'.  */
	    packet_change_dest (curr_packet->pred[0], curr_packet,
				pred_packet);
	  /* Remember: packet_change_src removes the link from succ by
	     shifting array `succ'.  */
	  packet_change_src (curr_packet->succ[0], curr_packet, NULL);
	  if (curr_packet->succ[0] != NULL)
	    abort ();
	  /* Set packet issue time, CPU state and delay for available
             ops.  */
	  if (pred_packet->issue_time < curr_packet->issue_time)
	    {
	      decr_time = curr_packet->issue_time - pred_packet->issue_time;
	      while (decr_time-- != 0)
		cpu_state_advance_cycle (pred_packet->cpu_state);
	      pred_packet->issue_time = curr_packet->issue_time;
	    }
	  else if (pred_packet->issue_time > curr_packet->issue_time)
	    {
	      decr_time = pred_packet->issue_time - curr_packet->issue_time;
	      while (decr_time-- != 0)
		cpu_state_advance_cycle (curr_packet->cpu_state);
	    }
	  cpu_state_union (pred_packet->cpu_state, curr_packet->cpu_state);
	  packet_delete (curr_packet);
	}
      else
	pred_packet = curr_packet;
    }
  if (pred_packet == NULL)
    abort ();
  pred_packet->next_frontier_packet = NULL;
  first_frontier_packet = frontier_packet_array [0];
}

/* The following function initiates work with the frontier packet list.  */
static void
start_frontier_packet_list ()
{
  first_frontier_packet = NULL;
  frontier_packet_array = NULL;
}

/* The following function finishes work with the frontier packet
   list.  */
static void
finish_frontier_packet_list ()
{
  if (frontier_packet_array != NULL)
    rcsp_free (frontier_packet_array = NULL);
}



/* This page contains functions for evaluating CFG achievability,
   dominator and postdominator relation.  */

/* This recursive function returns
     o IGNORE_ACHIEVABILITY if we should ignore this path to calculate
       achievability: there is a cycle in CFG and there are no other
       paths (FROM, TO).
     o UNACHIEVABLE if TO is not achievable from FROM and there are no
       CFG cycles containing path (FROM, TO).
     o BASIC_BLOCK_PATH if TO is achievable from FROM and (TO,FROM) is
       a basic block.
     o OTHER_PATH otherwise.

   Remember the function may be called for formed packets too,
   therefore there may be cycles in CFG.  */
static achievable_mode_t
achievable_1 (from, to)
     packet_t from, to;
{
  int succ_number;
  int ignore_p;
  achievable_mode_t result;

  if (from->visit_number1 == last_packet_visit_number1)
    return from->achievable_result;
  from->visit_number1 = last_packet_visit_number1;
  /* This is needed if we walk around a cycle in CFG.  */
  from->achievable_result = IGNORE_ACHIEVABILITY;
  if (from == to)
    result = BASIC_BLOCK_PATH;
  else
    {
      result = UNACHIEVABLE;
      ignore_p = FALSE;
      for (succ_number = 0; from->succ [succ_number] != NULL; succ_number++)
	if ((result = achievable_1 (from->succ [succ_number], to))
	    == IGNORE_ACHIEVABILITY)
	  ignore_p = TRUE;
	else if (result != UNACHIEVABLE)
	  break;
      if (result == UNACHIEVABLE && ignore_p)
	result = IGNORE_ACHIEVABILITY;
      else if (result == BASIC_BLOCK_PATH
	       && (succ_number > 0
		   || from->succ [succ_number + 1] != NULL
		   || from->succ[0]->pred [1] != NULL))
	/* There are side exit from the patch (from, to) or more one
	   enter into the path.  */
	result = OTHER_PATH;
    }
#ifdef USE_ACHIEVABLE_CACHE
  if (CFG_shift <= MAX_CFG_SHIFT && result != IGNORE_ACHIEVABILITY)
    {
      CFG_validation [(from->number << CFG_shift) + to->number] = CFG_tick;
      CFG_achievable [(from->number << CFG_shift) + to->number] = result;
    }
#endif
  from->achievable_result = result;
  return result;
}


/* See comments for function `achievable_1'.  There is one exception
   -- the function never returns IGNORE_ACHIEVABILITY.  */
static achievable_mode_t
achievable (from, to)
     packet_t from, to;
{
  achievable_mode_t result;

  last_packet_visit_number1++;
  if (to == NULL || from == NULL)
    abort ();
  result = achievable_1 (from, to);
  if (result == IGNORE_ACHIEVABILITY)
    {
      /* We found cycle in CFG and there are no other paths (FROM,
         TO).  So TO is not accessible from FROM.  Remember that such
         conclusion may be not true for any intermediate CFG node.  */
      result = UNACHIEVABLE;
#ifdef USE_ACHIEVABLE_CACHE
      if (CFG_shift <= MAX_CFG_SHIFT)
	{
	  CFG_validation [(from->number << CFG_shift) + to->number] = CFG_tick;
	  CFG_achievable [(from->number << CFG_shift) + to->number] = result;
	}
#endif
    }
  return result;
}

/* The following recursive function processes PACKET (which may be on
   a path (FROM, TO)) and returns TRUE if FROM is a dominator of
   TO.  */
static int
dom_p_1 (from, packet, to)
     packet_t from, packet, to;
{
  int succ_number, pred_number;
  packet_t succ, pred;
  achievable_mode_t path;

  path = ACHIEVABLE (packet, to);
  if (path == UNACHIEVABLE)
    return TRUE;
  else if (path == BASIC_BLOCK_PATH)
    return TRUE;
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    {
      if (!dom_p_1 (from, succ, to))
	return FALSE;
      for (pred_number = 0;
	   (pred = succ->pred [pred_number]) != NULL;
	   pred_number++)
	if (pred != succ && ACHIEVABLE (from, pred) == UNACHIEVABLE)
	  return FALSE;
    }
  return TRUE;
}

/* The following function returns TRUE if FROM is a dominator of TO.
   TO and FROM should be not formed packets.  In this case traversed
   CFG part has no cycles.  */
static int
dom_p (from, to)
     packet_t from, to;
{
  if (to->formed_packet_p || from->formed_packet_p)
    abort ();
  return dom_p_1 (from, from, to);
}

/* The following recursive function returns TRUE if TO is a
   post-dominator of FROM.  */
static int
postdom_p_1 (to, from)
     packet_t to, from;
{
  int succ_number;
  packet_t succ;
  achievable_mode_t path;

  path = ACHIEVABLE (from, to);
  if (path == UNACHIEVABLE)
    return FALSE;
  else if (path == BASIC_BLOCK_PATH)
    return TRUE;
  for (succ_number = 0;
       (succ = from->succ [succ_number]) != NULL;
       succ_number++)
    if (!postdom_p_1 (to, succ))
      return FALSE;
  return TRUE;
}

/* The following function returns TRUE if TO is a post-dominator of
   FROM.  TO and FROM should be not formed packets.  In this case
   traversed CFG part has no cycles.  */
static int
postdom_p (to, from)
     packet_t to, from;
{
  if (to->formed_packet_p || from->formed_packet_p)
    abort ();
  return postdom_p_1 (to, from);
}




/* This page contains functions for defining does moving operation
   kill a live location.  */

/* When we are processing operation FROM on which OP is depended, the
   following function returns TRUE if live LOC is killed after moving
   OP to PACKET.  */
static int
kill_loc_move_p (op, packet, from, loc)
     op_t op ATTRIBUTE_UNUSED;
     packet_t packet;
     op_t from;
     location_t loc;
{
  dep_t curr_out_dep;

  for (curr_out_dep = dep_out_find (from, loc);
       curr_out_dep != NULL;
       curr_out_dep = curr_out_dep->next_out_loc_dep)
    if (curr_out_dep->dep_mode == DEF_USE
	/* We can place OP after insn which uses LOC in the same package: */
	&& packet != curr_out_dep->to->packet
	/* Ignore OP itself: */
	&& op->packet != curr_out_dep->to->packet
	&& ACHIEVABLE (packet, curr_out_dep->to->packet) != UNACHIEVABLE)
      return TRUE;
  return FALSE;
}

/* The following function returns TRUE if a live location is killed
   after moving OP to PACKET.  */
static int
kill_move_p (op, packet)
     op_t op;
     packet_t packet;
{
  dep_t curr_in_dep;
  location_t curr_loc;
  int i;

  if (ACHIEVABLE (packet, op->packet) == BASIC_BLOCK_PATH)
    /* Moving in basic block.  */
    return FALSE;
  for (curr_in_dep = op->in_deps;
       curr_in_dep != NULL;
       curr_in_dep = curr_in_dep->next_in)
    if (curr_in_dep->dep_mode == DEF_DEF
	&& ACHIEVABLE (curr_in_dep->from->packet, packet) != UNACHIEVABLE)
      {
        if (kill_loc_move_p (op, packet, curr_in_dep->from,
			     curr_in_dep->to_loc))
          return TRUE;
	for (i = 0; ; i++)
          {
            curr_loc = loc_super (curr_in_dep->to_loc, i == 0);
            if (curr_loc == NULL)
              break;
            if (kill_loc_move_p (op, packet, curr_in_dep->from, curr_loc))
              return TRUE;
          }

	for (i = 0; ; i++)
	  {
	    curr_loc = loc_sub (curr_in_dep->to_loc, i == 0);
	    if (curr_loc == NULL)
	      break;
	    if (kill_loc_move_p (op, packet, curr_in_dep->from, curr_loc))
	      return TRUE;
	  }
      }
  return FALSE;
}



/* This page contains functions for evaluation of set "covered".  */

/* The following recursive function marks (transitively) all packets
   (except for PACKET) which are are achievable from PACKET and which
   are sources of dependencies for OP.  */
static void
mark_src_packets (op, packet)
     op_t op;
     packet_t packet;
{
  packet_t from_packet;
  dep_t curr_in_dep;

  if (op->packet->visit_number2 == last_packet_visit_number2)
    return;
  for (curr_in_dep = op->in_deps;
       curr_in_dep != NULL;
       curr_in_dep = curr_in_dep->next_in)
    {
      from_packet = curr_in_dep->from->packet;
      if (packet != from_packet
	  && ACHIEVABLE (packet, from_packet) != UNACHIEVABLE)
	{
	  mark_src_packets (curr_in_dep->from, packet);
	  from_packet->visit_number2 = last_packet_visit_number2;
	}
    }
}

/* The following variable is list of packets which are in set
   `covered' (see comments for function `mark_covered').  */
static packet_t covered_list;

/* The following recursive function traverses CFG starting with PACKET
   and marks packets which are in set `covered' for OP and increments
   COVERED_SIZE after each marking.  The function returns TRUE if
   PACKET is in set `covered'.  */
static int
mark_covered_1 (op, packet, covered_size)
     op_t op;
     packet_t packet;
     int *covered_size;
{
  int succ_number;
  packet_t succ;

  if (packet->visit_number2 == last_packet_visit_number2 - 1)
    /* Op is src of a dependency for OP.  It can not be in set
       `covered'.  It is supposed that before this traversing CFG,
       traversing for marking packets which are sources of
       dependencies for OP has been made.  */
    return FALSE;
  else if (packet->visit_number2 == last_packet_visit_number2)
    /* Op is already in set `covered' */
    return TRUE;
  else if (op->packet == packet)
    {
      /* Mark that packet of OP is in set `covered' because it is end
	 of all paths.  */
      packet->visit_number2 = last_packet_visit_number2;
      packet->next_packet = covered_list;
      covered_list = packet;
      (*covered_size)++;
      return TRUE;
    }
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (ACHIEVABLE (succ, op->packet) != UNACHIEVABLE
	&& mark_covered_1 (op, succ, covered_size)
        && packet->visit_number2 != last_packet_visit_number2)
      {
        /* Mark that packet is in set `covered'.  */
        packet->visit_number2 = last_packet_visit_number2;
	packet->next_packet = covered_list;
	covered_list = packet;
        (*covered_size)++;
      }
  return packet->visit_number2 == last_packet_visit_number2;
}

/* The following function marks all packets which are in set `covered'
   for OP and PACKET.  The packet is in the set if there is a path
   (PACKET, OP) in CFG and OP is not data dependent from any packets
   on the path.  The function returns size of the set.  If OP can be
   moved into PACKET (i.e. there is a such path), the set contains
   PACKET and OP too.  The function also forms list of packets in set
   `covered'.  */
static int
mark_covered (op, packet)
     op_t op;
     packet_t packet;
{
  int covered_size = 0;

  if (op->packet == packet)
    abort ();
  last_packet_visit_number2++;
  mark_src_packets (op, packet);
  last_packet_visit_number2++;
  covered_list = NULL;
  mark_covered_1 (op, packet, &covered_size);
  return covered_size;
}



/* This page contains functions for defining move consistence and cost
   of moving operation.  */

/* The following recursive function traverses CFG starting with PACKET
   and counts number of packets which should be copied in order to
   make moving OP (see function move_cost).  It is supposed that
   marking packets in set `covered' made before this.  */
static int
traverse_CFG_for_consistent_cost (op, packet)
     op_t op;
     packet_t packet;
{
  int cost = 0;
  int succ_number;
  packet_t succ;

  if (packet->visit_number3 == last_packet_visit_number3)
    return 0;
  packet->visit_number3 = last_packet_visit_number3;
  if (packet->visit_number2 == last_packet_visit_number2)
    /* packet is in set `covered'.  */
    cost++;
  if (packet != op->packet)
    for (succ_number = 0;
	 (succ = packet->succ [succ_number]) != NULL;
	 succ_number++)
      if (ACHIEVABLE (succ, op->packet) != UNACHIEVABLE)
        cost += traverse_CFG_for_consistent_cost (op, succ);
  return cost;
}

/* The following function evaluates cost of moving OP into PACKET.
   The cost is number of necessary packet copies for making CGF
   consistent for moving OP and size of covered set when OP is a
   branch.  Negative cost means that OP can not be moved into PACKET.
   Otherwise, the function also sets up parameter INCONSISTENCE_P when
   making CFG move consistence is required.  */
static int
move_cost (op, packet, inconsistence_p)
     op_t op;
     packet_t packet;
     int *inconsistence_p;
{
  packet_t curr_packet;
  dep_t curr_in_dep;
  int cost, covered_size;
  
  if (op->packet == packet)
    abort ();
  if (ACHIEVABLE (packet, op->packet) == UNACHIEVABLE
      || kill_move_p (op, packet))
    return -1;
  covered_size = mark_covered (op, packet);
  if (covered_size == 0)
    return -1;
  if (covered_size < 2)
    abort ();
  cost = 0;
  /* Add number packet copies for making CFG move consistent.  */
  last_packet_visit_number3++;
  for (curr_in_dep = op->in_deps;
       curr_in_dep != NULL;
       curr_in_dep = curr_in_dep->next_in)
    if (packet != curr_in_dep->from->packet
	&& ACHIEVABLE (packet, curr_in_dep->from->packet) != UNACHIEVABLE)
      cost += traverse_CFG_for_consistent_cost (op, curr_in_dep->from->packet);
  /* If OP is achievable from another frontier packet, take into
     account necessary copies on path from the frontier packet to OP.  */
  if (!dom_p (packet, op->packet))
    for (curr_packet = first_frontier_packet;
	 curr_packet != NULL;
	 curr_packet = curr_packet->next_frontier_packet)
      {
	if (curr_packet == packet)
	  abort ();
	if (ACHIEVABLE (curr_packet, op->packet) != UNACHIEVABLE)
	  cost += traverse_CFG_for_consistent_cost (op, curr_packet);
      }
  if (cost != 0 && op->use_def_p)
    return -1;
  *inconsistence_p = cost != 0;
  return (cond_branch_p (op->insn) ? covered_size - 2 : 0) + cost;
}



/* The following function returns negative value if OP is not ready
   for PACKET (i.e. there is no a path from root to OP through PACKET
   on which all ops whose outputs are OP inputs have been issued).
   Otherwise the function returns absolute maximal time when all the
   inputs will be ready.  */
static int
get_op_ready_time (op, packet)
     op_t op;
     packet_t packet;
{
  int loc_value_ready_p;
  dep_t curr_in_dep;
  location_t prev_loc;
  int issue_time, ready_time;

  prev_loc = NULL;
  loc_value_ready_p = TRUE;
  ready_time = 0;
  for (curr_in_dep = op->in_deps;
       curr_in_dep != NULL;
       curr_in_dep = curr_in_dep->next_in)
    if (curr_in_dep->dep_mode == DEF_USE)
      {
	if (prev_loc == NULL || !loc_eq_p (prev_loc, curr_in_dep->to_loc))
	  {
	    if (!loc_value_ready_p)
	      return -1;
	    loc_value_ready_p = FALSE;
	    prev_loc = curr_in_dep->to_loc;
	  }
	if (curr_in_dep->from->packet == packet
	    || (curr_in_dep->from->packet->formed_packet_p
		&& ACHIEVABLE (curr_in_dep->from->packet, packet)
		   != UNACHIEVABLE))
	  {
	    loc_value_ready_p = TRUE;
	    issue_time = curr_in_dep->from->packet->issue_time;
	    if (issue_time < 0)
	      abort ();
	    if (ready_time < issue_time + curr_in_dep->latency)
	      ready_time = issue_time + curr_in_dep->latency;
	  }
      }
  return (loc_value_ready_p ? ready_time : -1);
}



/* This page contains functions to make CFG move consistent.  */

/* The following variable is list of copied packets in order to make
   CFG is move consistent.  */
static packet_t copied_packets_4_move_consistence;

/* The following function sets up in/out dependencies for copied
   packets.  */
static void
copy_dependencies_4_consistent_move (packet)
     packet_t packet;
{
  dep_t in_deps, out_deps, curr_dep;

  if (packet->visit_number2 != last_packet_visit_number2
      || packet->visit_number3 != last_packet_visit_number3)
    abort ();
  if (packet->first_op == NULL || packet->first_op->next_op != NULL)
    abort ();
  in_deps = packet->first_op->in_deps;
  out_deps = packet->first_op->out_deps;
  if (packet->copy->first_op->next_op != NULL)
    abort ();
  for (curr_dep = out_deps; curr_dep != NULL; curr_dep = curr_dep->next_out)
    {
      if (curr_dep->to->packet->first_op->next_op != NULL)
	abort ();
      if (curr_dep->to->packet->visit_number2 == last_packet_visit_number2)
	{
	  if (curr_dep->to->packet->visit_number3 != last_packet_visit_number3)
	    abort ();
	  dep_create (packet->copy->first_op,
		      curr_dep->to->packet->copy->first_op,
		      curr_dep->to_loc, curr_dep->dep_mode, curr_dep->latency);
	}
      else
	dep_create (packet->copy->first_op, curr_dep->to->packet->first_op,
		    curr_dep->to_loc, curr_dep->dep_mode, curr_dep->latency);
    }
  for (curr_dep = in_deps; curr_dep != NULL; curr_dep = curr_dep->next_in)
    if (curr_dep->from->packet->visit_number2 == last_packet_visit_number2
	&& curr_dep->from->packet->visit_number3 == last_packet_visit_number3)
      {
	op_t curr_op, curr_copy_op;

	/* Input dependency may be from formed packet which may have
           more one operation.  */
	for (curr_op = curr_dep->from->packet->first_op,
	       curr_copy_op = curr_dep->from->packet->copy->first_op;
	     curr_op != curr_dep->from;
	     curr_op = curr_op->next_op, curr_copy_op = curr_copy_op->next_op)
	  ;
	dep_create (curr_copy_op, packet->copy->first_op, curr_dep->to_loc,
		    curr_dep->dep_mode, curr_dep->latency);
      }
    else
      dep_create (curr_dep->from, packet->copy->first_op,
		  curr_dep->to_loc, curr_dep->dep_mode, curr_dep->latency);
}

/* The following function removes obsolete (from unachievable packet
   operations) input dependencies of operation of interior PACKET and
   its copy.  */
static void
remove_obsolete_dependencies_4_move_consistence (packet)
     packet_t packet;
{
  dep_t curr_dep, next_dep;

  if (packet->first_op == NULL || packet->first_op->next_op != NULL)
    abort ();
  for (curr_dep = packet->first_op->in_deps;
       curr_dep != NULL;
       curr_dep = next_dep)
    {
      next_dep = curr_dep->next_in;
      if (ACHIEVABLE (curr_dep->from->packet, packet) == UNACHIEVABLE)
	dep_delete (curr_dep);
    }
  for (curr_dep = packet->copy->first_op->in_deps;
       curr_dep != NULL;
       curr_dep = next_dep)
    {
      next_dep = curr_dep->next_in;
      if (ACHIEVABLE (curr_dep->from->packet, packet->copy) == UNACHIEVABLE)
	dep_delete (curr_dep);
    }
}

/* The following function modifies CFG which was not valid after
   making move consistent packets copies (PACKET is a packet for which
   move consistent copy has been made).  Actually the order of
   processing packets for which move consistent copies have been made
   is not important.  Practically each packet is processed twice: as
   pred and as succ.  */
static void
modify_packet_CFG (packet)
     packet_t packet;
{
  packet_t pred, succ;
  int pred_number, succ_number;

  if (packet->visit_number2 != last_packet_visit_number2
      || packet->visit_number3 != last_packet_visit_number3)
    abort ();
  for (pred_number = 0;
       (pred = packet->pred [pred_number]) != NULL;
       pred_number++)
    if (pred->visit_number2 == last_packet_visit_number2)
      {
	/* pred is in `covered' */
	if (pred->visit_number3 != last_packet_visit_number3)
	  /* No copy of pred is made: remove arc (pred, packet->copy).  */
	  packet_change_dest (pred, packet->copy, packet);
	else
	  /* Copy is made: remove possible arc (pred->copy, packet)
	     and add arc (pred->copy, packet->copy).  */
	  packet_change_dest (pred->copy, packet, packet->copy);
      }
    else
      {
	/* Pred is not in `covered' (it means there is no copy):
	   remove arc (pred, packet) and add arc (pred, packet->copy).
	   Remember: packet_change_dest removes the link from pred by
	   shifting array `pred'.*/
	packet_change_dest (pred, packet, packet->copy);
	pred_number--;
      }

  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (succ->visit_number2 == last_packet_visit_number2)
      {
	/* succ is in `covered' */
	if (succ->visit_number3 != last_packet_visit_number3)
	  /* It should be visited.  */
	  abort ();
	packet_change_src (succ->copy, packet, packet->copy);
      }
    else
      /* We do not change true or false target in branch because they
         are the same after copying.  */
      packet_connect (packet->copy, succ, FALSE, FALSE);
}

/* The following function makes GFG consistent for moving OP into
   packet TO.  */
static void
make_move_consistent (op, to)
     op_t op;
     packet_t to;
{
  packet_t curr_packet, next_packet;
  int inconsistence_p;
  packet_t curr_frontier_packet;
  int ready_p, copy_ready_p;

  if (move_cost (op, to, &inconsistence_p) <= 0 || !inconsistence_p)
    abort ();
  copied_packets_4_move_consistence = NULL;
  for (curr_packet = covered_list;
       curr_packet != NULL;
       curr_packet = next_packet)
    {
      next_packet = curr_packet->next_packet;
      if (curr_packet->visit_number3 == last_packet_visit_number3)
	{
	  if (curr_packet->visit_number2 != last_packet_visit_number2)
	    abort ();
	  curr_packet->copy = packet_copy (curr_packet);
	  curr_packet->next_packet = copied_packets_4_move_consistence;
	  copied_packets_4_move_consistence = curr_packet;
	}
    }
  /* Modify CFG.  */
  for (curr_packet = copied_packets_4_move_consistence;
       curr_packet != NULL;
       curr_packet = curr_packet->next_packet)
    modify_packet_CFG (curr_packet);
  /* Modify DDG.  */
  for (curr_packet = copied_packets_4_move_consistence;
       curr_packet != NULL;
       curr_packet = curr_packet->next_packet)
    copy_dependencies_4_consistent_move (curr_packet);
  dep_sort_added_in_deps_op ();
  for (curr_packet = copied_packets_4_move_consistence;
       curr_packet != NULL;
       curr_packet = curr_packet->next_packet)
    remove_obsolete_dependencies_4_move_consistence (curr_packet);
#if 1
  /* ??? Update ready ops.  After copying packets, copied ops are also
     ready ops if their originals are.  Some of ops (copied ones and
     originals) may be not really ready because of CFG changes.  Is it
     worth to remove them from ready ops.  To remain them is in ready
     ops is safe but it requires more time to evaluate available ops
     for packets.  */
  for (curr_packet = copied_packets_4_move_consistence;
       curr_packet != NULL;
       curr_packet = curr_packet->next_packet)
    if (curr_packet->first_op->ready_p)
      {
	if (!curr_packet->copy->first_op->ready_p)
	  abort ();
	copy_ready_p
	  = get_op_ready_time (curr_packet->copy->first_op, to) >= 0;
	ready_p = get_op_ready_time (curr_packet->first_op, to) >= 0;
	for (curr_frontier_packet = first_frontier_packet;
	     curr_frontier_packet != NULL;
	     curr_frontier_packet = curr_frontier_packet->next_frontier_packet)
	  {
	    if (!ready_p && get_op_ready_time (curr_packet->first_op,
					       curr_frontier_packet) >= 0)
	      ready_p = TRUE;
	    if (!copy_ready_p
		&& get_op_ready_time (curr_packet->copy->first_op,
				      curr_frontier_packet) >= 0)
	      copy_ready_p = TRUE;
	    if (copy_ready_p && ready_p)
	      break;
	  }
	if (!ready_p)
	  op_become_not_ready (curr_packet->first_op);
	if (!copy_ready_p)
	  op_become_not_ready (curr_packet->copy->first_op);
      }
#endif
}



/* This page contains the rest of software pipeliner function.  */

/* The following function adds new dependencies after moving OP on LOC
   used (if USAGE_P) or defined in OP when we are processing operation
   FROM on which OP is depended.  */
static void
update_loc_deps (op, from, loc, usage_p)
     op_t op;
     op_t from;
     location_t loc;
     int usage_p;
{
  dep_t curr_out_dep;
  dep_mode_t dep_mode;

  for (curr_out_dep = dep_out_find (from, loc);
       curr_out_dep != NULL;
       curr_out_dep = curr_out_dep->next_out_loc_dep)
    if ((!usage_p || curr_out_dep->dep_mode != DEF_USE)
	&& op != curr_out_dep->to
	/* Ignore prev ops in OP packet.  */
	&& op->packet != curr_out_dep->to->packet
	&& ACHIEVABLE (op->packet, curr_out_dep->to->packet) != UNACHIEVABLE)
      {
	/* DEF_USE is not possible here because we could kill a live
	   location value.  It is guaranteed by function
	   `kill_move_p'.  */
	dep_mode = (usage_p ? USE_DEF : DEF_DEF);
	dep_create (op, curr_out_dep->to, curr_out_dep->to_loc, dep_mode,
		    insn_dep_time (op->insn, curr_out_dep->to->insn,
				   dep_mode));
      }
}

/* The following function updates DDG after OP has been moved.  */
static void
update_deps (op)
     op_t op;
{
  dep_t curr_in_dep;
  location_t curr_loc;
  int i;

  for (curr_in_dep = op->in_deps;
       curr_in_dep != NULL;
       curr_in_dep = curr_in_dep->next_in)
    {
      update_loc_deps (op, curr_in_dep->from, curr_in_dep->to_loc,
                       curr_in_dep->dep_mode == DEF_USE);
      for (i = 0; ; i++)
        {
          curr_loc = loc_super (curr_in_dep->to_loc, i == 0);
          if (curr_loc == NULL)
            break;
          update_loc_deps (op, curr_in_dep->from, curr_loc,
                           curr_in_dep->dep_mode == DEF_USE);
        }

      for (i = 0; ; i++)
	{
	  curr_loc = loc_sub (curr_in_dep->to_loc, i == 0);
	  if (curr_loc == NULL)
	    break;
	  update_loc_deps (op, curr_in_dep->from,
			   curr_in_dep->to_loc,
			   curr_in_dep->dep_mode == DEF_USE);
	}
    }
}

/* The following function creates new packet with unconditional jump
   between packets FROM and TO.  The function returns created
   operation.  */
static op_t
create_simple_jump (from, to, initiator_op)
     packet_t from, to;
     op_t initiator_op;
{
  packet_t new_packet;
  op_t new_op;
  insn_t new_insn;

  new_packet = packet_create (-1);
  new_packet->exit_packet_p = from->exit_packet_p;
  new_packet->formed_packet_p = to->formed_packet_p;
  new_insn = create_uncond_branch (to->first_op->insn);
  new_op = op_create (new_insn, to->first_op->it_num);
  new_op->initiator_op = initiator_op;
  packet_append_op (new_packet, new_op, FALSE);
  packet_change_dest (from, to, new_packet);
  packet_connect (new_packet, to, TRUE, TRUE);
  return new_op;
}

/* The following function creates new avail list element of PACKET
   with given characteristics (OP, AVAIL_P, COST, INCONSISTENCE_P,
   READY_TIME).  */
static void
create_avail_list_el (packet, op, avail_p, cost, inconsistence_p, ready_time)
     packet_t packet;
     op_t op;
     int avail_p;
     int cost;
     int inconsistence_p;
     int ready_time;
{
  avail_list_t new_el;

  if (op->packet->visit_number4 == last_packet_visit_number4)
    /* It is already in available ops list.  */
    return;
  op->packet->visit_number4 = last_packet_visit_number4;
  new_el = avail_list_el_create (packet, op, avail_p);
  new_el->cost = cost;
  new_el->inconsistence_p = inconsistence_p;
  new_el->delay = ready_time - curr_time;
  if (new_el->delay < 0)
    new_el->delay = 0;
  new_el->next_avail = packet->avail_list;
  if (packet->avail_list != NULL)
    packet->avail_list->prev_avail = new_el;
  packet->avail_list = new_el;
  if (!op->packet->exit_packet_p && op->it_num < packet->min_it_n && avail_p)
    /* It is not possible situation for available operations.  */
    abort ();
}

/* The following function returns TRUE if we need a unconditional jump
   insn between PACKET and its successor SUCC.  */
static int
needs_jump_p (packet, succ)
     packet_t packet, succ;
{
  packet_t pred;
  int pred_number;
  op_t branch;

  if (packet_find_branch (packet, succ, TRUE) != NULL)
    return dest_packet_requires_jump_p (succ);
  for (pred_number = 0;
       (pred = succ->pred [pred_number]) != NULL;
       pred_number++)
    if (pred != packet)
      {
	branch = packet_find_branch (pred, succ, FALSE);
	if (branch == NULL
	    || (cond_branch_p (branch->insn)
		&& ((branch->true_jump_p && branch->false_target == succ)
		    || (!branch->true_jump_p && branch->true_target == succ))
		&& branch->target_fixed_p))
	  break;
      }
  return pred != NULL || dest_packet_requires_jump_p (succ);
}

/* The following function processes all new conditional branches in
   PACKET and adds necessary jumps or converts them for CFG
   consistence.  */
static void
add_consistent_jumps (packet)
     packet_t packet;
{
  op_t curr_op, branch, new_op;
  packet_t next_packet, jump_packet, pred_packet, pred_packet1;
  packet_t pred;
  int pred_number;
  int require_jump_p, can_be_converted_p;
  avail_list_t new_el ATTRIBUTE_UNUSED;

  for (curr_op = packet->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    if (!insn_dummy_p (curr_op->insn) && cond_branch_p (curr_op->insn)
	&& !curr_op->processed_4_consistent_jumps_p)
      {
	curr_op->processed_4_consistent_jumps_p = TRUE;
	pred_packet = packet;
	if (curr_op->true_jump_p)
	  {
	    next_packet = curr_op->false_target;
	    jump_packet = curr_op->true_target;
	  }
	else
	  {
	    next_packet = curr_op->true_target;
	    jump_packet = curr_op->false_target;
	  }
	while (next_packet != NULL)
	  {
	    if (next_packet->first_op != NULL
		&& !insn_dummy_p (next_packet->first_op->insn))
	      {
		if (uncond_branch_p (next_packet->first_op->insn))
		  break;
	      }
 	    require_jump_p = needs_jump_p (pred_packet, next_packet);
 	    if (require_jump_p)
	      {
		can_be_converted_p = FALSE;
		if (!require_jump_p && pred_packet == packet
		    && !curr_op->target_fixed_p && curr_op->can_be_converted_p)
		  {
		    can_be_converted_p = TRUE;
		    pred_packet1 = packet;
		    /* Look at the jump path.  */
		    while (jump_packet != NULL)
		      {
			if (jump_packet->first_op != NULL
			    && !insn_dummy_p (jump_packet->first_op->insn))
			  {
			    if (uncond_branch_p (jump_packet->first_op->insn))
			      break;
			    if (jump_packet->first_op->next_op != NULL)
			      abort ();
			  }
			for (pred_number = 0;
			     (pred = jump_packet->pred [pred_number]) != NULL;
			     pred_number++)
			  if (pred != pred_packet1)
			    {
			      branch = packet_find_branch (pred, jump_packet,
							   FALSE);
			      if (branch == NULL
				  ||
				  (cond_branch_p (branch->insn)
				   && ((branch->true_jump_p
					&& branch->false_target == jump_packet)
				       ||
				       (!branch->true_jump_p
					&& branch->true_target == jump_packet))
				   && branch->target_fixed_p))
				break;
			    }
			if (pred != NULL)
			  {
			    /* There is another fall into JUMP_PACKET
                               and this path has been already
                               fixed.  */
			    can_be_converted_p = FALSE;
			    break;
			  }
			pred_packet1 = jump_packet;
			if (jump_packet->succ [0] != NULL
			    && jump_packet->succ [1] != NULL)
			  /* We already processed the branch in JUMP_PACKET
                             or will look at it later.  */
			  break;
			jump_packet = jump_packet->succ [0];
		      }
		  }
		if (can_be_converted_p)
		  {
		    /* convert curr_op */
		    curr_op->true_jump_p = !curr_op->true_jump_p;
		    convert_cond_branch (curr_op->insn);
		  }
		else
		  {
		    /* Insert jump.  */
		    new_op = create_simple_jump (pred_packet, next_packet,
						 curr_op);
#if 0
		    {
		      int cost, inconsistence_p;

		      /* ??? The following code will be needed if we
			 add jumps during scheduling itself (see
			 comments for add_consistent_jumps and
			 convert_cond_branch).  */
		      /* We require that the unconditional jump does
			 not create new data dependencies.  */
		      /* Add the jump into ready op list and available
			 ops list of PACKET.  */
		      op_become_ready (new_op);
		      cost = move_cost (new_op, packet, &inconsistence_p);
		      if (cost < 0)
			abort (0);
		      create_avail_list_el (packet, new_op, TRUE,
					    cost, inconsistence_p, curr_time);
		    }
#endif
		  }
		/* We inserted jump or converted the branch.  So we
                   can start processing of the next operation of
                   PACKET.  */
		break;
	      }
	    pred_packet = next_packet;
	    if (next_packet->succ [0] != NULL && next_packet->succ [1] != NULL)
	      /* We already processed the branch in NEXT_PACKET or
                 will look at it later.  */
	      break;
	    next_packet = next_packet->succ [0];
	  }
	curr_op->target_fixed_p = TRUE;
      }
}

/* The following function updates ready ops list after scheduling OP
   and available ops list for packet of OP if AVAIL_P is TRUE.  */
static void
update_ready_avail_ops (op, avail_p)
     op_t op;
     int avail_p;
{
  avail_list_t next_el, curr_el;
  dep_t curr_dep;
  op_t candidate;
  int ready_time, cost, inconsistence_p;
  packet_t op_packet = op->packet;

  if (avail_p)
    {
      /* Modify existing list.  Move of OP may transform given
         operation from one requiring making CFG move consistent into
         one not requiring making CFG move consistent.  Consequently
         the cost may also be changed.  Speculated move of OP may make
         available operation unavailable because the operation will
         kill a location of the moved OP.  */
      for (curr_el = op_packet->avail_list; curr_el != NULL; curr_el = next_el)
	{
	  next_el = curr_el->next_avail;
	  cost = move_cost (curr_el->op, op_packet, &inconsistence_p);
	  if (cost < 0)
	    {
	      /* Remove curr_el from available list.  */
	      if (curr_el->prev_avail == NULL)
		op_packet->avail_list = next_el;
	      else
		curr_el->prev_avail->next_avail = next_el;
	      if (next_el != NULL)
		next_el->prev_avail = curr_el->prev_avail;
	      avail_list_el_free (curr_el);
	    }
	  else
	    {
	      curr_el->cost = cost;
	      curr_el->inconsistence_p = inconsistence_p;
	    }
	}
    }
  for (curr_dep = op->out_deps;
       curr_dep != NULL;
       curr_dep = curr_dep->next_out)
    {
      candidate = curr_dep->to;
      ready_time = -1;
      if (avail_p
	  && candidate->it_num < op_packet->min_it_n + iteration_look_ahead)
	{
	  ready_time = get_op_ready_time (candidate, op_packet);
	  if (ready_time < 0)
	    continue;
	  cost = move_cost (candidate, op_packet, &inconsistence_p);
	  if (cost >= 0)
	    create_avail_list_el (op_packet, candidate, TRUE,
				  cost, inconsistence_p, ready_time);
	}
      if (!candidate->ready_p)
	{
	  if (ready_time < 0)
	    {
	      ready_time = get_op_ready_time (candidate, op_packet);
	      if (ready_time < 0)
		continue;
	    }
	  op_become_ready (candidate);
	}
    }
}

/* The following function adds OP of an interior packet to frontier
   PACKET and updates CFG and DDG.  */
static void
schedule_op (op, packet)
     op_t op;
     packet_t packet;
{
  packet_t op_packet = op->packet;
  int basic_block_move_p;
  op_t curr_op;
  packet_t true_target, false_target, curr_packet, succ, pred;
  int succ_number, pred_number;

  if (op == NULL || op->next_op != NULL)
    abort ();
  basic_block_move_p = ACHIEVABLE (packet, op_packet) == BASIC_BLOCK_PATH;
  if (packet->succ[0] != NULL && packet->succ[1] != NULL
      && cond_branch_p (packet->last_op->insn))
    {
      curr_op = packet->last_op;
#if 0
      /* This code may be present when we add jumps during scheduling.  */
      if (!curr_op->target_fixed_p)
	/* It should be fixed in add_consistent_jumps.  */
	abort ();
#endif
      if (curr_op->true_jump_p)
	curr_op->false_target = NULL;
      else
        curr_op->true_target = NULL;
    }
  if (nonbranch_p (op->insn) || uncond_branch_p (op->insn))
    {
      if (*op_packet->succ == NULL || op_packet->succ[1] != NULL)
        abort ();
      packet_append_op (packet, op, FALSE);
      op_packet->first_op = op_packet->last_op = NULL;
      /* Remove the packet from CFG.  */
      succ = *op_packet->succ;
      for (pred_number = 0;
	   (pred = op_packet->pred [pred_number]) != NULL;
	   pred_number++)
	packet_change_src (succ, op_packet, pred);
      while ((pred = op_packet->pred[0]) != NULL)
	/* Remember: packet_change_dest removes the link from pred by
	   shifting array `pred'.  */
	packet_change_dest (pred, op_packet, succ);
    }
  else if (cond_branch_p (op->insn))
    {
      true_target = op->true_target;
      false_target = op->false_target;
      if (true_target == NULL || false_target == NULL
	  || *op_packet->succ == NULL || op_packet->succ [1] == NULL
	  || op_packet->succ [2] != NULL)
	abort ();
      if (packet->succ[0] != NULL && packet->succ[1] != NULL)
	/* ??? Now we place only one branch into packet.  See comments
           in get_best_candidate_avail_op.  If we permit more one
           branch, the following code should be changed.  */
	abort ();
      /* Create copies of packets in "covered" except for the root and
         the branch.  */
      mark_covered (op, packet);
      for (curr_packet = covered_list;
	   curr_packet != NULL;
	   curr_packet = curr_packet->next_packet)
        if (curr_packet != op_packet && curr_packet != packet)
	  {
	    if (curr_packet->succ[1] != NULL
		&& (curr_packet->succ[0] == op_packet
		    || curr_packet->succ[1] == op_packet)
		&& (curr_packet->succ[0] == true_target
		    || curr_packet->succ[0] == false_target
		    || curr_packet->succ[1] == true_target
		    || curr_packet->succ[1] == false_target))
	      /* We move cond branch through the 2nd cond branch and
                 2nd branch will refer only for one packet.  */
	      abort ();
	    curr_packet->copy = packet_copy (curr_packet);
	    if (true_target->exit_packet_p)
	      curr_packet->exit_packet_p = TRUE;
	    if (false_target->exit_packet_p)
	      curr_packet->copy->exit_packet_p = TRUE;
	  }

      /* Form the second control flow path from packet into the branch
         packet.  */
      for (curr_packet = covered_list;
	   curr_packet != NULL;
	   curr_packet = curr_packet->next_packet)
        if (curr_packet != op_packet && curr_packet != packet)
          {
	    for (succ_number = 0;
		 (succ = curr_packet->succ [succ_number]) != NULL;
		 succ_number++)
	      if (succ != op_packet)
		{
		  if (succ->visit_number2 == last_packet_visit_number2)
		    {
		      /* The successor belongs to "covered" */
		      packet_change_dest (curr_packet->copy, succ, succ->copy);
		      packet_change_src (succ->copy, curr_packet,
					 curr_packet->copy);
		    }
		  else
		    packet_connect (curr_packet->copy, succ, FALSE, FALSE);
		}
	    
	    for (pred_number = 0;
		 (pred = curr_packet->pred [pred_number]) != NULL;
		 pred_number++)
	      if (pred->visit_number2 != last_packet_visit_number2)
		/* The predecessor does not belong to "covered" --
                   side path.  Make the side path entering only into
                   original packets.  */
		packet_change_dest (pred, curr_packet->copy, curr_packet);
	  }
      
      /* Moving op from OP_PACKET into PACKET.  */
      packet_append_op (packet, op, FALSE);
      op_packet->first_op = op_packet->last_op = NULL;
      /* Changes src of {false, true} targets and dest of the
         packet.  */
      succ = packet->succ[0];
      if (succ == op_packet)
	{
	  packet_change_src (true_target, op_packet, packet);
	  packet_change_src (false_target, op_packet, packet);
	  packet_change_dest (packet, op_packet, true_target);
	}
      else
	{
	  if (succ->visit_number2 != last_packet_visit_number2)
	    abort ();
	  packet_change_dest (packet, true_target, succ);
	  packet_change_dest (packet, false_target, succ->copy);
	  while ((pred = op_packet->pred[0]) != NULL)
	    {
	      if (pred->visit_number2 != last_packet_visit_number2
		  || pred == packet)
		abort ();
	      packet_change_dest (pred->copy, op_packet, false_target);
	      packet_change_src (false_target, op_packet, pred->copy);
	      packet_change_dest (pred, op_packet, true_target);
	      packet_change_src (true_target, op_packet, pred);
	    }
	}

      /* Create dependencies for copied packets.  */
      for (curr_packet = covered_list;
	   curr_packet != NULL;
	   curr_packet = curr_packet->next_packet)
        if (curr_packet != op_packet && curr_packet != packet)
	  {
	    dep_t in_deps, out_deps, curr_dep, next_dep;

	    in_deps = curr_packet->first_op->in_deps;
	    out_deps = curr_packet->first_op->out_deps;
	    for (curr_dep = in_deps;
		 curr_dep != NULL;
		 curr_dep = curr_dep->next_in)
	      if (curr_dep->from->packet->visit_number2
		  == last_packet_visit_number2
		  && curr_dep->from->packet != packet)
		dep_create (curr_dep->from->packet->copy->first_op,
			    curr_packet->copy->first_op, curr_dep->to_loc,
			    curr_dep->dep_mode, curr_dep->latency);
	      else if (ACHIEVABLE (curr_dep->from->packet,
				   curr_packet->copy) != UNACHIEVABLE)
		dep_create (curr_dep->from, curr_packet->copy->first_op,
			    curr_dep->to_loc, 
			    curr_dep->dep_mode, curr_dep->latency);	      
	    for (curr_dep = out_deps; curr_dep != NULL; curr_dep = next_dep)
	      {
		next_dep = curr_dep->next_out;
		if (curr_dep->to->packet->visit_number2
		    != last_packet_visit_number2)
		  {
		    if (ACHIEVABLE (curr_packet, curr_dep->to->packet)
			== UNACHIEVABLE)
		      dep_delete (curr_dep);
		    if (ACHIEVABLE (curr_packet->copy,
				    curr_dep->to->packet) != UNACHIEVABLE)
		      dep_create (curr_packet->copy->first_op, curr_dep->to,
				  curr_dep->to_loc,
				  curr_dep->dep_mode, curr_dep->latency);
		  }

	      }
	  }
    }
  else
    abort ();
  if (!basic_block_move_p)
    update_deps (op);
  dep_sort_added_in_deps_op ();
  packet_delete (op_packet);
}

/* The following function is used as comparison function to sort
   available list to choose the best operation for scheduling.  */
static int
avail_op_cmp (avail_el1, avail_el2)
     avail_list_t *avail_el1;
     avail_list_t *avail_el2;
{
  avail_list_t av1 = *avail_el1;
  avail_list_t av2 = *avail_el2;
  
  if (av1->op == av2->op)
    /* No duplications.  */
    abort ();
  if (av1->avail_p && !av2->avail_p)
    return -1;
  else if (!av1->avail_p && av2->avail_p)
    return 1;
  else if (!av1->op->packet->exit_packet_p && av2->op->packet->exit_packet_p)
    return -1;
  else if (av1->op->packet->exit_packet_p && !av2->op->packet->exit_packet_p)
    return 1;
  else if (av1->delay < av2->delay)
    return -1;
  else if (av1->delay > av2->delay)
    return 1;
  else if (av1->cost < av2->cost)
    return -1;
  else if (av1->cost > av2->cost)
    return 1;
  else if (av1->op->it_num < av2->op->it_num)
    return -1;
  else if (av1->op->it_num > av2->op->it_num)
    return 1;
  /* We are trying branches as the last choice for scheduling.  */
  else if (!insn_dummy_p (av2->op->insn)
	   && (uncond_branch_p (av2->op->insn)
	       || cond_branch_p (av2->op->insn))
           && (insn_dummy_p (av1->op->insn)
	       || (!uncond_branch_p (av1->op->insn)
		   && !cond_branch_p (av1->op->insn))))
    return -1;
  else if (!insn_dummy_p (av1->op->insn)
	   && (uncond_branch_p (av1->op->insn)
	       || cond_branch_p (av1->op->insn))
           && (insn_dummy_p (av2->op->insn)
	       || (!uncond_branch_p (av2->op->insn)
		   && !cond_branch_p (av2->op->insn))))
    return 1;
  else
    /* ??? Additional ordering.  May be moving in basic block or
       speculative moving.  */
    return 0;
}

/* The following function creates empty packets for all successors and
   place them between PACKET and the successors.  The function also
   defines available lists for the packets and adds the packets to the
   list of frontier packets.  */
static void
add_next_packets (packet)
     packet_t packet;
{
  packet_t *target_ptr, r, old_to;

  if (packet->first_op == NULL)
    {
      /* It is empty packet.  We should merge it lately (with frontier
         packets which have the same successors.  */
      packet->formed_packet_p = FALSE;
      add_frontier_packet (packet);
      return;
    }
  for (target_ptr = packet->succ; *target_ptr != NULL; target_ptr++)
    if (!(*target_ptr)->stop_packet_p)
      {
	r = packet_create (curr_time);
	memcpy (r->cpu_state, packet->cpu_state, cpu_state_length);
	old_to = *target_ptr;
	r->exit_packet_p = old_to->exit_packet_p;
	packet_change_dest (packet, old_to, r);
	packet_connect (r, old_to, FALSE, FALSE);
      }
  for (target_ptr = packet->succ; *target_ptr != NULL; target_ptr++)
    if (!(*target_ptr)->stop_packet_p)
      add_frontier_packet (*target_ptr);
}

/* The function chooses PACKET available op list element with the best
   op for scheduling.  *NO_MORE_CANDIDATE_OP_P == TRUE means that the
   next call we will not find an available op.  */
static avail_list_t
get_best_candidate_avail_op (packet, no_more_candidate_op_p)
     packet_t packet;
     int *no_more_candidate_op_p;
{
  avail_list_t pred_el, curr_el;
  packet_t curr_packet ATTRIBUTE_UNUSED;
  int pred_number;
  packet_t pred;

  *no_more_candidate_op_p = TRUE;
  for (pred_el = NULL, curr_el = packet->avail_list;
       curr_el != NULL;
       pred_el = curr_el, curr_el = curr_el->next_avail)
    {
      if (!curr_el->avail_p)
	/* Current packet available list should contain only available ops.  */
	abort ();
      if (!packet->exit_packet_p && curr_el->op->packet->exit_packet_p)
	/* We never move op from exit packet into the loop body.  */
	continue;
#if 0
      /* This is restriction on choosing op for scheduling.  Such code
         would generate more compact pipelined loops but less
         effective.  ??? How less effective.  We need to examine
         this.  */
      for (curr_packet = first_frontier_packet;
	   curr_packet != NULL;
	   curr_packet = curr_packet->next_frontier_packet)
	if (ACHIEVABLE (curr_packet, curr_el->op->packet) != UNACHIEVABLE)
	  break;
      if (curr_packet != NULL)
	/* Given op belongs more one frontier packet -- ignore it.  */
	continue;
#endif
      
      if ((insn_risky_speculative_move_p (curr_el->op->insn, packet)
	   && !postdom_p (curr_el->op->packet, packet))
	  /* Don't make such move of branch.  Sometimes it results in
             quick increase of insn copy number.  But it is ignored
             for an exit packet because such packets are achieved from
             different iterations and dominator constraint is too
             strict.  */
	  || (!packet->exit_packet_p
	      && (uncond_branch_p (curr_el->op->insn)
		  || cond_branch_p (curr_el->op->insn))
	      && !dom_p (packet, curr_el->op->packet)))
	continue;
      if (curr_el->delay > 0)
	{
	  *no_more_candidate_op_p = FALSE;
	  return NULL;
	}
      memcpy (temp_cpu_state, packet->cpu_state, cpu_state_length);
      if (!cpu_state_issue (curr_el->op->insn, temp_cpu_state))
	*no_more_candidate_op_p = FALSE;
      else
	{
	  int op_can_be_moved_p = TRUE;
	  op_t curr_op;
	  
	  if (!cond_branch_p (curr_el->op->insn)
	      && !uncond_branch_p (curr_el->op->insn))
	    for (curr_op = packet->first_op;
		 curr_op != NULL;
		 curr_op = curr_op->next_op)
	      /* ??? Currently gcc (gcc rtl) does not support muti-way
                 branch architectures.  */
	      if (uncond_branch_p (curr_op->insn)
		  || cond_branch_p (curr_op->insn))
		{
		  op_can_be_moved_p = FALSE;
		  break;
		}

	  if (cond_branch_p (curr_el->op->insn)
	      /*|| uncond_branch_p (curr_el->op->insn)*/)
	    for (curr_op = packet->first_op;
		 curr_op != NULL;
		 curr_op = curr_op->next_op)
	      /* ??? Currently we prohibit more one branch in a
		 packet.  Some processors can issue more one branch
		 per cycle.  So we could implement this because we
		 have some infrastructure for this see members and
		 comments for struct op.  The implementation should be
		 depended on semantics of such case (the first branch
		 is taken or the last one).  */
	      if (uncond_branch_p (curr_op->insn)
		  || cond_branch_p (curr_op->insn))
		{
		  op_can_be_moved_p = FALSE;
		  break;
		}
	  if (op_can_be_moved_p && uncond_branch_p (curr_el->op->insn)
	      && curr_el->op->packet->succ[0]->pred [1] != NULL
	      && curr_el->op->packet->pred [0] != packet)
	    op_can_be_moved_p = FALSE;
	  if (op_can_be_moved_p)
	    for (pred_number = 0;
		 (pred = curr_el->op->packet->pred [pred_number]) != NULL;
		 pred_number++)
	      if (pred->succ[1] != NULL
		  && (pred->succ[0] == curr_el->op->packet->succ [0]
		      || pred->succ[1] == curr_el->op->packet->succ [0]
		      || (cond_branch_p (curr_el->op->insn)
			  &&
			  (pred->succ[0] == curr_el->op->packet->succ [1]
			   ||
			   pred->succ[1] == curr_el->op->packet->succ [1]))))
		{
		  /* We move an op through the cond branch and the
		     cond branch will refer only for one packet.
		     ??? We could remove the cond branch if it has no
		     side effects.  */
		  op_can_be_moved_p = FALSE;
		  break;
		}
	  if (op_can_be_moved_p)
	    {
	      *no_more_candidate_op_p = FALSE;
	      return curr_el;
	    }
	}
    }
  return NULL;
}

/* The following function makes available ops list from the ready ops
   list.  */
static void
make_packet_avail_list (packet)
     packet_t packet;
{
  int it_num, ready_time, cost;
  int inconsistence_p;
  op_t curr_op;

  /* Remove the current available ops list if it exists.  */
  avail_list_free (packet->avail_list);
  packet->avail_list = NULL;
  for (it_num = min_ready_op_it_num;
       it_num < ready_ops_size
	 && (packet->min_it_n < 0
	     || it_num < packet->min_it_n + iteration_look_ahead);
       it_num++)
    for (curr_op = ready_ops [it_num];
	 curr_op != NULL;
	 curr_op = curr_op->next_ready_op)
      {
	ready_time = get_op_ready_time (curr_op, packet);
	if (ready_time < 0)
	  continue;
	cost = move_cost (curr_op, packet, &inconsistence_p);
	if (cost < 0)
	  continue;
	if (!curr_op->packet->exit_packet_p
	    && (packet->min_it_n < 0 || it_num < packet->min_it_n))
	  packet->min_it_n = it_num;
	create_avail_list_el (packet, curr_op, TRUE,
			      cost, inconsistence_p, ready_time);
      }
}

#ifndef WRONG_ORIGINAL_RCSP

/* This recursive function processes packets (CURR_PACKET) achievable
   from PACKET and adds the packet ops to the available list of
   PACKET.  */
static void
add_achievable_ops (packet, curr_packet)
     packet_t packet, curr_packet;
{
  op_t op = curr_packet->first_op;
  packet_t succ;
  int succ_number;

  if (curr_packet->formed_packet_p || curr_packet->frontier_packet_p
      || packet->min_it_n < 0)
    abort ();
  if (op == NULL || op->next_op != NULL)
    abort ();
  if (op->it_num >= packet->min_it_n + iteration_look_ahead)
    return;
  create_avail_list_el (packet, op, FALSE, 0, 0, curr_time);
  for (succ_number = 0;
       (succ = curr_packet->succ [succ_number]) != NULL;
       succ_number++)
    add_achievable_ops (packet, succ);
}
#endif /* #ifndef WRONG_ORIGINAL_RCSP */

/* The following function fills up PACKET by operations from available
   list when functional units reservation is described by
   packet->cpu_state.  The function updates the available list and cpu
   state for PACKET.  F is output file for debugging information.
   PREFIX is prefix of each new line in debugging output.  */
static void
schedule_packet (packet, f, prefix)
     packet_t packet;
     FILE *f;
     const char *prefix;
{
  avail_list_t el;
  int added_p = FALSE, no_more_op;
  avail_list_t curr;
  int succ_number;
  packet_t succ;
  int inconsistence_p;

#ifdef RCSP_DEBUG
  if (rcsp_debug_level >= 2)
    fprintf (f, "\n%s\n%s scheduling packet %3d:",
	     prefix, prefix, packet->number);
  if (rcsp_debug_level >= 3)
    print_ready_ops (f, prefix);
#endif
  last_packet_visit_number4++;
  make_packet_avail_list (packet);
  for (;;)
    {
      packet->avail_list
	= avail_list_sort (packet->avail_list, avail_op_cmp);
#ifdef RCSP_DEBUG
      if (rcsp_debug_level >= 3)
	print_avail_list (f, packet->avail_list, prefix, FALSE);
#endif
      el = get_best_candidate_avail_op (packet, &no_more_op);
      if (el == NULL)
	{
	  if (no_more_op && !added_p)
	    {
	      if (first_frontier_packet == NULL)
		abort ();
	      /* We cannot place any op into the packet in this case.  */
	      break;
	    }
	  /* Advance time.  */
	  for (curr = packet->avail_list;
	       curr != NULL;
	       curr = curr->next_avail)
	    if (curr->delay != 0)
	      curr->delay--;
	  curr_time++;
	  cpu_state_advance_cycle (packet->cpu_state);
#ifdef RCSP_DEBUG
	  if (rcsp_debug_level >= 3)
	    fprintf (f, "\n%s   advancing time (time = %3d)",
		     prefix, curr_time);
#endif
	  if (added_p)
	    break;
	}
      else
	{
	  op_t op = el->op;

	  packet->issue_time = curr_time;
	  if (el->inconsistence_p)
	    {
	      make_move_consistent (op, packet);
	      if (move_cost (op, packet, &inconsistence_p) < 0
		  || inconsistence_p)
		/* After making CFG move consistent PACKET should be
                   able to be moved and move consistent */
		abort ();
	    }
	  schedule_op (op, packet);
	  if (packet->min_it_n < 0 || op->it_num < packet->min_it_n)
	    packet->min_it_n = op->it_num;
#if 0
	  /* ??? This unfinished code is for adding jumps during
             scheduling itself (see comments for add_consistent_jumps
             and convert_cond_branch).  */
	  if (cond_branch_p (op->insn))
	    /* We are trying to make it here too because a new jump
	       may be added into avail list and consequently into this
	       packet too.  */
	    add_consistent_jumps (packet);
#endif
	  op_become_scheduled (op);
	  /* Remove el from available list.  */
	  if (el->prev_avail == NULL)
	    packet->avail_list = el->next_avail;
	  else
	    el->prev_avail->next_avail = el->next_avail;
	  if (el->next_avail != NULL)
	    el->next_avail->prev_avail = el->prev_avail;
	  avail_list_el_free (el);
	  update_ready_avail_ops (op, TRUE);
	  if (!cpu_state_issue (op->insn, packet->cpu_state))
	    abort ();
#ifdef RCSP_DEBUG
	  if (rcsp_debug_level >= 3)
	    fprintf (f, "\n%s   move op (iter %3d, time = %3d) %s",
		     prefix, op->it_num, curr_time,
		     insn_name (op->insn, FALSE));
#endif
	  added_p = TRUE;
	}
    }
  packet->formed_packet_p = TRUE;
  if (packet->min_it_n < 0)
    abort ();
#ifndef WRONG_ORIGINAL_RCSP
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    add_achievable_ops (packet, succ);
  packet->avail_list = avail_list_sort (packet->avail_list, avail_op_cmp);
#ifdef RCSP_DEBUG
  if (rcsp_debug_level >= 3)
    print_avail_list (f, packet->avail_list, prefix, TRUE);
#endif
#endif /* #ifndef WRONG_ORIGINAL_RCSP */
}

/* The following recursive function traverses CFG starting with PACKET
   and makes all independent ops ready.  */
static void
add_independent_ready_ops (packet)
     packet_t packet;
{
  int succ_number;

  if (packet->visit_number1 == last_packet_visit_number1)
    return;
  packet->visit_number1 = last_packet_visit_number1;
  if (packet->first_op != NULL && packet->first_op->next_op != NULL)
    abort ();
  if (!packet->stop_packet_p
      && packet->first_op != NULL && packet->first_op->in_deps == NULL)
    op_become_ready (packet->first_op);
  for (succ_number = 0;
       packet->succ [succ_number] != NULL;
       succ_number++)
    add_independent_ready_ops (packet->succ [succ_number]);
}

/* The following recursive function traverses all achievable packets
   and marks them.  */
static void
mark_achievable_packets (packet)
     packet_t packet;
{
  int succ_number;

  if (packet->visit_number1 == last_packet_visit_number1)
    return;
  packet->visit_number1 = last_packet_visit_number1;
  for (succ_number = 0;
       packet->succ [succ_number] != NULL;
       succ_number++)
    mark_achievable_packets (packet->succ [succ_number]);
}

/* The following recursive function traverses PACKET and its
   successors and links all unachievable packets.  It is supposed that
   packets are marked by mark_achievable_packets before the function
   call.  It is important that all successors of packet finally will
   be after the packet in the list.  */
static packet_t
link_unachievable_packets (packet, first_packet)
     packet_t packet, first_packet;
{
  int succ_number;

  /* It is achievable packet or we already linked the packet.  */
  if (packet->visit_number1 == last_packet_visit_number1 - 1
      || packet->visit_number1 == last_packet_visit_number1)
    return first_packet;
  packet->visit_number1 = last_packet_visit_number1;
  if (packet->frontier_packet_p)
    abort ();
  for (succ_number = 0;
       packet->succ [succ_number] != NULL;
       succ_number++)
    first_packet = link_unachievable_packets (packet->succ [succ_number],
					      first_packet);
  packet->next_packet = first_packet;
  first_packet = packet;
  return first_packet;
}

/* The following function redirects packet FROM onto TO.  ROOT is a
   top packet from which all other packets are achievable.  F and
   PREFIX are used for debug output.  */
static void
redirect_packet (root, from, to, f, prefix)
     packet_t root, from, to;
     FILE *f;
     const char *prefix;
{
  dep_t curr_dep, next_dep;
  op_t curr_op, next_op;
  packet_t curr_packet, next_packet;

#ifdef RCSP_DEBUG
  if (rcsp_debug_level >= 2)
    fprintf (f, "\n%s\n%s changing packet %3d onto packet %3d:",
	     prefix, prefix, from->number, to->number);
#endif
  while (from->pred[0] != NULL)
    /* Remember: packet_change_dest removes the link from
       pred by shifting array `pred'.  */
    packet_change_dest (from->pred[0], from, to);
  last_packet_visit_number1++;
  mark_achievable_packets (root);
  last_packet_visit_number1++;
  link_unachievable_packets (from, NULL);
  for (curr_packet = from; curr_packet != NULL; curr_packet = next_packet)
    {
      next_packet = curr_packet->next_packet;
      while (curr_packet->succ[0] != NULL)
	/* Remember: packet_change_src removes the link from
	   succ by shifting array `succ'.  */
	packet_change_src (curr_packet->succ[0], curr_packet, NULL);
      /* Remove ops from packet.  Ops of packet may be in ready ops
	 list.  */
      for (curr_op = curr_packet->first_op; curr_op != NULL; curr_op = next_op)
	{
	  next_op = curr_op->next_op;
	  for (curr_dep = curr_op->out_deps;
	       curr_dep != NULL;
	       curr_dep = next_dep)
	    {
	      next_dep = curr_dep->next_out;
	      dep_delete (curr_dep);
	    }
	  for (curr_dep = curr_op->in_deps;
	       curr_dep != NULL;
	       curr_dep = next_dep)
	    {
	      next_dep = curr_dep->next_in;
	      dep_delete (curr_dep);
	    }
	  op_delete (curr_op);
	}
      curr_packet->first_op = curr_packet->last_op = NULL;
      packet_delete (curr_packet);
    }
}

/* The following recursive function traverses packets, processes
   packets containing conditional jumps, and adds unconditional jumps
   to make CFG consistent.  */
static void
add_jumps_4_cond_branches (packet)
     packet_t packet;
{
  int succ_number;
  packet_t succ;

  if (packet->visit_number5 == last_packet_visit_number5)
    return;
  packet->visit_number5 = last_packet_visit_number5;
  /* We use post order traversing here to process the last op whose
     successor is a loop exit as the first op and possibly remove
     additional jump.  */
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    add_jumps_4_cond_branches (succ);
  add_consistent_jumps (packet);
}

/* The following recursive function finds back edges in the formed
   loop and adds unconditional jumps for them if they are needed.  */
static void
add_jumps_on_back_edges (packet)
     packet_t packet;
{
  int succ_number;
  packet_t succ;

  packet->visit_number5 = last_packet_visit_number5;
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (succ->visit_number5 != last_packet_visit_number5)
      add_jumps_on_back_edges (succ);
    else if (succ->visit_number4 != last_packet_visit_number4)
      {
	/* packet->succ is a back edge.  */
	if (packet_find_branch (packet, succ, TRUE) == NULL)
	  /* Add jump on back edge.  */
	  (void) create_simple_jump (packet, succ, NULL);
      }
  packet->visit_number4 = last_packet_visit_number4;
}

/* The following recursive function traverses packets, processes
   packets adds the rest of unconditional jumps to make CFG
   consistent.  */
static void
add_the_rest_jumps (packet)
     packet_t packet;
{
  int succ_number;
  packet_t succ;

  if (packet->visit_number5 == last_packet_visit_number5)
    return;
  packet->visit_number5 = last_packet_visit_number5;
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    {
      add_the_rest_jumps (succ);
      if (needs_jump_p (packet, succ))
	(void) create_simple_jump (packet, succ, NULL);
    }
}

/* The following recursive function may add jump onto main exit
   packet.  This jump is necessary to guarantee that the main exit
   will be the last packet during outputting packets.  It is supposed
   that for each packet the first output packet will be fall-through
   packet then all jumps which will be output in the same order as
   they are presented in array succ of PACKET.  */
static void
add_exit_jump (packet, there_are_unprocessed_packets_p)
     packet_t packet;
     int there_are_unprocessed_packets_p;
{
  packet_t succ, next_pc_packet;
  int succ_number, unvisited_n;

  if (packet->visit_number5 == last_packet_visit_number5)
    return;
  packet->visit_number5 = last_packet_visit_number5;
  unvisited_n = 0;
  next_pc_packet = NULL;
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (packet_find_branch (packet, succ, TRUE) == NULL)
      next_pc_packet = succ;
    else if (succ->visit_number5 != last_packet_visit_number5)
      unvisited_n++;
  if (next_pc_packet != NULL)
    {
      there_are_unprocessed_packets_p
	= there_are_unprocessed_packets_p || unvisited_n != 0;
      if (next_pc_packet->stop_packet_p && next_pc_packet->main_exit_p
	  && there_are_unprocessed_packets_p)
	{
	  (void) create_simple_jump (packet, next_pc_packet, NULL);
	  return;
	}
      add_exit_jump (next_pc_packet, there_are_unprocessed_packets_p);
    }
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (succ != next_pc_packet)
      {
	unvisited_n--;
	add_exit_jump (succ,
		       there_are_unprocessed_packets_p || unvisited_n != 0);
      }
}

/* The following function adds unconditional jumps in order to make
   code linear.  ??? If we do not implement adding and scheduling
   unconditional jumps on the fly in the future, we should simplify
   this code by removing functions add_jumps_4_cond_branches and
   add_consistent_jumps.  */
static void
add_jumps (packet)
     packet_t packet;
{
  last_packet_visit_number5++;
  add_jumps_4_cond_branches (packet);
  last_packet_visit_number4++;
  last_packet_visit_number5++;
  add_jumps_on_back_edges (packet);
  last_packet_visit_number5++;
  add_the_rest_jumps (packet);
  last_packet_visit_number5++;
  add_exit_jump (packet, FALSE);
}

#ifdef RCSP_DEBUG

/* The following recursive function checks consistence of CFG starting
   with PACKET.  */
static void
check_CFG (packet)
     packet_t packet;
{
  int pred_number, succ_number;
  packet_t pred, succ;
  op_t curr_op;

  if (packet->visit_number5 == last_packet_visit_number5)
    return;
  packet->visit_number5 = last_packet_visit_number5;
  if (packet->formed_packet_p && packet->frontier_packet_p)
    abort ();
  /* Check consistence of pred and succ: */
  for (pred_number = 0;
       (pred = packet->pred [pred_number]) != NULL;
       pred_number++)
    if (packet_find_in_arr (pred->succ, packet) == NULL)
      abort ();
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (packet_find_in_arr (succ->pred, packet) == NULL)
      abort ();
  for (curr_op = packet->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    if (curr_op->false_target != NULL
	&& packet_find_in_arr (packet->succ, curr_op->false_target) == NULL)
      abort ();
    else if (curr_op->true_target != NULL
	     && packet_find_in_arr (packet->succ,
				    curr_op->true_target) == NULL)
      abort ();
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    check_CFG (succ);
}

/* The following recursive function checks consistence of DDG starting
   with PACKET.  */
static void
check_DDG (packet)
     packet_t packet;
{
  op_t curr_op;
  dep_t curr_dep;
  int succ_number;
  packet_t succ;

  if (packet->visit_number5 == last_packet_visit_number5)
    return;
  packet->visit_number5 = last_packet_visit_number5;
  /* Check DDG.  */
  for (curr_op = packet->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    {
      for (curr_dep = curr_op->in_deps;
	   curr_dep != NULL;
	   curr_dep = curr_dep->next_in)
	if (curr_dep->to->packet != packet
	    && ACHIEVABLE (curr_dep->from->packet, packet) == UNACHIEVABLE)
	  abort ();
      for (curr_dep = curr_op->out_deps;
	   curr_dep != NULL;
	   curr_dep = curr_dep->next_out)
	if (curr_dep->from->packet != packet
	    && ACHIEVABLE (packet, curr_dep->to->packet) == UNACHIEVABLE)
	  abort ();
    }
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    check_DDG (succ);
}

/* The following function checks consistence of CFG and DDG starting
   with packet START.  */
static void
check_graphs (start)
     packet_t start;
{
  last_packet_visit_number5++;
  check_CFG (start);
  last_packet_visit_number5++;
  check_DDG (start);
}

/* The following function outputs all ready ops into file F with
   output line PREFIX.  */
static void
print_ready_ops (f, prefix)
     FILE *f;
     const char *prefix;
{
  int it_num;
  op_t curr_op;
  
  for (it_num = min_ready_op_it_num;
       it_num < ready_ops_size;
       it_num++)
    {
      if (ready_ops [it_num] != NULL)
	fprintf (f, "\n%s    ready list (iter = %3d):", prefix, it_num);
      for (curr_op = ready_ops [it_num];
	   curr_op != NULL;
	   curr_op = curr_op->next_ready_op)
	fprintf (f, "\n%s       packet = %3d, op %s",
		 prefix, curr_op->packet->number,
		 insn_name (curr_op->insn, FALSE));
    }
}

/* The following function outputs AVAIL_LIST into file F with output
   line PREFIX.  IF FULL_P is TRUE, the output is believed to contain
   also achievable ops.  */
static void
print_avail_list (f, avail_list, prefix, full_p)
     FILE *f;
     avail_list_t avail_list;
     const char *prefix;
     int full_p;
{
  avail_list_t curr_el;
  
  fprintf (f, (full_p ? "\n%s    available & achievable list:"
	       : "\n%s    available list:"), prefix);
  for (curr_el = avail_list; curr_el != NULL; curr_el = curr_el->next_avail)
    {
      fprintf (f, "\n%s       ", prefix);
      if (curr_el->avail_p)
	fprintf (f, "delay = %3d, cost = %3d (%s) ",
		 curr_el->delay, curr_el->cost,
		 (curr_el->inconsistence_p ? "I" : "C"));
      fprintf (f, "packet = %3d, iter = %3d, %sop %s",
	       curr_el->op->packet->number, curr_el->op->it_num,
	       (curr_el->avail_p ? "" : "achievable "),
	       insn_name (curr_el->op->insn, FALSE));
    }
}

/* The following macro value is constraint on maximal length of packet
   name.  */
#define PACKET_NAME_LONG 200

/* The following function returns symbolic name of PACKET.  If GRAPH_P
   is TRUE the name will be shorter.  The result is stored in static
   memory.  So the next call will rewrite the memory.  */
static const char *
get_packet_name (packet, graph_p)
     packet_t packet;
     int graph_p;
{
  static char name [PACKET_NAME_LONG];
  const char *prefix, *postfix;

  if (packet->formed_packet_p)
    prefix = (graph_p ? "" : "formed ");
  else if (packet->frontier_packet_p)
    prefix = (graph_p ? "F" : "frontier ");
  else
    prefix = (graph_p ? "I" : "");
  if (packet->stop_packet_p)
    postfix = (graph_p ? "S" : " (stop packet)");
  else if (packet->exit_packet_p)
    postfix = (graph_p ? "E" : " (exit packet)");
  else
    postfix = "";
  if (graph_p)
    sprintf (name, "\"%s%s %d:", prefix, postfix, packet->number);
  else
    sprintf (name, "%spacket %4d%s", prefix, packet->number, postfix);
  if (graph_p)
    {
      op_t curr_op;
      char *out = name;

      out += strlen (out);
      for (curr_op = packet->first_op;
	   curr_op != NULL;
	   curr_op = curr_op->next_op)
	{
	  out += strlen (out);
#if 0
	  if (curr_op == packet->first_op)
	    {
	      *out++ = '\\';
	      *out++ = 'n';
	    }
#endif
	  sprintf (out," %s(%d)", insn_name (curr_op->insn, TRUE),
		   curr_op->it_num);
	}
      out += strlen (out);
      strcpy (out, "\"");
    }
  if (strlen (name) >= PACKET_NAME_LONG)
    abort ();
  return name;
}

/* The following function prints information about PACKET into file F.
   PREFIX is output after each newline.  If GRAPH_P is TRUE the output
   is in a graphic vizualization program format.  */
static void
print_packet (f, packet, prefix, graph_p)
     FILE *f;
     packet_t packet;
     const char *prefix;
     int graph_p;
{
  char packet_name [PACKET_NAME_LONG];
  int pred_number, succ_number;
  packet_t pred, succ;
  op_t curr_op;
  dep_t curr_dep;

  strcpy (packet_name, get_packet_name (packet, graph_p));
  if (graph_p)
    fprintf (f, "\n  %s;", packet_name);
  else
    {
      fprintf (f, "\n%s%s", prefix, packet_name);
      fprintf (f, ":\n%s  packet predecessors: ", prefix);
      for (pred_number = 0;
	   (pred = packet->pred [pred_number]) != NULL;
	   pred_number++)
	fprintf (f, " %4d", pred->number);
      fprintf (f, "\n%s  packet successors: ", prefix);
    }
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    if (graph_p)
      fprintf (f, "\n  %s -> %s;", packet_name,
	       get_packet_name (succ, graph_p));
    else
      fprintf (f, " %4d", succ->number);
  if (!graph_p)
    {
      fprintf (f, "\n%s  packet operations:\n%s", prefix, prefix);
      for (curr_op = packet->first_op;
	   curr_op != NULL;
	   curr_op = curr_op->next_op)
	{
	  fprintf (f, "  iter = %3d op = %s", curr_op->it_num,
		   insn_name (curr_op->insn, FALSE));
	  fprintf (f, "\n%s      input dependencies:", prefix);
	  for (curr_dep = curr_op->in_deps;
	       curr_dep != NULL;
	       curr_dep = curr_dep->next_in)
	    fprintf
	      (f, "\n%s        %s (%3d) : packet %3d, iter = %3d, op = %s",
	       prefix,
	       (curr_dep->dep_mode == DEF_USE
		? "DEF-USE"
		: (curr_dep->dep_mode == DEF_DEF ? "DEF-DEF" : "USE-DEF")),
	       curr_dep->latency, curr_dep->from->packet->number,
	       curr_dep->from->it_num,
	       insn_name (curr_dep->from->insn, FALSE)); 
	  fprintf (f, "\n%s      output dependencies:", prefix);
	  for (curr_dep = curr_op->out_deps;
	       curr_dep != NULL;
	       curr_dep = curr_dep->next_out)
	    fprintf
	      (f, "\n%s        %s (%3d) : packet %3d, iter = %3d, op = %s",
	       prefix,
	       (curr_dep->dep_mode == DEF_USE
		? "DEF-USE"
		: (curr_dep->dep_mode == DEF_DEF ? "DEF-DEF" : "USE-DEF")),
	       curr_dep->latency, curr_dep->to->packet->number,
	       curr_dep->to->it_num, insn_name (curr_dep->to->insn, FALSE)); 
	  fprintf (f, "\n%s", prefix);
	}
    }
}

/* The following recursive function prints CFG and DDG starting with
   PACKET into file F.  PREFIX is output after each newline.  If
   GRAPH_P is TRUE the output is in a graphic visualization program
   format.  */
static void
print_packets_1 (f, packet, prefix, graph_p)
     FILE *f;
     packet_t packet;
     const char *prefix;
     int graph_p;
{
  int succ_number;
  packet_t succ;

  if (packet->visit_number5 == last_packet_visit_number5)
    return;
  packet->visit_number5 = last_packet_visit_number5;
  print_packet (f, packet, prefix, graph_p);
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    print_packets_1 (f, succ, prefix, graph_p);
}

/* The following function prints CFG and DDG starting with PACKET into
   file F.  PREFIX is output after each newline.  If GRAPH_P is TRUE
   the output is in a graphic visualization program format.  */
static void
print_packets (f, start, prefix, graph_p)
     FILE *f;
     packet_t start;
     const char *prefix;
     int graph_p;
{
  last_packet_visit_number5++;
  if (graph_p)
    {
      fprintf (f, "\ndigraph CFG {\n");
      fprintf (f, "  node [shape=box, fontsize=200];\n");
      fprintf (f, "  ratio=fill;\n");
      fprintf (f, "  page = \"8.5, 11\"; // inches\n");
      fprintf (f, "  size = \"7.5, 10\"; // inches\n");
    }
  print_packets_1 (f, start, prefix, graph_p);
  if (graph_p)
    fprintf (f, "\n}\n");
}

#endif /* #ifdef RCSP_DEBUG */

/* This entry function makes software pipelining of loop body started
   with START_PACKET.  The function returns TRUE if loop has been
   pipelined successfully.  DEBUG_FILE and PREFIX are used for
   debugging output.  */
static packet_t
pipeline (start_packet, debug_file, prefix)
     packet_t start_packet;
     FILE *debug_file;
     const char *prefix;
{
  packet_set_t scheduled_before;
  packet_t r, n, scheduled_packet, pred;
  int ok_p;
  int loop_ops_number;
  op_t branch ATTRIBUTE_UNUSED;
  int pred_number ATTRIBUTE_UNUSED;

  if (iteration_look_ahead <= 0 || RCSP_UNROLL_FACTOR <= 1
      || RCSP_EXPANSION_LOOP_FACTOR <= 1)
    abort ();
  last_packet_visit_number1 = 0;
  last_packet_visit_number2 = 0;
  last_packet_visit_number3 = 0;
  last_packet_visit_number4 = 0;
  last_packet_visit_number5 = 0;
  loop_ops_number
    = packets_number / (RCSP_UNROLL_FACTOR * iteration_look_ahead + 1);
  if (loop_ops_number <= 0)
    abort ();
  scheduled_before = packet_set_create ();
  curr_time = 0;
  r = packet_create (curr_time);
  packet_connect (r, start_packet, FALSE, FALSE);
  build_deps (r);
  top_packet->formed_packet_p = TRUE;
  update_ready_avail_ops (top_packet->first_op, FALSE);
  last_packet_visit_number1++;
  add_independent_ready_ops (r);
  cpu_state_reset (r->cpu_state);
#ifdef RCSP_DEBUG
  if (rcsp_debug_level >= 1)
    {
      fprintf (debug_file, "\n%s\n%s ORIGINAL LOOP:\n%s",
	       prefix, prefix, prefix);
      print_packets (debug_file, r, prefix, FALSE);
      check_graphs (r);
    }
#endif
  add_frontier_packet (r);
  while (first_frontier_packet != NULL)
    {
      frontier_packet_list_sort_and_merge (debug_file, prefix);
      n = frontier_packet_list_head_remove ();
      curr_time = n->issue_time;
      schedule_packet (n, debug_file, prefix);
      scheduled_packet = packet_set_insert (scheduled_before, n);
      pred = NULL;
#if 0
      /* ??? This unfinished code is trying to redirect only packets
         which are predecessors of branches.  As consequence it
         results in less added jumps.  Is it worth to do? */
      if (scheduled_packet != n)
	/* There is already such packet.  Can we redirect pred
           packets? */
	for (pred_number = 0;
	     (pred = n->pred [pred_number]) != NULL;
	     pred_number++)
	  {
	    branch = packet_find_branch (pred, n, FALSE);
	    if (branch == NULL
		|| (branch->true_target == n && !branch->true_jump_p)
		|| (branch->false_target == n && branch->true_jump_p))
	      break;
	  }
#endif
      if (scheduled_packet == n || pred != NULL)
	{
	  add_next_packets (n);
	  /* Pipelined loop is already too big.  */
	  if (scheduled_ops_number
	      > RCSP_EXPANSION_LOOP_FACTOR * loop_ops_number)
	    break;
	  if (n->min_it_n >= (RCSP_UNROLL_FACTOR - 1) * iteration_look_ahead)
	    /* The conditional branch of the last iteration which is
               backward jump will have only one successor.  Therefore
               we have one more iteration after the lookahead.  No
               repeated patterns -- do not process other frontier
               packets.  */
	    break;
	}
      else
	redirect_packet (r, n, scheduled_packet, debug_file, prefix);
#ifdef RCSP_DEBUG
      if (rcsp_debug_level >= 1)
	check_graphs (r);
#endif
    }
  ok_p = first_frontier_packet == NULL;
  if (ok_p)
    add_jumps (r);
#ifdef RCSP_DEBUG
  if (rcsp_debug_level >= 1)
    {
      if (!ok_p)
	fprintf (debug_file, "\n%s\n%s THE LOOP HAS NOT BEEN PIPELINED!\n%s\n",
		 prefix, prefix, prefix);
      else
	{
	  fprintf (debug_file, "\n%s\n%s FINAL PIPELINED LOOP:\n%s\n",
		   prefix, prefix, prefix);
	  print_packets (debug_file, r, prefix, FALSE);
	}
    }
#endif
  packet_set_delete (scheduled_before);
  return (ok_p ? r : NULL);
}

/* ============ End of RCSP software pipeliner. =================== */



#ifndef RCSP_DO_NOT_USE_OBSTACK

#include "obstack.h"

#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free

#endif

/* This page contains functions and data needed for interface of
   the software pipeliner.  */

/* The following array contains starts of chains of locations
   referenced by the same insn.  */
static struct loc **insn_locs;

/* The following variable is length of the previous array.  */
static int insn_locs_length;

/* This structure is an element of the list which contains references
   for software pipeliner locations.  */
struct loc_link
{
  struct loc *loc;
  struct loc_link *next;
};

/* This structure is reference for a software pipeliner location.  */
struct loc
{
  /* Rtx representation of the location.  */
  rtx loc;
  /* Flag of that the location is used.  */
  int usage_p;
  /* Type of writing if the location is written.  */
  int write_kind;
  /* All the structures which correspond to hard registers or memory are
     chained by the following members.  */
  struct loc *next_hard_reg;
  struct loc *next_mem;
  /* All location references which are in the same insn are chained by
     the following member.  */
  struct loc *next_insn_loc;
  /* Links to locations which correspondingly contain the location and
     are contained in the location.  */
  struct loc_link *supers;
  struct loc_link *subs;
};

/* This structure is a software pipeliner insn.  */
struct insn
{
  /* The order number (0, 1, ...) of software pipeliner insn.  */
  int number;
  /* Source file and line of the insn.  The value of `file' is NULL if
     source file and line is not known.  This is always NULL for dummy
     insn.  */
  char *file;
  int line;
  /* Original insn.  Corresponding insns from different iterations
     have the same member value.  The value is NULL only for dummy
     insn.  */
  rtx orig_insn;
  /* Output rtx representation of the insn.  It means that the
     corresponding insns from different iterations may be different
     (e.g. register in insn may be renamed).  The value is NULL only
     for dummy insn.  */
  rtx insn;
  /* Notes (and labels) which should be output before and after the
     insn.  */
  rtx notes_before;
  rtx notes_after;
  /* The following member value is code label of exit label of the
     loop. The value is not NULL only for stop packets which
     correspond exit jumps from the loop.  */
  rtx exit_label;
  /* The following member value is TRUE if the insn sets up a register
     living after the loop.  */
  int live_reg_set_p;
};

/* The following variable contains array of all software pipeliner
   insns.  Index corresponds to order numbers of insns.  */
static varray_type insns;

/* The following variables contain arrays of locations which are read,
   may be written, are written, and are write barriers.  */
static varray_type insn_reads;
static varray_type insn_possible_writes;
static varray_type insn_write_barriers;
static varray_type insn_writes;

/* The following variable values are actual lengths of the
   corresponding arrays.  */
static int insns_length;
static int insn_reads_length;
static int insn_possible_writes_length;
static int insn_write_barriers_length;
static int insn_writes_length;

/* The following function adds pointer EL to varray of pointers VECTOR
   which already contains *LENGTH elements.  Then *LENGTH is
   incremented.  */
static void
add_vector_element (vector, el, length)
     varray_type *vector;
     void *el;
     int *length;
{
  unsigned size = VARRAY_SIZE (*vector);

  if ((unsigned) *length >= size)
    {
      size *= 2;
      if (size < (unsigned) *length)
	size = (unsigned) *length;
      VARRAY_GROW (*vector, size);
    }
  VARRAY_GENERIC_PTR (*vector, *length) = el;
  (*length)++;
}

/* The following variables refer for all created locations which are a
   hard register and a memory correspondingly.  */
static struct loc *hard_reg_loc_chain, *mem_loc_chain;

/* The following variable refers for location corresponding to all
   memory.  */
static struct loc *all_memory_loc;

/* The following function creates and returns new location link
   structure with given member values LOC and NEXT.  */
static struct loc_link *
new_loc_link (loc, next)
     struct loc *loc;
     struct loc_link *next;
{
  struct loc_link *result = rcsp_malloc (sizeof (struct loc_link));

  result->loc = loc;
  result->next = next;
  return result;
}

/* The following function creates and returns new location structure
   with given member values LOC, USAGE_P, WRITE_KIND, and
   NEXT_INSN_LOC.  */
static struct loc *
new_loc (loc, usage_p, write_kind, next_insn_loc)
     rtx loc;
     int usage_p, write_kind;
     struct loc *next_insn_loc;
{
  struct loc *result = rcsp_malloc (sizeof (struct loc));

  result->loc = loc;
  result->usage_p = usage_p;
  result->write_kind = write_kind;
  result->next_insn_loc = next_insn_loc;
  result->supers = result->subs = NULL;
  return result;
}

/* The following function creates and returns new location structure
   with given member values LOC, USAGE_P, and WRITE_KIND.  The new
   location is added to the list locations of INSN and lists of hard
   register and memory locations if it is applicable to LOC.  */
static struct loc *
new_insn_loc (insn, loc, usage_p, write_kind)
     insn_t insn;
     rtx loc;
     int usage_p, write_kind;
{
  struct loc *result;

  result = new_loc (loc, usage_p, write_kind, insn_locs [insn->number]);
  insn_locs [insn->number] = result;
  if (GET_CODE (loc) == REG && REGNO (loc) < FIRST_PSEUDO_REGISTER)
    {
      result->next_hard_reg = hard_reg_loc_chain;
      hard_reg_loc_chain = result;
    }
  else if (GET_CODE (loc) == MEM)
    {
      result->next_mem = mem_loc_chain;
      mem_loc_chain = result;
    }
  return result;
}

/* The following function frees location LOC and loc links of the
   locations.  */
static void
free_loc (loc)
     struct loc *loc;
{
  struct loc_link *curr_link, *next_link; 

  for (curr_link = loc->supers; curr_link != NULL; curr_link = next_link)
    {
      next_link = curr_link->next;
      rcsp_free (curr_link);
    }
  for (curr_link = loc->subs; curr_link != NULL; curr_link = next_link)
    {
      next_link = curr_link->next;
      rcsp_free (curr_link);
    }
  rcsp_free (loc);
}

/* The following function frees chain of locations starting with
   LOC.  */
static void
free_insn_locs (loc)
     struct loc *loc;
{
  struct loc *next_loc;

  while (loc != NULL)
    {
      next_loc = loc->next_insn_loc;
      free_loc (loc);
      loc = next_loc;
    }
}

/* The following function processes a single SET, CLOBBER, PRE_DEC,
   POST_DEC, PRE_INC or POST_INC rtx, X, setting up locations of INSN.  */
static void
process_writes (x, insn)
     rtx x;
     insn_t insn;
{
  register int regno;
  register rtx dest = XEXP (x, 0);
  enum rtx_code code = GET_CODE (x);

  if (dest == 0)
    return;

  if (GET_CODE (dest) == PARALLEL
      && GET_MODE (dest) == BLKmode)
    {
      register int i;
      for (i = XVECLEN (dest, 0) - 1; i >= 0; i--)
	process_writes (XVECEXP (dest, 0, i), insn);
      if (GET_CODE (x) == SET)
	process_reads (SET_SRC (x), insn);
      return;
    }

  while (GET_CODE (dest) == STRICT_LOW_PART || GET_CODE (dest) == SUBREG
      || GET_CODE (dest) == ZERO_EXTRACT || GET_CODE (dest) == SIGN_EXTRACT)
    {
      if (GET_CODE (dest) == ZERO_EXTRACT || GET_CODE (dest) == SIGN_EXTRACT)
	{
	  /* The second and third arguments are values read by this insn.  */
	  process_reads (XEXP (dest, 1), insn);
	  process_reads (XEXP (dest, 2), insn);
	}
      dest = XEXP (dest, 0);
    }

  if (GET_CODE (dest) == REG)
    {
      register int i;

      regno = REGNO (dest);

      /* A hard reg in a wide mode may really be multiple registers.
         If so, mark all of them just like the first.  */
      if (regno < FIRST_PSEUDO_REGISTER)
	{
	  i = HARD_REGNO_NREGS (regno, GET_MODE (dest));
	  while (--i >= 0)
	    {
	      rtx reg = gen_rtx_REG (VOIDmode, regno + i);

	      /* Clobbers need not be ordered with respect to one
		 another, but sets must be ordered with respect to a
		 pending clobber.  */
	      if (code == SET)
		new_insn_loc (insn, reg, 0 /* usage_p */, 0 /* usual write */);
	      else
		new_insn_loc (insn, reg, 0 /* usage_p */, 1 /* may write */);
	    }
	}
      else
	{
	  if (code == SET)
	    new_insn_loc (insn, dest, 0 /* usage_p */, 0 /* usual write */);
	  else
	    new_insn_loc (insn, dest, 0 /* usage_p */, 1 /* may write */);

	  /* Pseudos that are REG_EQUIV to something may be replaced
	     by that during reloading.  We need only add dependencies for
	     the address in the REG_EQUIV note.  */
	  if (!reload_completed
	      && reg_known_equiv_p[regno]
	      && GET_CODE (reg_known_value[regno]) == MEM)
	    process_reads (XEXP (reg_known_value[regno], 0), insn);
	}
    }
  else if (GET_CODE (dest) == MEM)
    {
      /* Writing memory.  */
      if (MEM_VOLATILE_P (x))
	abort ();
      new_insn_loc (insn, dest, 0 /* usage_p */, 0 /* usual write */);
      process_reads (XEXP (dest, 0), insn);
    }
  
  /* Analyze reads.  */
  if (GET_CODE (x) == SET)
    process_reads (SET_SRC (x), insn);
}

/* The following functions processes the uses of memory and registers
   in rtx X of INSN to set up insn's locations.  */
static void
process_reads (x, insn)
     rtx x;
     insn_t insn;
{
  register int i;
  register int j;
  register enum rtx_code code;
  register const char *fmt;

  if (x == 0)
    return;

  code = GET_CODE (x);

  switch (code)
    {
    case CONST_INT:
    case CONST_DOUBLE:
    case SYMBOL_REF:
    case CONST:
    case LABEL_REF:
      /* Ignore constants.  Note that we must handle CONST_DOUBLE here
         because it may have a cc0_rtx in its CONST_DOUBLE_CHAIN field, but
         this does not mean that this insn is using cc0.  */
      return;

#ifdef HAVE_cc0
    case CC0:
      /* ??? This code is not implemented because now we do not make
         pipeline for machines with cc0.  */
      new_insn_loc (insn, cc0_rtx, 1 /* usage_p */, 0);
      break;
#endif

    case REG:
      {
	int regno = REGNO (x);

	if (regno < FIRST_PSEUDO_REGISTER)
	  {
	    int i;

	    i = HARD_REGNO_NREGS (regno, GET_MODE (x));
	    while (--i >= 0)
	      new_insn_loc (insn, gen_rtx_REG (VOIDmode, regno + i),
			    1 /* usage_p */, 0);
	  }
	else
	  {
	    new_insn_loc (insn, x, 1 /* usage_p */, 0);

	    /* Pseudos that are REG_EQUIV to something may be replaced
	       by that during reloading.  We need only add dependencies for
	       the address in the REG_EQUIV note.  */
	    if (!reload_completed
		&& reg_known_equiv_p[regno]
		&& GET_CODE (reg_known_value[regno]) == MEM)
	      process_reads (XEXP (reg_known_value[regno], 0), insn);
	  }
	return;
      }

    case MEM:
      if (MEM_VOLATILE_P (x))
	abort ();
      new_insn_loc (insn, x, 1 /* usage_p */, 0);
      /* Take advantage of tail recursion here.  */
      process_reads (XEXP (x, 0), insn);
      return;
      
    /* Make memory barrier to in case a trap handler needs them.  */
    case TRAP_IF:
      new_insn_loc (insn, all_memory_loc->loc,
		    0 /* usage_p */, 2 /* write barrier */);
      break;

    case ASM_OPERANDS:
    case ASM_INPUT:
    case UNSPEC_VOLATILE:
      abort ();

    case PRE_DEC:
    case POST_DEC:
    case PRE_INC:
    case POST_INC:
      /* These both read and modify the result.  We must handle them as writes
         to get proper dependencies for following instructions.  We must handle
         them as reads to get proper dependencies from this to previous
         instructions.  Thus we need to pass them to both process_writes
         and process_reads.  */
      process_reads (XEXP (x, 0), insn);
      process_writes (x, insn);
      return;

    default:
      break;
    }

  /* Other cases: walk the insn.  */
  fmt = GET_RTX_FORMAT (code);
  for (i = GET_RTX_LENGTH (code) - 1; i >= 0; i--)
    {
      if (fmt[i] == 'e')
	process_reads (XEXP (x, i), insn);
      else if (fmt[i] == 'E')
	for (j = 0; j < XVECLEN (x, i); j++)
	  process_reads (XVECEXP (x, i, j), insn);
    }
}

/* The following function processes INSN to set up insn's
   locations.  */
static void
process_insn (insn)
     insn_t insn;
{
  rtx pattern;
  RTX_CODE code;
  int i;

  if (insn == NULL || insn->insn == NULL_RTX)
    /* Dummy op.  */
    abort ();
  pattern = PATTERN (insn->insn);
  code = GET_CODE (insn->insn);
  if (code == CALL
      || (code == NOTE
	  && (NOTE_LINE_NUMBER (insn->insn) == NOTE_INSN_LOOP_BEG
	      || NOTE_LINE_NUMBER (insn->insn) == NOTE_INSN_LOOP_END
	      || NOTE_LINE_NUMBER (insn->insn) == NOTE_INSN_EH_REGION_BEG
	      || NOTE_LINE_NUMBER (insn->insn) == NOTE_INSN_EH_REGION_END
	      || NOTE_LINE_NUMBER (insn->insn) == NOTE_INSN_SETJMP)))
    abort ();

  code = GET_CODE (pattern);
  if (code == COND_EXEC)
    {
      process_reads (COND_EXEC_TEST (pattern), insn);
      pattern = COND_EXEC_CODE (pattern);
      code = GET_CODE (pattern);
    }
  if (code == SET || code == CLOBBER)
    process_writes (pattern, insn);
  else if (code == PARALLEL)
    {
      for (i = XVECLEN (pattern, 0) - 1; i >= 0; i--)
	{
	  rtx sub = XVECEXP (pattern, 0, i);

	  code = GET_CODE (sub);
	  if (code == COND_EXEC)
	    {
	      process_reads (COND_EXEC_TEST (sub), insn);
	      sub = COND_EXEC_CODE (sub);
	      code = GET_CODE (sub);
	    }
	  if (code == SET || code == CLOBBER)
	    process_writes (sub, insn);
	  else
	    process_reads (sub, insn);
	}
    }
  else
    process_reads (pattern, insn);
}

/* The following function creates and returns new insn structure with
   rtx INSN with given attributes source FILE and source LINE.  */
static struct insn *
new_insn (insn, file, line)
     rtx insn;
     char *file;
     int line;
{
  struct insn *result = rcsp_malloc (sizeof (struct insn));
  
  result->number = insns_length;
  result->orig_insn = result->insn = insn;
  result->notes_before = result->notes_after = NULL_RTX;
  result->exit_label = NULL_RTX;
  result->live_reg_set_p = FALSE;
  result->file = file;
  result->line = line;
  add_vector_element (&insns, result, &insns_length);
  return result;
}

/* The following variable refers for the table which contains
   information about packets with jumps.  */
static htab_t jumps_table;

/* The structure describes a packet which contains jump.  The labels
   table entry refers for chain of packets which contain jump onto the
   same label.  */
struct jump_packet
{
  /* The following member value is LABEL onto which the jump is made.
     This is the key of jumps_table entry.  */
  rtx label;
  /* The packet which contains the jump.  */
  packet_t packet;
  /* The following member value is destination packet after connection
     of the packet with the destination packet.  */
  packet_t dest_packet;
  /* The next jump packet with the same label code.  */
  struct jump_packet *next_jump_packet;
  /* The following member is TRUE if we should connect the packet with
     label as jump, otherwise as fall through.  */
  char jump_p;
};

/* The following function returns hash code of JUMP_PACKET.  */
static unsigned
jump_packet_hash (jump_packet)
     const void *jump_packet;
{
  return (unsigned) ((struct jump_packet *) jump_packet)->label;
}

/* The following function returns TRUE if JUMP_PACKET1 and
   JUMP_PACKET2 refer for the same label.  */
static int
jump_packet_eq (jump_packet1, jump_packet2)
     const void *jump_packet1, *jump_packet2;
{
  return (((struct jump_packet *) jump_packet1)->label
	  == ((struct jump_packet *) jump_packet2)->label);
}

/* The following function frees jump_packets in chains starting with
   JUMP_PACKET.  */
static void
jump_packet_free (jump_packet)
     void *jump_packet;
{
  struct jump_packet *curr_jump_packet, *next_jump_packet;

  for (curr_jump_packet = jump_packet;
       curr_jump_packet != NULL;
       curr_jump_packet = next_jump_packet)
    {
      next_jump_packet = curr_jump_packet->next_jump_packet;
      rcsp_free (curr_jump_packet);
    }
}

/* The following function returns information from the label table
   about packets which contains jump onto LABEL.  */
static struct jump_packet *
find_jump_packet (label)
     rtx label;
{
  struct jump_packet temp;

  temp.label = label;
  return htab_find (jumps_table, &temp);
}

/* The following function inserts LABEL and PACKET which contains jump
   onto the label into the labels table.  */
static void
insert_jump_packet (label, packet, jump_p)
     rtx label;
     packet_t packet;
     int jump_p;
{
  void **slot;
  struct jump_packet *jump_packet;

  jump_packet
    = (struct jump_packet *) rcsp_malloc (sizeof (struct jump_packet));
  jump_packet->label = label;
  jump_packet->packet = packet;
  jump_packet->dest_packet = FALSE;
  jump_packet->jump_p = jump_p;
  slot = htab_find_slot (jumps_table, jump_packet, TRUE);
  jump_packet->next_jump_packet = *slot;
  *slot = jump_packet;
}

/* The following function call FUNC for all jump labels table entries
   with additional PARAM.  */
static void
traverse_jumps_table (func, param)
     int (*func) PARAMS ((void **, void *));
     void *param;
{
  htab_traverse (jumps_table, func, param);
}

/* The following function starts work with the jump labels table.  */
static void
initiate_jumps_table ()
{
  jumps_table = htab_create (100, jump_packet_hash, jump_packet_eq,
			     jump_packet_free);
}

/* The following function finishes work with the jump labels table.  */
static void
finish_jumps_table ()
{
  htab_delete (jumps_table);
}

/* The following variable contains registers which live after the
   loop.  */
static bitmap global_regs_living_after_loop;

/* The following function adds global registers living at the start of
   bb containing BB_INSN to set of register living after the loop.  */
static void
update_regs_living_after_loop (bb_insn)
     rtx bb_insn;
{
  for (;;)
    {
      if (bb_insn == NULL
	  || (GET_CODE (bb_insn) == NOTE
	      && NOTE_LINE_NUMBER (bb_insn) == NOTE_INSN_BASIC_BLOCK))
	break;
      bb_insn = NEXT_INSN (bb_insn);
    }
  if (bb_insn != NULL_RTX)
    {
      basic_block bb = NOTE_BASIC_BLOCK (bb_insn);

      IOR_REG_SET(global_regs_living_after_loop, bb->global_live_at_start);
    }
}

/* The following recursive function returns TRUE if X sets up a register
   living after the loop.  */
static int
rtx_sets_live_reg (x)
     rtx x;
{
  int regno, i;
  rtx dest = XEXP (x, 0);

  if (dest == NULL_RTX)
    return FALSE;

  if (GET_CODE (dest) == PARALLEL && GET_MODE (dest) == BLKmode)
    {
      for (i = XVECLEN (dest, 0) - 1; i >= 0; i--)
	if (rtx_sets_live_reg (XVECEXP (dest, 0, i)))
	  return TRUE;
      return FALSE;
    }

  while (GET_CODE (dest) == STRICT_LOW_PART || GET_CODE (dest) == SUBREG
      || GET_CODE (dest) == ZERO_EXTRACT || GET_CODE (dest) == SIGN_EXTRACT)
    dest = XEXP (dest, 0);
  
  if (GET_CODE (dest) == REG)
    {
      regno = REGNO (dest);

      /* A hard reg in a wide mode may really be multiple registers.
         If so, mark all of them just like the first.  */
      if (regno < FIRST_PSEUDO_REGISTER)
	{
	  i = HARD_REGNO_NREGS (regno, GET_MODE (dest));
	  while (--i >= 0)
	    if (REGNO_REG_SET_P (global_regs_living_after_loop, regno + i))
	      return TRUE;
	}
      else if (REGNO_REG_SET_P (global_regs_living_after_loop, regno))
	return TRUE;
    }
  return FALSE;
}

/* The following function returns TRUE if INSN sets up a register
   living after the loop.  */
static int
insn_sets_live_reg_p (insn)
     rtx insn;
{
  rtx pattern;
  RTX_CODE code;
  int i;

  if (insn == NULL_RTX)
    /* Dummy op.  */
    abort ();
  pattern = PATTERN (insn);
  code = GET_CODE (pattern);
  if (code == SET || code == CLOBBER)
    return rtx_sets_live_reg (pattern);
  else if (code == PARALLEL)
    {
      for (i = XVECLEN (pattern, 0) - 1; i >= 0; i--)
	{
	  code = GET_CODE (XVECEXP (pattern, 0, i));
	  if ((code == SET || code == CLOBBER)
	      && rtx_sets_live_reg (XVECEXP (pattern, 0, i)))
	    return TRUE;
	}
    }
  return FALSE;
}

/* The following function connects packets by backward jumps and for
   loop exits and update live registers after the loop.  START_LABEL
   is label of the loop entry.  */
static int
second_jumps_pass (jump_packet_ptr, start_label)
     void **jump_packet_ptr;
     void *start_label;
{
  struct jump_packet *curr_jump_packet;
  packet_t dest_packet;
  rtx start = start_label;

  dest_packet = NULL;
  /* Find destination packet.  */
  for (curr_jump_packet = *jump_packet_ptr;
       curr_jump_packet != NULL && curr_jump_packet->dest_packet == NULL;
       curr_jump_packet = curr_jump_packet->next_jump_packet)
    ;
  dest_packet
    = (curr_jump_packet == NULL ? NULL : curr_jump_packet->dest_packet);
  /* Make connection by backward jumps and to additional exit
     packets.  */
  for (curr_jump_packet = *jump_packet_ptr;
       curr_jump_packet != NULL;
       curr_jump_packet = curr_jump_packet->next_jump_packet)
    if (curr_jump_packet->dest_packet == NULL
	&& start != curr_jump_packet->label)
      {
	if (dest_packet == NULL)
	  {
	    /* This is an exit packet */
	    dest_packet = create_dummy_packet ();
	    dest_packet->first_op->insn = new_insn (NULL, NULL, 0);
	    dest_packet->stop_packet_p = dest_packet->exit_packet_p = TRUE;
	    dest_packet->first_op->insn->exit_label
	      = curr_jump_packet->label;
	    update_regs_living_after_loop (curr_jump_packet->label);
	  }
	packet_connect (curr_jump_packet->packet, dest_packet, TRUE,
			curr_jump_packet->jump_p);
	curr_jump_packet->dest_packet = dest_packet;
      }
  return TRUE;
}

/* The following function connects PACKET with the previous packet
   *PREV_PACKET (if it is not NULL).  The function also forms chain of
   all packets with the aid of parameter LAST_PACKET.  After that it
   sets up PREV_PACKET and *LAST_PACKET onto PACKET.  */
static void
connect_packets (packet, prev_packet, last_packet)
     packet_t packet, *prev_packet, *last_packet;
{
  if (*last_packet != NULL)
    (*last_packet)->next_packet = packet;
  if (*prev_packet != NULL)
    {
      if (cond_branch_p ((*prev_packet)->first_op->insn))
	(*prev_packet)->first_op->false_target = packet;
      if (!uncond_branch_p ((*prev_packet)->first_op->insn))
	packet_connect (*prev_packet, packet, FALSE, FALSE);
    }
  *last_packet = *prev_packet = packet;
}

/* The following function connects PACKET with packets in
   LAST_JUMP_PACKET in CFG.  */
static void
connect_jump_packets (packet, last_jump_packet)
     packet_t packet;
     struct jump_packet *last_jump_packet;
{
  struct jump_packet *curr_jump_packet;

  for (curr_jump_packet = last_jump_packet;
       curr_jump_packet != NULL;
       curr_jump_packet = curr_jump_packet->next_jump_packet)
    {
      packet_connect (curr_jump_packet->packet, packet, TRUE,
		      curr_jump_packet->jump_p);
      curr_jump_packet->dest_packet = packet;
    }
}

/* The following function redirects packets in chain LAST_JUMP_PACKET
   onto LABEL.  */
static void
redirect_jump_packets (last_jump_packet, label)
     struct jump_packet *last_jump_packet;
     rtx label;
{
  struct jump_packet *curr_jump_packet;

  for (curr_jump_packet = last_jump_packet;
       curr_jump_packet != NULL;
       curr_jump_packet = curr_jump_packet->next_jump_packet)
    {
      insert_jump_packet (label, curr_jump_packet->packet,
			  curr_jump_packet->jump_p);
      /* Don't process the label on the second jump pass.  */
      curr_jump_packet->dest_packet = curr_jump_packet->packet;
    }
}

/* The following recursive function clears exit packet flag of PACKET
   and all its predecessors.  */
static void
clear_exit_packet_flag (packet)
     packet_t packet;
{
  packet_t pred;
  int pred_number;

  if (packet->visit_number1 == last_packet_visit_number1)
    return;
  packet->visit_number1 = last_packet_visit_number1;
  packet->exit_packet_p = FALSE;
  for (pred_number = 0;
       (pred = packet->pred [pred_number]) != NULL;
       pred_number++)
    clear_exit_packet_flag (pred);
}

/* The following functions unrolls loop with body [START_INSN,
   BOUND_INSN) RCSP_UNROLL_FACTOR*ITERATION_LOOK_AHEAD + 1 times and
   returns the result as interior packets graph.  Unrolled loop
   contains only copied insn so we can make anything with them without
   disturbing other parts of scheduler.  Notes which should be placed
   at the beginning of pipelined loop are returned through
   START_NOTES.  */
static struct packet *
unroll (start_insn, bound_insn, start_notes)
     rtx start_insn, bound_insn, *start_notes;
{
  rtx insn, note;
  RTX_CODE code;
  packet_t start_packet, stop_packet, packet, default_prev_packet, last_packet;
  struct jump_packet *last_jump_packet;
  varray_type pending_jump_packets;
  int pending_jump_packets_length;
  op_t op;
  int i, pred_number;
  packet_t copy, pred;
  rtx start_label, exit_label;
  rtx notes_before = NULL_RTX;
  rtx notes_after = NULL_RTX;
  insn_t new = NULL;
  char *file = NULL;
  int line = 0;

  initiate_jumps_table ();
  start_label = prev_nonnote_insn (start_insn);
  if (start_label == NULL_RTX || GET_CODE (start_label) != CODE_LABEL)
    abort ();
  for (exit_label = bound_insn;
       exit_label != NULL_RTX && GET_CODE (exit_label) == BARRIER;
       exit_label = next_nonnote_insn (exit_label))
    ;
  if (exit_label != NULL_RTX && GET_CODE (exit_label) != CODE_LABEL)
    exit_label = NULL_RTX;
  start_packet = NULL;
  stop_packet = create_dummy_packet ();
  stop_packet->first_op->insn = new_insn (NULL, NULL, 0);
  stop_packet->stop_packet_p = stop_packet->exit_packet_p = TRUE;
  stop_packet->main_exit_p = TRUE;
  stop_packet->first_op->insn->exit_label = exit_label;
  default_prev_packet = last_packet = NULL;
  VARRAY_GENERIC_PTR_INIT (pending_jump_packets, 10, "pending jump packets");
  pending_jump_packets_length = 0;
  *start_notes = NULL;
  for (insn = start_insn; insn != NULL_RTX; insn = PREV_INSN (insn))
    if (GET_CODE (insn) == NOTE && NOTE_LINE_NUMBER (insn) > 0)
      {
	line = NOTE_LINE_NUMBER (insn);
	file = (char *) NOTE_SOURCE_FILE (insn);
	break;
      }
  /* First pass: creation of packets and forward connection of them.  */
  for (insn = start_insn; insn != bound_insn; insn = NEXT_INSN (insn))
    {
      code = GET_CODE (insn);
      
      if (code == INSN || (code == JUMP_INSN && !simplejump_p (insn)))
	{
	  new = new_insn (insn, file, line);
	  new->insn = copy_insn (new->insn);
	  new->notes_before = notes_before;
	  new->notes_after = notes_after;
	  notes_before = notes_after = NULL_RTX;
	  packet = packet_create (-1);
	  op = op_create (new, 0);
	  packet_append_op (packet, op, TRUE);
	  connect_packets (packet, &default_prev_packet, &last_packet);
	  while (pending_jump_packets_length != 0)
	    {
	      pending_jump_packets_length--;
	      connect_jump_packets
		(packet,
		 (struct jump_packet *)
		 VARRAY_GENERIC_PTR (pending_jump_packets,
				     pending_jump_packets_length));
	    }
	  if (start_packet == NULL)
	    start_packet = packet;
	  if (code == JUMP_INSN && !simplejump_p (insn))
	    {
	      op->target_fixed_p = !cond_jump_p (insn);
	      op->can_be_converted_p = !op->target_fixed_p;
	      if (op->can_be_converted_p)
		{
		  /* ??? Now it is switched off.  See comments for
                     `convert_cond_branch'.  */
		  op->can_be_converted_p = FALSE;
#if 0
		  if (!invert_exp (PATTERN (insn), insn))
		    op->can_be_converted_p = FALSE;
		  else if (!invert_exp (PATTERN (insn), insn))
		    /* This should just be putting it back the way it
                       was.  */
		    abort ();
#endif
		}
	      insert_jump_packet (JUMP_LABEL (insn), packet, TRUE);
	      if (uncond_jump_p (insn))
		default_prev_packet = NULL;
	    }
	}
      else if (code == JUMP_INSN)
	{
	  if (!simplejump_p (insn))
	    abort ();
	  if (default_prev_packet == NULL)
	    {
	      /* Jump to Jump.  */
	      if (pending_jump_packets_length == 0)
		abort ();
	      while (pending_jump_packets_length != 0)
		{
		  pending_jump_packets_length--;
		  redirect_jump_packets
		    ((struct jump_packet *)
		     VARRAY_GENERIC_PTR (pending_jump_packets,
					 pending_jump_packets_length),
		     JUMP_LABEL (insn));
		}
	    }
	  else
	    insert_jump_packet (JUMP_LABEL (insn), default_prev_packet, FALSE);
	  default_prev_packet = NULL;
	}
      else if (code == NOTE)
	{
	  if (NOTE_LINE_NUMBER (insn) > 0)
	    {
	      line = NOTE_LINE_NUMBER (insn);
	      file = (char *) NOTE_SOURCE_FILE (insn);
	    }
	  else if (NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_CONT
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_VTOP)
	    {
	      /* Software pipelined loop are too modified so there is
		 no sense to save such notes.  */
#if 0
	      note = rtx_alloc (NOTE);
	      INSN_UID (note) = INSN_UID (insn);
	      NOTE_SOURCE_FILE (note) = NOTE_SOURCE_FILE (insn);
	      NOTE_LINE_NUMBER (note) = NOTE_LINE_NUMBER (insn);
	      notes_before = alloc_EXPR_LIST (REG_SAVE_NOTE, note,
					      notes_before);
#endif
	    }
	  else if (NOTE_LINE_NUMBER (insn) == NOTE_INSN_DELETED
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_BLOCK_BEG
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_BLOCK_END)
	    {
	      note = rtx_alloc (NOTE);
	      INSN_UID (note) = INSN_UID (insn);
	      NOTE_SOURCE_FILE (note) = NOTE_SOURCE_FILE (insn);
	      NOTE_LINE_NUMBER (note) = NOTE_LINE_NUMBER (insn);
	      *start_notes = alloc_EXPR_LIST (REG_SAVE_NOTE, note,
					      *start_notes);
	    }
	  else if (NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_BEG
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_END
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_EH_REGION_BEG
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_EH_REGION_END
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_SETJMP)
	    abort ();
	}
      else if (code == CODE_LABEL)
	{
	  unsigned size = VARRAY_SIZE (pending_jump_packets);
	  
	  if ((unsigned) pending_jump_packets_length >= size)
	    {
	      size *= 2;
	      if (size < (unsigned) pending_jump_packets_length)
		size = (unsigned) pending_jump_packets_length;
	      VARRAY_GROW (pending_jump_packets, size);
	    }
	  VARRAY_GENERIC_PTR (pending_jump_packets,
			      pending_jump_packets_length)
	    = (void *) find_jump_packet (insn);
	  pending_jump_packets_length++;
	}
      else if (code != BARRIER)
	/* We could nullify DEFAULT_PREV_PACKET here.  But
           unfortunately the barrier for the last loop jump is not in
           the loop part.  */
	abort ();
    }
  global_regs_living_after_loop
    = INIT_REG_SET ((bitmap) rcsp_malloc (sizeof (bitmap_head)));
  CLEAR_REG_SET (global_regs_living_after_loop);
  if (default_prev_packet != NULL)
    update_regs_living_after_loop (bound_insn);
  last_jump_packet = (exit_label == NULL_RTX
		      ? NULL : find_jump_packet (exit_label));
#if 0
  if (default_prev_packet == NULL && last_jump_packet == NULL)
    /* It can not be for infinite loop or loop without main exit.  */
    abort ();
#endif
  connect_packets (stop_packet, &default_prev_packet, &last_packet);
  connect_jump_packets (stop_packet, last_jump_packet);
  /* The 2nd pass: connection of packets by backward jumps and exit
     packets and find registers living after the loop.  */
  traverse_jumps_table (second_jumps_pass, start_label);
  /* Find exit packets: */
  for (packet = start_packet; packet != NULL; packet = packet->next_packet)
    packet->exit_packet_p = TRUE;
  last_packet_visit_number1 = 0;
  for (last_jump_packet = find_jump_packet (start_label);
       last_jump_packet != NULL;
       last_jump_packet = last_jump_packet->next_jump_packet)
    clear_exit_packet_flag (last_jump_packet->packet);
  /* Set up live_reg_set_p for the ops and restore visit numbers.  */
  for (packet = start_packet; packet != NULL; packet = packet->next_packet)
    {
      if (packet->stop_packet_p && !packet->exit_packet_p)
	abort ();
      packet->visit_number1 = -1;
      if (!packet->stop_packet_p)
	packet->first_op->insn->live_reg_set_p
	  = insn_sets_live_reg_p (packet->first_op->insn->insn);
    }
  /* Repeat it RCSP_UNROLL_FACTOR * ITERATION_LOOK_AHEAD times.  */
  for (i = 1; i < RCSP_UNROLL_FACTOR * iteration_look_ahead + 1; i++)
    {
      for (packet = start_packet; packet != NULL; packet = packet->next_packet)
	{
	  if (packet->exit_packet_p)
	    {
	      for (pred_number = 0;
		   (pred = packet->pred [pred_number]) != NULL;
		   pred_number++)
		if (pred->next_frontier_packet != NULL)
		  /* We do not change {true, false} target in possible
		     branch because it is the same after copying.  */
		  packet_connect (pred->next_frontier_packet, packet,
				  FALSE, FALSE);
	    }
	  else
	    {
	      /* Form list of copies: next_frontier_packet is used as
		 the last copy of the packet.  */
	      copy = packet_copy (packet);
	      new = new_insn (packet->first_op->insn->orig_insn,
			      packet->first_op->insn->file,
			      packet->first_op->insn->line);
	      if (packet->first_op->insn->notes_before != NULL_RTX)
		new->notes_before
		  = copy_rtx (packet->first_op->insn->notes_before);
	      if (packet->first_op->insn->notes_after != NULL_RTX)
		new->notes_after
		  = copy_rtx (packet->first_op->insn->notes_after);
	      copy->first_op->insn = new;
	      copy->first_op->insn->live_reg_set_p
		= packet->first_op->insn->live_reg_set_p;
	      if (packet->next_frontier_packet == NULL)
		packet->copy = packet->next_frontier_packet = copy;
	      else
		{
		  packet->next_frontier_packet->copy = copy;
		  packet->next_frontier_packet = copy;
		}
	      copy->min_it_n = i;
	      copy->first_op->it_num = i;
	      copy->first_op->insn->insn
		= copy_insn (packet->first_op->insn->orig_insn);
	      for (pred_number = 0;
		   (pred = packet->pred [pred_number]) != NULL;
		   pred_number++)
		{
		  packet_change_dest (pred->next_frontier_packet,
				      packet, packet->next_frontier_packet);
		  packet_change_src (packet->next_frontier_packet, pred,
				     pred->next_frontier_packet);
		}
	    }
	}
    }
  /* Connect iterations.  */
  for (packet = start_packet, i = 1;
       i < RCSP_UNROLL_FACTOR * iteration_look_ahead + 1;
       i++, packet = packet->copy)
    for (last_jump_packet = find_jump_packet (start_label);
	 last_jump_packet != NULL;
	 last_jump_packet = last_jump_packet->next_jump_packet)
      {
	packet_connect (last_jump_packet->packet, packet->copy, TRUE,
			last_jump_packet->jump_p);
	last_jump_packet->packet = last_jump_packet->packet->copy;
      }
  finish_jumps_table ();
  CLEAR_REG_SET (global_regs_living_after_loop);
  rcsp_free (global_regs_living_after_loop);
  VARRAY_FREE (pending_jump_packets);
  return start_packet;
}

/* The following function set up inheritance relation between LOC and
   already processed locations.  */
static void
set_loc_inheritance (loc)
     struct loc *loc;
{
  RTX_CODE code = GET_CODE (loc->loc);
  int regno, curr_regno, nregs, curr_nregs;
  struct loc *curr_loc;
  
  if (code == REG && REGNO (loc->loc) < FIRST_PSEUDO_REGISTER)
    {
      regno = REGNO (loc->loc);
      nregs = HARD_REGNO_NREGS (regno, GET_MODE (loc->loc));
      for (curr_loc = hard_reg_loc_chain;
	   curr_loc != NULL;
	   curr_loc = curr_loc->next_hard_reg)
	{
	  curr_regno = REGNO (curr_loc->loc);
	  curr_nregs = HARD_REGNO_NREGS (curr_regno, GET_MODE (curr_loc->loc));
	  if ((regno != curr_regno || nregs != curr_nregs)
	      && ((regno >= curr_regno && regno < curr_regno + curr_nregs)
		  || (curr_regno >= regno && curr_regno < regno + nregs)))
	    {
	      if (regno >= curr_regno
		  && regno + nregs <= curr_regno + curr_nregs)
		loc->supers = new_loc_link (curr_loc, loc->supers);
	      else
		loc->subs = new_loc_link (curr_loc, loc->subs);

	      if (curr_regno >= regno
		  && curr_regno + curr_nregs <= regno + nregs)
		curr_loc->supers = new_loc_link (loc, curr_loc->supers);
	      else
		curr_loc->subs = new_loc_link (loc, curr_loc->subs);
	    }
	}
    }
  else if (code == MEM)
    {
      for (curr_loc = mem_loc_chain;
	   curr_loc != NULL;
	   curr_loc = curr_loc->next_mem)
	if (!mems_in_disjoint_alias_sets_p (loc->loc, curr_loc->loc))
	  {
	    loc->subs = new_loc_link (curr_loc, loc->subs);
	    curr_loc->subs = new_loc_link (loc, curr_loc->subs);
	  }
      all_memory_loc->subs = new_loc_link (loc, all_memory_loc->subs);
      loc->supers = new_loc_link (all_memory_loc, loc->supers);
    }
}

/* The following function returns TRUE if INSN besides (un)conditional
   jump which can be in software pipelined loop.  */
static int
nonjump_p (insn)
     rtx insn;
{
  /* ??? Now we do not pipeline loops with function calls and asm
     insns (see comments for may_be_pipelined_p).  */
  if (GET_CODE (insn) == JUMP_INSN
      || GET_CODE (insn) == CALL_INSN
      || GET_CODE (PATTERN (insn)) == ASM_INPUT
      || asm_noperands (PATTERN (insn)) >= 0)
    return FALSE;
  return TRUE;
}

/* The following function returns TRUE if INSN contains unconditional
   jump which can be in software pipelined loop.  */
static int
uncond_jump_p (insn)
     rtx insn;
{
  rtx x = PATTERN (insn);

  if (GET_CODE (x) == PARALLEL)
    x = XVECEXP (x, 0, 0);

  if (GET_CODE (x) != SET)
    return FALSE;
  if (GET_CODE (SET_DEST (x)) != PC)
    return FALSE;
  if (GET_CODE (SET_SRC (x)) == LABEL_REF)
    return TRUE;
  return FALSE;
}

/* The following function returns TRUE if INSN contains conditional
   branch which can be in software pipelined loop.  */
static int
cond_jump_p (insn)
     rtx insn;
{
  register rtx x = PATTERN (insn);

  if (GET_CODE (x) == PARALLEL)
    x = XVECEXP (x, 0, 0);

  if (GET_CODE (x) != SET)
    return FALSE;
  if (GET_CODE (SET_DEST (x)) != PC)
    return FALSE;
  if (GET_CODE (SET_SRC (x)) != IF_THEN_ELSE)
    return FALSE;
  if (XEXP (SET_SRC (x), 2) == pc_rtx
      && (GET_CODE (XEXP (SET_SRC (x), 1)) == LABEL_REF
	  || GET_CODE (XEXP (SET_SRC (x), 1)) == RETURN))
    return TRUE;
  if (XEXP (SET_SRC (x), 1) == pc_rtx
      && (GET_CODE (XEXP (SET_SRC (x), 2)) == LABEL_REF
	  || GET_CODE (XEXP (SET_SRC (x), 2)) == RETURN))
    return TRUE;
  return FALSE;
}

/* The following macro defines maximal size (in insns) of loop on
   which RCSP will be tried.  */
#ifndef RCSP_MAX_LOOP_SIZE
#define RCSP_MAX_LOOP_SIZE 20
#endif

/* The following function returns TRUE if we may pipeline the loop
   with body [START_INSN, BOUND_INSN) and FALSE otherwise.  Now there
   are the following constraints for pipelined loops:
     1. Machine does not use cc0.
     2. Loop body insns do not refer for volatile memory.
     3. Loop body insns do not contain asm operands, asm inputs, or
        unspecified volatile operations.
     4. There are no CALL, SETJUMP, computed jumps, and tablejumps in
        the loop body.
     5. There are no exception handlers in the loop.
     6. There is one start point of the loop.
     7. There are not multi-insn sequences (e.g. library calls) in
        the loop.
     8. Maximal loop size (in insns) <= RCSP_MAX_LOOP_SIZE.  */
static int
may_be_pipelined_p (start_insn, bound_insn)
     rtx start_insn, bound_insn;
{
  rtx insn;
  RTX_CODE code;
  int insns_number = 0;

#ifdef HAVE_cc0
  /* Is it worth to implement software pipelining for such machines???
     If yes, we should implement scheduling a group of insns in the
     software pipeliner.  */
  return FALSE;
#endif
  for (insn = start_insn;
       insn != bound_insn && insn != NULL_RTX;
       insn = NEXT_INSN (insn))
    {
      code = GET_CODE (insn);
      if (code == NOTE
	  && (NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_BEG
	      || NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_END
	      || NOTE_LINE_NUMBER (insn) == NOTE_INSN_EH_REGION_BEG
	      || NOTE_LINE_NUMBER (insn) == NOTE_INSN_EH_REGION_END
	      || NOTE_LINE_NUMBER (insn) == NOTE_INSN_SETJMP))
	return FALSE;
      else if (code == NOTE
	       && (NOTE_LINE_NUMBER (insn) > 0
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_BASIC_BLOCK
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_BLOCK_BEG
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_BLOCK_END
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_CONT
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_VTOP
		   || NOTE_LINE_NUMBER (insn) == NOTE_INSN_DELETED))
	continue;
      else if (code == BARRIER || code == CODE_LABEL)
	continue;
      else if (cond_jump_p (insn) || uncond_jump_p (insn))
	{
	  if (!simplejump_p (insn))
	    insns_number++;
	}
      else if (!nonjump_p (insn))
	return FALSE;
      else
	insns_number++;
      if (GET_RTX_CLASS (code) == 'i'
	  && (volatile_refs_p (PATTERN (insn))
	      /* We don't deal with insn sequences now.  ??? We could
                 implement it in the future.  */
	      || find_reg_note (insn, REG_LIBCALL, NULL_RTX) != NULL_RTX))
	return FALSE;
    }
  if (insn == NULL_RTX && bound_insn != NULL_RTX)
    /* This is weird loop in which bound insn is before start loop
       insn.  Currently we prevent to pipeline such loops.  ??? We
       could pipeline such loop if we correctly find bound insn from
       regions recognized by find_rgns.  */
    return FALSE;
  return insns_number > 0 && insns_number <= RCSP_MAX_LOOP_SIZE;
}

/* The following is start packet and exit packet of unrolled loop.  */
static packet_t first_loop_packet;

/* The following is notes which will be placed at the beginning of
   pipelined loop.  */
static rtx start_loop_notes;

/* The most of memory needed for the pipeliner is allocated here.  */
static struct obstack rcsp_memory;

/* The following function initiates work of the software pipeliner
   with look ahead K (in iterations) on loop with body [START_INSN,
   BOUND_INSN).  */
static void
start_RCSP_interface (start_insn, bound_insn, k)
     rtx start_insn, bound_insn;
     int k;
{
  int i;
  struct loc *curr_loc;

  iteration_look_ahead = k;
  rcsp_debug_level = (sched_verbose == 0 ? 0 : sched_verbose + 1);
#ifndef RCSP_DO_NOT_USE_OBSTACK
  obstack_init (&rcsp_memory);
#endif
  VARRAY_GENERIC_PTR_INIT (insn_reads, 10, "insn_reads");
  VARRAY_GENERIC_PTR_INIT (insn_possible_writes, 10, "insn_possible_writes");
  VARRAY_GENERIC_PTR_INIT (insn_write_barriers, 10, "insn_write_barriers");
  VARRAY_GENERIC_PTR_INIT (insn_writes, 10, "insn_writes");
  insns_length = 0;
  VARRAY_GENERIC_PTR_INIT (insns, 100, "insns");
  start_pipeline_data ();
  first_loop_packet = unroll (start_insn, bound_insn, &start_loop_notes);
  insn_locs_length = insns_length;
  insn_locs = rcsp_malloc (insn_locs_length * sizeof (struct loc *));
  memset ((char *) insn_locs, 0, insn_locs_length * sizeof (struct loc *));
  all_memory_loc = new_loc (gen_rtx_MEM (VOIDmode, gen_rtx_REG (Pmode, 0)),
			    0, 0, NULL);
  hard_reg_loc_chain = mem_loc_chain = NULL;
  /* Create locs.  */
  for (i = 0; i < insns_length; i++)
    if (((insn_t) VARRAY_GENERIC_PTR (insns, i))->insn != NULL_RTX)
      process_insn ((insn_t) VARRAY_GENERIC_PTR (insns, i));
  /* Make subs and supers.  */
  for (i = 0; i < insn_locs_length; i++)
    for (curr_loc = insn_locs [i];
	 curr_loc != NULL;
	 curr_loc = curr_loc->next_insn_loc)
      set_loc_inheritance (curr_loc);
}

/* The following recursive function generates labels which are in jump
   insns of PACKET and of packets which are achievable from PACKET.
   The function also connects the new labels to corresponding
   insns.  */
static void
add_jump_labels (packet)
     packet_t packet;
{
  op_t curr_op;
  packet_t target, succ;
  int succ_number;
  rtx nlabel;

  if (packet->visit_number1 == last_packet_visit_number1)
    return;
  packet->visit_number1 = last_packet_visit_number1;
  for (curr_op = packet->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    if (curr_op->true_target != NULL || curr_op->false_target != NULL)
      {
	target = curr_op->true_target;
	if (cond_branch_p (curr_op->insn) && !curr_op->true_jump_p)
	  target = curr_op->false_target;
	if (target->stop_packet_p
	    && target->first_op->insn->exit_label != NULL_RTX)
	  /* Another branch may be exit from the loop if we permit
             moving one branch insn through another.  */
	  nlabel = target->first_op->insn->exit_label;
	else
	  {
	    nlabel = gen_label_rtx ();
	    target->first_op->insn->notes_before
	      = alloc_EXPR_LIST (REG_SAVE_NOTE, nlabel,
				 target->first_op->insn->notes_before);
	  }
	if (JUMP_LABEL (curr_op->insn->insn) == NULL_RTX)
	  abort ();
	if (!redirect_jump (curr_op->insn->insn, nlabel, FALSE))
	  abort ();
	LABEL_NUSES (JUMP_LABEL (curr_op->insn->insn))--;
      }
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    add_jump_labels (succ);
}

/* The following variable values is file name and line number
   corresponding to the last output insn in emit_packet.  */
static char *output_last_file;
static int output_last_line; 

/* The following recursive function emits list of NOTES in reverse
   order.  */
static void
emit_notes_list (notes)
     rtx notes;
{
  if (notes == NULL_RTX)
    return;
  emit_notes_list (XEXP (notes, 1));
  if (GET_CODE (XEXP (notes, 0)) == NOTE)
    emit_note (NOTE_SOURCE_FILE (XEXP (notes, 0)),
	       NOTE_LINE_NUMBER (XEXP (notes, 0)));
  else
    emit (XEXP (notes, 0));
}

/* The following recursive function outputs insns of PACKET and all
   packets which are achievable from PACKET.  The function returns
   main loop exit packet if it is achievable from PACKET, NULL
   otherwise.  */
static packet_t
emit_packet (packet)
     packet_t packet;
{
  op_t curr_op;
  rtx insn;
  packet_t succ, pred, result, next_pc_packet, main_exit;
  int succ_number, pred_number;

  if (packet->visit_number1 == last_packet_visit_number1)
    return NULL;
  packet->visit_number1 = last_packet_visit_number1;
  for (curr_op = packet->first_op; curr_op != NULL; curr_op = curr_op->next_op)
    {
      if (curr_op->insn->file != NULL
	  && curr_op->insn->file != NULL
	  && (curr_op->insn->file != output_last_file
	      || curr_op->insn->line != output_last_line))
	{
	  output_last_file = curr_op->insn->file;
	  output_last_line = curr_op->insn->line;
	  emit_note (output_last_file, output_last_line);
	}
      emit_notes_list (curr_op->insn->notes_before);
      if (curr_op->insn->insn != NULL)
	{
	  /* Output non dummy insn.  */
	  emit (PATTERN (curr_op->insn->insn));
	  if (GET_CODE (curr_op->insn->insn) == JUMP_INSN
	      && JUMP_LABEL (curr_op->insn->insn) != NULL_RTX)
	    {
	      /* Function emit may emit also BARRIER.  */
	      insn = get_last_insn ();
	      if (GET_CODE (insn) == BARRIER)
		insn = prev_nonnote_insn (insn);
	      JUMP_LABEL (insn) = JUMP_LABEL (curr_op->insn->insn);
	    }
	}
      emit_notes_list (curr_op->insn->notes_after);
    }
  result = NULL;
  next_pc_packet = NULL;
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    {
      curr_op = packet_find_branch (packet, succ, TRUE);
      if (curr_op == NULL &&
	  (!succ->stop_packet_p || !succ->main_exit_p))
	{
	  /* Output next pc packet but the exit packet.  */
	  result = emit_packet (succ);
	  next_pc_packet = succ;
	  break;
	}
    }
  for (succ_number = 0;
       (succ = packet->succ [succ_number]) != NULL;
       succ_number++)
    {
      /* Remember: We should output main exit as the last packet.  */
      main_exit = NULL;
      if (succ->stop_packet_p && succ->main_exit_p)
	main_exit = succ;
      else if (succ != next_pc_packet)
	{
	  /* Find pc predecessor of succ.  */
	  for (pred_number = 0;
	       (pred = succ->pred [pred_number]) != NULL;
	       pred_number++)
	    {
	      curr_op = packet_find_branch (pred, succ, TRUE);
	      if (curr_op == NULL)
		/* pred->succ is not a jump.  */
		break;
	    }
	  if (pred == packet)
	    /* We already output the next pc packet.  */
	    abort ();
	  if (pred == NULL)
	    /* There are only jumps onto succ or  */
	    main_exit = emit_packet (succ);
	}
      if (main_exit != NULL)
	{
	  if (result != NULL && result != main_exit)
	    abort ();
	  result = main_exit;
	}
    }
  return result;
}

/* The following function calls pipeliner with DEBUG_FILE.  It should
   be called only after the pipeliner interface initialization.  */
static rtx
call_RCSP (debug_file)
     FILE *debug_file;
{
  rtx temp, since, first_insn;
  packet_t res, main_exit;

  since = rtx_alloc (CONST_INT);
  first_insn = NULL;
  if (first_loop_packet != NULL && first_loop_packet->first_op != NULL)
    first_insn = first_loop_packet->first_op->insn->insn;
  res = pipeline (first_loop_packet, debug_file, ";; ");
  if (res == NULL)
    return NULL;
  start_sequence ();
  last_packet_visit_number1++;
  add_jump_labels (res);
  output_last_file = NULL;
  output_last_line = -1;
  for (; first_insn != NULL_RTX; first_insn = PREV_INSN (first_insn))
    if (GET_CODE (first_insn) == NOTE && NOTE_LINE_NUMBER (first_insn) > 0)
      {
	output_last_line = NOTE_LINE_NUMBER (first_insn);
	output_last_file = (char *) NOTE_SOURCE_FILE (first_insn);
	break;
      }
  emit_note (NULL, NOTE_SOFTWARE_PIPELINED_LOOP_BEGIN);
  emit_notes_list (start_loop_notes);
  /* Emit pipelined loop insns.  */
  last_packet_visit_number1++;
  main_exit = emit_packet (res);
  if (main_exit != NULL)
    /* Main exit can be unaccessible for infinite loops.  */
    emit_packet (main_exit);
  emit_note (NULL, NOTE_SOFTWARE_PIPELINED_LOOP_END);
  temp = get_insns ();
  end_sequence ();
  return temp;
}

/* The following function finishes work of the software pipeliner on
   the current loop.  */
static void
finish_RCSP_interface (void)
{
  int i;

  finish_pipeline_data ();
  VARRAY_FREE (insns);
  VARRAY_FREE (insn_writes);
  VARRAY_FREE (insn_write_barriers);
  VARRAY_FREE (insn_possible_writes);
  VARRAY_FREE (insn_reads);
  free_loc (all_memory_loc);
  for (i = 0; i < insn_locs_length; i++)
    free_insn_locs (insn_locs [i]);
  rcsp_free (insn_locs);
#ifndef RCSP_DO_NOT_USE_OBSTACK
  obstack_free (&rcsp_memory, NULL);
#endif
}



/* This page contains interface of software pipeliner to GCC.  All
   these functions are called by software pipeliner (abstract data)
   and should be implemented.  */

/* Allocate memory by RCSP.  */
static void *
rcsp_malloc (size)
     size_t size;
{
#ifndef RCSP_DO_NOT_USE_OBSTACK
  void *res;

  obstack_blank (&rcsp_memory, size);
  res = obstack_base (&rcsp_memory);
  obstack_finish (&rcsp_memory);
  return res;
#else
  return xmalloc (size);
#endif
}

/* Reallocate memory by RCSP.  */
static void *
rcsp_realloc (addr, old_size, new_size)
     void *addr;
     size_t old_size ATTRIBUTE_UNUSED;
     size_t new_size;
{
#ifndef RCSP_DO_NOT_USE_OBSTACK
  void *res;

  res = rcsp_malloc (new_size);
  memcpy (res, addr, (old_size < new_size ? old_size : new_size));
  return res;
#else
  return xrealloc (addr, new_size);
#endif
}

/* Free memory allocated by RCSP.  */
static void
rcsp_free (addr)
#ifdef RCSP_DO_NOT_USE_OBSTACK
     void *addr;
#else
     void *addr ATTRIBUTE_UNUSED;
#endif
{
#ifdef RCSP_DO_NOT_USE_OBSTACK
  free (addr);
#endif
}

/* The following function returns TRUE if INSN is a software pipeline
   insn besides (un)conditional branch.  */
static int
nonbranch_p (insn)
     insn_t insn;
{
  return nonjump_p (insn->insn);
}

/* The following function returns TRUE if software pipeliner's INSN
   contains a unconditional branch.  */
static int
uncond_branch_p (insn)
     insn_t insn;
{
  return uncond_jump_p (insn->insn);
}

/* The following function returns TRUE if software pipeliner's INSN
   contains a conditional branch.  */
static int
cond_branch_p (insn)
     insn_t insn;
{
  return cond_jump_p (insn->insn);
}

/* The following function reverses conditional branch INSN, i.e. jump
   and fall trough of the branch is exchanged by reversing the branch
   condition.  The function should make it successfully for branch
   with can_be_converted_p == TRUE.  */
static void
convert_cond_branch (insn)
     insn_t insn ATTRIBUTE_UNUSED;
{
  rtx cond ATTRIBUTE_UNUSED;

  /* ??? Now we marks all condtional branch as unconverable.  So we do
     not call this function.  We need implement this function if we
     are going to schedule jumps added to make CFG jump consistent.
     But is it really necessary.  I think the current approach is not
     bad because the added jump with preceding conditional branch
     (this is a frequent case) is usually transformed into one
     conditional branch by jump optimizer.  */
  abort ();
#if 0
  if (!cond_jump_p (insn->insn))
    abort ();
  cond = XEXP (SET_SRC (PATTERN (insn->insn)), 0);
  /* ??? We still need code for change validation.  */
  XEXP (SET_SRC (PATTERN (insn->insn)), 0)
    = gen_rtx (reverse_condition (GET_CODE (cond)), VOIDmode,
	       XEXP (cond, 0), XEXP (cond, 1));
#endif
}

/* The following function returns new unconditional branch onto insn
   TARGET.  The important requiremnt for the branch is that it should
   not create new data dependencies.  */
static insn_t
create_uncond_branch (target)
     insn_t target;
{
  rtx label;
  insn_t insn;

  if (target->exit_label != NULL_RTX)
    {
      insn = new_insn (make_jump_insn_raw (gen_jump (target->exit_label)),
		       NULL, 0);
      JUMP_LABEL (insn->insn) = target->exit_label;
    }
  else
    {
      /* We generate labels for such jumps in add_jump_labels.  */
      label = gen_label_rtx ();
      insn = new_insn (make_jump_insn_raw (gen_jump (label)), NULL, 0);
      JUMP_LABEL (insn->insn) = label;
    }
  return insn;
}

/* The following function returns TRUE if we need jump onto PACKET even
   there is only one packet predecessor without unconditional jump.  Now
   this is true for all loop exits.  We make the jump even for main loop exit
   (which correspond the first insn after the loop) although jump in
   this case is not necessary.  Fortunately, such jump is removed by
   jump optimizer.  */
static int
dest_packet_requires_jump_p (packet)
     packet_t packet;
{
  if (packet->first_op->insn->exit_label != NULL_RTX)
    {
      /* Here we can say that insn is in a stop packet.  */
      if (packet->first_op->insn->insn != NULL)
	/* It should be non-dummy insn in a stop packet.  */
	abort ();
      return TRUE;
    }
  return FALSE;
}

/* The following function returns TRUE if INSN1 and INSN2 are the same
   insn possibly from different iterations.  */
static int
insn_eq_p (insn1, insn2)
     insn_t insn1, insn2;
{
  return insn1->orig_insn == insn2->orig_insn;
}

/* The folowing function returns hash code of INSN.  The same insns
   from different iterations have the same hash code.  */
static unsigned
insn_hash (insn)
     insn_t insn;
{
  return (unsigned) insn->orig_insn;
}

/* The following function returns (compact if DENSE_P) symbolic INSN
   representation.  For brevity we simply return UID of INSN.  The
   representation stores in static memory so the next function call
   may rewrite it.  */
static const char *
insn_name (insn, dense_p)
     insn_t insn;
     int dense_p;
{
  if (insn == NULL || insn->insn == NULL_RTX)
    return (dense_p ? "dummy" : "dummy op");
  else
    {
      static char name [20];
      
      sprintf (name, (dense_p ? "%d" : "%4d"), INSN_UID (insn->insn));
      return name;
    }
}

/* The following function returns TRUE if it is dummy insn (which
   should be ignored in many cases) for software pipeliner.  Dummy
   insns are placed in stop packets and in the root packet.  */
static int
insn_dummy_p (insn)
     insn_t insn;
{
  return insn == NULL || insn->insn == NULL_RTX;
}

/* The following function makes and returns copy of RCSP INSN.  */
static insn_t
insn_copy (insn)
     insn_t insn;
{
  insn_t copy;

  copy = new_insn (insn->orig_insn, insn->file, insn->line);
  if (insn->notes_before != NULL_RTX)
    copy->notes_before = copy_rtx (insn->notes_before);
  if (insn->notes_after != NULL_RTX)
    copy->notes_after = copy_rtx (insn->notes_after);
  copy->live_reg_set_p = insn->live_reg_set_p;
  copy->insn = copy_insn (insn->insn);
  return copy;
}

/* The following function returns latency time on dependency SRC->DEST
   of type DEP_MODE.  */
static int
insn_dep_time (src, dest, dep_mode)
     insn_t src, dest;
     int dep_mode;
{
  recog_memoized (src->insn);
  recog_memoized (dest->insn);

  if (INSN_CODE (src->insn) < 0 || INSN_CODE (dest->insn) < 0)
    return 0;
  if (dep_mode == USE_DEF)
    return 0;
  else if (dep_mode == DEF_USE)
    return (bypass_p (src->insn)
	    ? insn_latency (src->insn, dest->insn)
	    : insn_default_latency (src->insn));
  else if (dep_mode != DEF_DEF)
    abort ();
  else
    {
      int cost
	= insn_default_latency (src->insn) - insn_default_latency (dest->insn);

      return (cost < 0 ? 0 : cost);
    }
}

/* The following function returns TRUE if speculative move INSN into
   PACKET is risky.  In other words such move is prohibited.  ??? Now
   we get the information only from insn itself.  We could use
   information about context too (e.g. all path from PACKET contains
   load of the same location as INSN does).  */
static int
insn_risky_speculative_move_p (insn, packet)
     insn_t insn;
     packet_t packet  ATTRIBUTE_UNUSED;
{
  int insn_class;

  if (insn == NULL || insn->insn == NULL)
    abort ();

  if (insn->live_reg_set_p)
    return TRUE;

  insn_class = haifa_classify_insn (insn->insn);

  /* Handle non-load insns.  */
  switch (insn_class)
    {
    case TRAP_FREE:
      return FALSE;
    case TRAP_RISKY:
      return TRUE;
    default:;
    }

  /* Handle loads.  */
  if (!flag_schedule_speculative_load)
    return TRUE;
  switch (insn_class)
    {
    case IFREE:
      return FALSE;
    case IRISKY:
      return TRUE;
    default:;
    }
  return TRUE;
}

/* The following function initializes CPU_STATE as if all functional
   units were free.  */
static void
cpu_state_reset (cpu_state)
     cpu_state_t cpu_state;
{
  state_reset (cpu_state);
}

/* The following function returns FALSE if INSN can not be issued in
   this CPU_STATE.  Otherwise, the function returns TRUE and changes
   CPU_STATE as if INSN were issued.  */
static int
cpu_state_issue (insn, cpu_state)
     insn_t insn;
     cpu_state_t cpu_state;
{
  int delay;
#if FIRST_CYCLE_MULTIPASS_SCHEDULING && defined (SCHEDULER_BUBBLE)
  int i;
  rtx bubble;
  state_t temp_state = alloca (dfa_state_size);
#endif

  recog_memoized (insn->insn);
  if (INSN_CODE (insn->insn) < 0)
    return 1;
  delay = state_transition (cpu_state, insn->insn);
  if ( delay < 0)
    return 1;
  else if (delay > 0)
    return 0;
#if FIRST_CYCLE_MULTIPASS_SCHEDULING && defined (SCHEDULER_BUBBLE)
  for (i = 0;
       (bubble = SCHEDULER_BUBBLE (i)) != NULL_RTX;
       i++)
    {
      memcpy (temp_state, cpu_state, dfa_state_size);
      if (state_transition (temp_state, bubble) < 0
	  && state_transition (temp_state, insn->insn) < 0)
	break;
    }
  if (bubble != NULL_RTX)
    {
      memcpy (cpu_state, temp_state, dfa_state_size);
      return 1;
    }
  else
#endif
    return 0;
}

/* The following function changes CPU_STATE for cpu cycle tick.  */
static void
cpu_state_advance_cycle (cpu_state)
     cpu_state_t cpu_state;
{
#ifdef DFA_SCHEDULER_PRE_CYCLE_INSN
  state_transition (cpu_state, DFA_SCHEDULER_PRE_CYCLE_INSN);
#endif
  state_transition (cpu_state, NULL);
#ifdef DFA_SCHEDULER_POST_CYCLE_INSN
  state_transition (cpu_state, DFA_SCHEDULER_POST_CYCLE_INSN);
#endif
}

/* The following function returns size of data which represents all
   current and projected functional unit reservations.  This is simply
   DFA-based pipeline description state.  */

static int
cpu_state_size ()
{
  return state_size ();
}

/* The following function makes union CPU_STATE and CPU_STATE2.  The
   result will be placed in CPU_STATE.  */

static void
cpu_state_union (cpu_state, cpu_state2)
     cpu_state_t cpu_state ATTRIBUTE_UNUSED, cpu_state2 ATTRIBUTE_UNUSED;
{
  /* ??? Now we have no implementation of this in our DFA-based
     pipeline description.  Such implementation could result in much
     bigger automata.  So another solution could be implementation of
     CPU state as a list of automaton states.  For the most loops the
     lists will be short because merging frontier packets is not
     frequent event.  */
}

/* The following function returns TRUE if LOC1 refers for the same
   location as LOC2. ??? We need better recognition of aliases.  */

static int
loc_eq_p (loc1, loc2)
     location_t loc1, loc2;
{
  if (GET_CODE (loc1->loc) != GET_CODE (loc2->loc))
    return FALSE;
  if (GET_CODE (loc1->loc) == REG)
    {
      if (REGNO (loc1->loc) != REGNO (loc2->loc))
	return FALSE;
      if (REGNO (loc1->loc) >= FIRST_PSEUDO_REGISTER)
	return TRUE;
      return GET_MODE (loc1->loc) == GET_MODE (loc2->loc);
    }
    
  if (GET_CODE (loc1->loc) != MEM)
    abort ();
  return loc1->loc == loc2->loc;
}

/* The following function returns hash code of location LOC.  */

static unsigned
loc_hash (loc)
     location_t loc;
{
  if (GET_CODE (loc->loc) == REG)
    {
      if (REGNO (loc->loc) >= FIRST_PSEUDO_REGISTER)
	return (unsigned) REGNO (loc->loc);
      else
	return ((unsigned) REGNO (loc->loc)
		+ ((unsigned) GET_MODE (loc->loc) << 16));
    }

  if (GET_CODE (loc->loc) != MEM)
    abort ();
  return (unsigned) loc->loc;
}

/* The following function returns the first (if FIRST_P) or the next
   (otherwise) sub-location (direct or indirect) of LOC.  If there are
   no more sub-locations the function returns NULL.  */

static location_t
loc_sub (loc, first_p)
     location_t loc;
     int first_p;
{
  static struct loc_link *curr_link = NULL;

  if (first_p)
    curr_link = loc->subs;
  else if (curr_link != NULL)
    curr_link = curr_link->next;
  if (curr_link == NULL)
    return NULL;
  else
    return curr_link->loc;
}

/* The following function returns the first (if FIRST_P) or the next
   (otherwise) super-location (direct or indirect) of LOC.  If there
   are no more super-locations, the function returns NULL.  */
static location_t
loc_super (loc, first_p)
     location_t loc;
     int first_p;
{
  static struct loc_link *curr_link;

  if (first_p)
    curr_link = loc->supers;
  else if (curr_link != NULL)
    curr_link = curr_link->next;
  if (curr_link == NULL)
    return NULL;
  else
    return curr_link->loc;
}

/* The following function returns N-th locations of INSN.  Flags of
   that INSN uses the returned location is passed through USAGE_P.  If
   the location is written, type of writing is returned through
   WRITE_KIND.  Used locations should be returned the first.  */

static location_t
getloc (insn, n, usage_p, write_kind)
     insn_t insn;
     int n;
     int *usage_p;
     enum write_kind *write_kind;
{
  /* Dummy operation does not refer for locations.  */
  if (insn == NULL || insn->insn == NULL_RTX)
    abort ();
  if (n == 0)
    {
      struct loc *curr_loc;

      insn_reads_length = insn_possible_writes_length
	= insn_write_barriers_length = insn_writes_length = 0;
      for (curr_loc = insn_locs [insn->number];
	   curr_loc != NULL;
	   curr_loc = curr_loc->next_insn_loc)
	if (curr_loc->usage_p)
	  add_vector_element (&insn_reads, curr_loc, &insn_reads_length);
	else if (curr_loc->write_kind == USUAL_WRITE)
	  add_vector_element (&insn_writes, curr_loc, &insn_writes_length);
	else if (curr_loc->write_kind == POSSIBLE_WRITE)
	  add_vector_element (&insn_possible_writes, curr_loc,
			      &insn_possible_writes_length);
	else if (curr_loc->write_kind != WRITE_BARRIER)
	  abort ();
	else
	  add_vector_element (&insn_write_barriers, curr_loc,
			      &insn_write_barriers_length);
    }

  *write_kind = 0;
  *usage_p = FALSE;
  if (n < insn_reads_length)
    {
      *usage_p = TRUE;
      return (location_t) VARRAY_GENERIC_PTR (insn_reads, n);
    }
  else if ((n -= insn_reads_length) < insn_possible_writes_length)
    {
      *write_kind = 1;
      return (location_t) VARRAY_GENERIC_PTR (insn_possible_writes, n);
    }
  else if ((n -= insn_possible_writes_length) < insn_writes_length)
    return (location_t) VARRAY_GENERIC_PTR (insn_writes, n);
  else if ((n -= insn_writes_length) < insn_write_barriers_length)
    {
      *write_kind = 2;
      return (location_t) VARRAY_GENERIC_PTR (insn_write_barriers, n);
    }
  else
    return NULL;
}

#endif /* #if RCSP_SOFTWARE_PIPELINING && AUTOMATON_PIPELINE_INTERFACE */
/* END CYGNUS LOCAL -- vmakarov/RCSP */

#endif /* INSN_SCHEDULING */
